#!/usr/bin/perl

use strict;
use warnings;

use CGI;
use CGI::Carp;
use CGI::Session;

use DateTime;
use DBI;
use MIME::Entity;

use MVHub::AuthAccount;
use MVHub::AgencyFieldDefs;
use MVHub::Common;
use MVHub::Utils::ConfigSimple;

use MVHub::DieHandler;
use MVHub::Utils;

my $DEFAULT_FIELD_SIZE    = 50;
my $DEFAULT_TEXTAREA_COLS = 45;

my $FIRST_PROGRAM_TEMPLATE = 'first_program.tmpl';

# The form template, and some of the non-form template variables it
# uses:
my $AGENCY_FORM_TEMPLATE_FILE  = 'agency_form.tmpl';
my $TOP_ERROR_MESSAGE_TAG      = 'TOP_ERROR_MESSAGE_TAG';
my $IS_ADMIN_TAG               = 'IS_ADMIN_TAG';
my $IS_DATA_ENTRY_OPERATOR_TAG = 'is_data_entry_operator';
my $AGENCY_NAME_TAG            = 'agency_name';
my $USER_REAL_NAME_TAG         = 'full_name';
my $USERNAME_TAG               = 'username';

{    # main
    my $cgi = CGI->new();
    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from(
        $ENV{MV_CONFIG_FILE} );

    MVHub::Utils::clean_cgi_params($cgi);

    my $agency_id = reset_invalid_agency_id( $dbh, $cgi->param('agency_id') );
    $cgi->param( 'agency_id', $agency_id );

    my ( $is_admin, $is_agency, $is_data_entry_operator, $error_message )
        = MVHub::AuthAccount::authenticate_session($cgi);

    # Now we have:
    #   session id: either a valid alphanumeric string or empty string)
    #   agency_id (from the client): valid number or $NO_AGENCY_ID; never 0

    my $form = prepare_agency_form(
        cgi       => $cgi,
        agency_id => $agency_id,
        is_admin  => $is_admin,
        is_agency => $is_agency,
    );

    if ( !$form->submitted() ) {
        if ( $agency_id == $MVHub::Common::NO_AGENCY_ID ) {

            # send a blank form.
            MVHub::Page::print_page_and_exit( body => $form->render(), );
        }
        else {
            my $values_href = MVHub::Utils::DB::select_agency_record(
                dbh                => $dbh,
                agency_id          => $agency_id,
                allow_empty_result => 1
            );
            MVHub::Page::print_page_and_exit(
                body => $form->render( values => $values_href ), );
        }    # else ! $agency_id == $MVHub::Common::NO_AGENCY_ID
    }         # if ! form->submitted
    else {    # form is submitted
        if ( !$form->validate() or !validate_more($form) ) {
            my $announcement = << "MSG";
	    <p><b>We had problem(s) with your submission.
	    Please see the message(s) highlighted below.</b></p>
MSG
            if ( $agency_id == $MVHub::Common::NO_AGENCY_ID && !$is_admin ) {

                # warn  HTML standards emptied password field
                $announcement .= << "MSG";
		<p><b>Also, for the security 
                    of your account, please
		    re-enter your passwords.
		</b></p>
MSG
            }
            $form->tmpl_param( $TOP_ERROR_MESSAGE_TAG, $announcement );
            MVHub::Page::print_page_and_exit( body => $form->render() );
        }
        else {    # $form_validate or $validate_more
            my $values_href = $form->field();
            prepare_record_for_db($values_href);

            # NO_AGENCY_ID, --> insert new record
            if ( $agency_id == $MVHub::Common::NO_AGENCY_ID ) {
                $agency_id = insert_agency_record($values_href);
                if ( !$is_admin ) {    # an agency submitted a new record
                    send_admin_notification($values_href);
                    send_confirmation_email( $cfg, $values_href );

                    # Log them in, and display the First Program wizard page
                    my $tmpl
                        = HTML::Template->new(
                        filename => $FIRST_PROGRAM_TEMPLATE )
                        or die
                        "Could not open template: $FIRST_PROGRAM_TEMPLATE";
                    $tmpl->param( agency_id => $agency_id );
                    $tmpl->param(
                        program_id => $MVHub::Common::NO_PROGRAM_ID );
                    MVHub::Page::print_page_and_exit(
                        cookie => MVHub::AuthAccount::init_session(
                            caller_id => $agency_id
                        ),
                        body => $tmpl->output()
                    );
                }         # if is_admin
                else {    # is_adnin
                       # Redirect to the agency home page. If we're not admin,
                     # then we'll need to create a session (i.e. log the user in,
                     # in the process.
                    MVHub::Common::display_agency_home_page_and_exit(
                        agency_id    => $agency_id,
                        init_session => 0,
                        message      => "Your record has been saved."
                    );
                }    # else $is_admin
            }         # if NO_AGENCY_ID
            else {    # UPDATE the agency record
                update_agency_record( $values_href, $agency_id );

                MVHub::Common::display_agency_home_page_and_exit(
                    agency_id    => $agency_id,
                    init_session => 0,
                    message      => "Your record has been updated."
                );
            }    # else UPDATE the agency record
        }    # else form->validate() and validate_more($form)
    }    # else form is submitted
}    # main

sub send_confirmation_email {
    my ( $cfg, $values_href ) = @_;

    my %params = build_confirmation_email_params( $cfg, $values_href );
    my $top = MIME::Entity->build(
        Type    => "text/plain",
        From    => $params{from},
        To      => $params{to},
        Subject => $params{subject},
        Data    => $params{body}
    );

    MVHub::Common::sendmail($top);
}

sub build_confirmation_email_params {
    my ( $cfg, $values_href ) = @_;

    my $from = $cfg->param('NOTIFICATION.team_name')
        or die "bad team name";
    my $admin_email = $cfg->param('NOTIFICATION.admin_email')
        or die "bad admin_email";
    $from .= " <$admin_email>";

    my $to = $values_href->{contact_email}
        or die "bad contact_email";

    my $subject = "$values_href->{agency_name} added to: ";
    $subject .= $cfg->param('SITE.website_name');

    my $body = build_confirmation_email_body( $cfg, $values_href );
    my %params
        = ( from => $from, to => $to, subject => $subject, body => $body, );
    return %params;

}

sub build_confirmation_email_body {
    my ( $cfg, $values_href ) = @_;

    my $website_name = $cfg->param('SITE.website_name');
    my $team_name    = $cfg->param('NOTIFICATION.team_name');

    my $body = << "BODY";
Hello $values_href->{contact_first_name}

$values_href->{agency_name} was just added to $website_name

You are now listed as our contact for $values_href->{agency_name}.
If you didn't enter this information and feel that you should 
not be the contact for this agency, please contact us.

For future reference, your agency's password is: 

    $values_href->{password}

Thank you for registering with $website_name.

$team_name;
BODY

    return $body;
}

# set empty or missing $agency_id to special CONSTANT
sub reset_invalid_agency_id {
    my ( $dbh, $agency_id ) = @_;
    my $is_valid_agency_id = MVHub::Utils::DB::select_agency_record(
        'dbh'                => $dbh,
        'agency_id'          => $agency_id,
        'allow_empty_result' => 1
    );

    if ( !$agency_id || !$is_valid_agency_id ) {
        $agency_id = $MVHub::Common::NO_AGENCY_ID;
    }

    return $agency_id;
}

sub insert_agency_record {
    my $values_href = shift;
    my $dbh         = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

    # Note: We first select the next available ID from agency id
    # sequence, then we insert the row with the ID. Although we
    # could do this as single operation, we wouldn't know what the
    # ID actually was (and we need it to pass back to the caller).
    my $agency_id = MVHub::Utils::DB::get_next_in_sequence( $dbh,
        'agency_id_sequence' );
    $values_href->{agency_id} = $agency_id;
    if ( !defined $values_href->{password} ) {
        $values_href->{password} = generate_random_password();
    }

    MVHub::Utils::DB::insert_one_record(
        dbh         => $dbh,
        table       => 'agency',
        values_href => $values_href
    );

    return $agency_id;
}

# Returns true iff the given agency name is not already in use
# in the database by another agency. If the given agency ID is real (i.e not
# $MVHub::Common::NO_AGENCY_ID), then any record with that ID will be excluded
# from the search. Returns false otherwise.
sub is_agency_name_unique {
    my %args        = @_;
    my $agency_name = delete $args{'agency_name'};
    my $agency_id   = delete $args{'agency_id'};
    MVHub::Utils::assert( defined $agency_name && $agency_name ne "",
        "Missing agency_name argument" );
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );
    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

    my @bind_variables = ($agency_name);
    my $sql            = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_X_IS_NAME_AVAILABLE');

    if ( $agency_id != $MVHub::Common::NO_AGENCY_ID ) {
        @bind_variables = ( $agency_name, $agency_id );
        $sql = MVHub::Utils::DB::get_sql_select_statement(
            'AGENCY_X_IS_NAME_AVAILABLE_NOT_ID');
    }

    my @row = $dbh->selectrow_array( $sql, undef, @bind_variables );

    return ( $row[0] == 0 );
}

sub prepare_record_for_db {
    my ( $values_href, $username, $is_data_entry_operator, $update_status )
        = @_;
    foreach my $key ( keys %$values_href ) {
        my $value = $values_href->{$key};
        if ( $key =~ /_verification$/ ) {
            delete $values_href->{$key};
        }
        elsif ( !defined $value ) {
            next;
        }

        # ONLY if we get here are we certain value is defined (so we
        # don't have to constantly check for it)
        elsif ( $value =~ /^\s*$/ ) {
            $values_href->{$key} = undef;
        }
        elsif ( $key eq 'email' or $key eq 'contact_email' ) {
            $values_href->{$key} = lc($value);
        }
        elsif ( grep( /$key/, @MVHub::Common::PHONE_FIELDS ) ) {
            $values_href->{$key} = MVHub::Common::normalize_phone($value);
        }
    }

# If the user had the option of checking date_updating_option but didn't, it'll exist in values_href
# but be undefined. In that case we don't want to update the last updated date. In all other cases,
# we do:
    if (  !exists $values_href->{'date_updating_option'}
        || defined $values_href->{'date_updating_option'} )
    {
        $values_href->{'last_updated'} = DateTime->now()->ymd();
    }
    delete $values_href->{'date_updating_option'};

}

sub update_agency_record {
    my ( $values_href, $agency_id ) = @_;

    # Reset number of reminders sent
    $values_href->{reminders_sent} = 0;
    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );
    MVHub::Utils::DB::update_one_record(
        dbh         => $dbh,
        table       => 'agency',
        values_href => $values_href,
        where_href  => { 'agency_id' => $agency_id }
    );
}

# This is the place for error checking that can't be done within
# CGI::FormBuilder->validate(), e.g.:
#   - complex validations that require access to other variables
#   - validations that need to produce a custom error message
sub validate_more {
    my $form       = shift;
    my $return_val = 1;

    if (!is_agency_name_unique(
            agency_name => $form->field('agency_name'),
            agency_id   => $form->field('agency_id'),
        )
        )
    {
        my $msg = "The agency name you entered is "
            . "already being used by another agency. Please enter a different one:";
        $form->tmpl_param( 'custom_error-agency_name', $msg );

        $return_val = 0;
    }

  # Since the e-mail field isn't required, the e-mail verification field never
  # gets validated by form-builder if it is left blank - thus we have to check
  # here
    if ((   defined $form->field('email')
            && !defined $form->field('email_verification')
        )
        || ((      defined $form->field('email')
                && defined $form->field('email_verification')
            )
            && ( $form->field('email_verification') ne $form->field('email') )
        )
        )
    {
        $form->tmpl_param( 'custom_error-email_verification',
            "The e-mail and e-mail verification fields must match." );
        $return_val = 0;
    }

    if (   $form->field('mail_address1')
        || $form->field('mail_address2')
        || $form->field('mail_city')
        || $form->field('mail_state')
        || $form->field('mail_zip') )
    {
        if (!(     $form->field('mail_address1')
                && $form->field('mail_city')
                && $form->field('mail_state')
                && $form->field('mail_zip')
            )
            )
        {
            $form->tmpl_param( 'custom_error-mail_address1',
                "If anything in the Mailing Address section is filled out, "
                    . "then it must be filled out entirely." );
            $return_val = 0;
        }
    }

    return $return_val;
}

sub send_admin_notification {
    my $values_href = shift;
    my $body;

    foreach my $field (@MVHub::AgencyFieldDefs::FIELDS) {
        if ( exists $values_href->{$field} )
        {    # Since we deleted the _verification fields
            $body 
                .= uc($field) . ": "
                . (
                ( defined $values_href->{$field} )
                ? $values_href->{$field}
                : ''
                ) . "\n\n";
        }
    }
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from(
        $ENV{MV_CONFIG_FILE} );
    my $to      = $cfg->param('NOTIFICATION.admin_email');
    my $website = $cfg->param('SITE.website_name');
    my $subject = "New: $website " . $values_href->{agency_name};

    MVHub::Common::notification_email(
        to      => $to,
        subject => $subject,
        message =>
            "The following agency was added to the Community Services Directory:\n\n$body"
    );
}

# Returns a FormBuilder object representing the appropriate agency form
# given the user and whether it's a new or existing form.
sub prepare_agency_form {
    my %args                   = @_;
    my $cgi                    = delete $args{cgi};
    my $agency_id              = delete $args{agency_id};
    my $username               = delete $args{username};
    my $is_admin               = delete $args{is_admin};
    my $is_agency              = delete $args{is_agency};
    my $is_data_entry_operator = delete $args{is_data_entry_operator};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $field_defs_href
        = MVHub::AgencyFieldDefs::get_agency_field_definitions_href();

    # Delete fields which don't apply to our circumstances (if any), i.e. ...

# ...only show the password fields when an agency person (not admin) is filling out
# a new form (not an existing one).
    if ( $is_admin || ( $agency_id != $MVHub::Common::NO_AGENCY_ID ) ) {
        delete @$field_defs_href{ 'password',
            'password_verification' };    # slice!
    }

# ...only admin gets a choice of whether or not to alter the last_updated field
# and only when changing an existing record.
    delete $field_defs_href->{'date_updating_option'}
        unless ( $is_admin && $agency_id != $MVHub::Common::NO_AGENCY_ID );

    # Same goes for the admin_notes field.
    delete $field_defs_href->{'admin_notes'}
        unless ( $is_admin && $agency_id != $MVHub::Common::NO_AGENCY_ID );

    my $form = MVHub::Common::prepare_form(
        form_attributes_href => {
            name     => 'agency_form',
            params   => $cgi,
            template => $AGENCY_FORM_TEMPLATE_FILE
        },
        fields_href => $field_defs_href
    );

    # For displaying instructions that only an admin should see:
    $form->tmpl_param( $IS_ADMIN_TAG, $is_admin );

    return $form;
}

# Generate an easy to remember password between 6 & 12 characters (inclusive)
sub generate_random_password {
    open( FILE, '/usr/share/dict/words' )
        or die "Could not open dictionary file for reading: $!";
    my @words = <FILE>;
    close FILE;

    my $password;
    do {
        my $word1 = $words[ int( rand( ( scalar @words ) - 1 ) ) ];
        chomp($word1);
        my $word2 = $words[ int( rand( ( scalar @words ) - 1 ) ) ];
        chomp($word2);
        my $num = int( rand(8) ) + 2;
        $password = "$word1$num$word2";
    } while ( length($password) < 6 || length($password) > 12 );
    return lc($password);
}

