#!/usr/bin/perl

use strict;
use warnings;

use MVHub::DieHandler;
use CGI;
use DateTime;
use MVHub::Utils;
use Set::Array;

use MVHub::AuthAccount;
use MVHub::Common;
use MVHub::Utils;
use MVHub::Utils::ConfigSimple;
use MVHub::Utils::DB;
use MVHub::HTML;
use MVHub::Page;
use MVHub::ProgramFieldDefs;

my $MODE_PARAM = 'mode';

# These are like run modes we get from forms/links/redirections
# that tell us what to do:
my $MODE_REMOVE_PROGRAM           = 'remove';
my $MODE_HIDE_PROGRAM             = 'hide';
my $MODE_UNHIDE_PROGRAM           = 'unhide';
my $MODE_SET_PROGRAM_CATEGORIES   = 'set_program_categories';
my $MODE_SHOW_NEW_CATEGORIES      = 'show_new_categories';
my $MODE_SHOW_EXISTING_CATEGORIES = 'show_existing_categories';
my $MODE_PROGRAM_FORM             = 'program_form';

# The form template, and some of the template variables it
# uses.
my $PROGRAM_FORM_TEMPLATE_FILE = 'program_form.tmpl';
my $TOP_ERROR_MESSAGE_TAG      = 'TOP_ERROR_MESSAGE_TAG';
my $IS_ADMIN_TAG               = 'IS_ADMIN_TAG';
my $CATEGORY_COLUMNS
    = 4;    # The number of columns when displaying category checkboxes

my $PF_AREAS_SERVED_LOOP_TAG         = 'areas_served';
my $PF_AREAS_SERVED_RADIO_BUTTON_TAG = 'radio_button';
my $PF_AREAS_SERVED_CHECKBOXES_TAG   = 'checkboxes';

# This is the prefix for the names of checkboxes in the areas served section.
my $AREAS_SERVED_CBOX_PREFIX = 'cboxarea_';

my $PROGRAM_NAME_TAG = 'program_name';
my $AGENCY_ID_TAG    = 'agency_id';
my $AGENCY_NAME_TAG  = 'agency_name';

my $PROGRAM_APPROVED_TAG = 'program_approved';

my @FIELDS_TO_COPY_FROM_AGENCY = qw [
    address1
    address2
    city
    state
    zip
    main_phone
    toll_free_phone
    hotline
    fax
    website
    email
    email_verification

    mail_address1
    mail_address2
    mail_city
    mail_state
    mail_zip

    resource_library
    resource_library_description
    meeting_space
    meeting_space_description

    contact_first_name
    contact_last_name
    contact_phone
    contact_fax
    contact_email
    contact_email_verification
    contact_email_available
];

my $JAVASCRIPT = <<JS;
    <script type="text/javascript" src="/all/js/areas_served.js"></script>
JS

{    # main
    my $cgi = CGI->new();
    MVHub::Utils::clean_cgi_params($cgi);

    my $agency_id = $cgi->param('agency_id');
    if ( !$agency_id ) {
        MVHub::Common::display_error_page_and_exit(
            "You are trying to access an invalid URL.");
    }

    my ( $is_admin, $is_agency, $error_message )
        = MVHub::AuthAccount::authenticate_session($cgi);

    my $program_id = $cgi->param('program_id');
    if ( !$program_id ) {
        $program_id = $MVHub::Common::NO_PROGRAM_ID;
        $cgi->param( 'program_id', $MVHub::Common::NO_PROGRAM_ID );
    }

    my $mode = $cgi->param($MODE_PARAM);

    if ( !defined $mode ) {

        #TMP:
        $mode = $MODE_PROGRAM_FORM;
        handle_main_program_form(
            cgi      => $cgi,
            is_admin => $is_admin,
        );
    }
    elsif ( $mode eq $MODE_REMOVE_PROGRAM ) {
        remove_program(
            program_name => $cgi->param('program_name'),
            program_id   => $program_id,
            agency_id    => $agency_id,
            is_admin     => $is_admin
        );
    }
    elsif ( $mode eq $MODE_SHOW_EXISTING_CATEGORIES ) {
        show_categories_selection(
            cgi      => $cgi,
            existing => 1
        );
    }
    elsif ( $mode eq $MODE_SHOW_NEW_CATEGORIES ) {
        show_categories_selection(
            cgi      => $cgi,
            existing => 0
        );
    }
    elsif ( $mode eq $MODE_SET_PROGRAM_CATEGORIES ) {
        set_program_categories( cgi => $cgi );
    }
    elsif ( $mode eq $MODE_HIDE_PROGRAM ) {
        hide_program(
            program_id => $program_id,
            agency_id  => $agency_id,
            is_admin   => $is_admin,
            reason     => $cgi->param('reason')
        );
    }
    elsif ( $mode eq $MODE_UNHIDE_PROGRAM ) {
        unhide_program(
            program_id => $program_id,
            agency_id  => $agency_id,
            is_admin   => $is_admin
        );
    }
    else {
        MVHub::Utils::assert( 0, "Unknown mode parameter, $mode" );
    }

    MVHub::Utils::assert( 0,
        "Sub in mode switch did not exit (mode = $mode)" );

}    # main

sub get_defaults_from_agency_form {
    my $agency_id   = shift;
    my $dbh         = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );
    my $agency_href = MVHub::Utils::DB::select_agency_record(
        dbh       => $dbh,
        agency_id => $agency_id
    );
    my %copied_fields;
    foreach (@FIELDS_TO_COPY_FROM_AGENCY) {
        MVHub::Utils::assert( exists $agency_href->{$_},
            "Can't find field name ($_) in agency record" );
        $copied_fields{$_} = $agency_href->{$_};
    }
    return \%copied_fields;
}

sub handle_main_program_form {
    my %args       = @_;
    my $cgi        = delete $args{'cgi'};
    my $is_admin   = delete $args{'is_admin'};
    my $agency_id  = $cgi->param('agency_id') or die "Missing agency id";
    my $program_id = $cgi->param('program_id') or die "Missing program id";

    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

    my $field_defs_href
        = MVHub::ProgramFieldDefs::get_program_field_definitions_href();

    # Delete fields that don't apply to our circumstances (if any), i.e. ...

# ...only admin gets a choice of whether or not to alter the last_updated field,
# and only on changing an existing record.
    delete $field_defs_href->{date_updating_option}
        unless ( $is_admin && $program_id != $MVHub::Common::NO_PROGRAM_ID );

    # Same goes for the admin_notes field.
    delete $field_defs_href->{'admin_notes'}
        unless ( $is_admin && $program_id != $MVHub::Common::NO_PROGRAM_ID );

    my $form = MVHub::Common::prepare_form(
        form_attributes_href => {
            name     => 'program_form',
            params   => $cgi,
            template => $PROGRAM_FORM_TEMPLATE_FILE
        },
        fields_href => $field_defs_href
    );

    # For displaying instructions that only an admin should see:
    $form->tmpl_param( $IS_ADMIN_TAG, $is_admin );

    $form->tmpl_param( $AGENCY_ID_TAG, $agency_id );

    if ( !$form->submitted() ) {
        my $values_href;
        my $agency_name;
        my $error_message;
        if ( $program_id == $MVHub::Common::NO_PROGRAM_ID ) {

            $values_href = get_defaults_from_agency_form($agency_id);
            $values_href->{area_type} = 'EVERYWHERE';
        }
        else {

            $values_href
                = MVHub::Utils::DB::select_program_record( $dbh, $agency_id,
                $program_id );
        }

        # Since the areas served section is too complicated to define using
        # typical FormBuilder methods, we hack it in here. values_href
        # is passed in to provide information about which radio & possible
        # checkbox(es) should be pre-checked.
        add_areas_served_to_form(
            form        => $form,
            program_id  => $program_id,
            values_href => $values_href
        );

        # The transit routes ("bus lines") section of the form is only
        # feasible with FormBuilder for one transit company.
        add_transit_routes_to_form(
            form        => $form,
            program_id  => $program_id,
            values_href => $values_href
        );
        MVHub::Page::print_page_and_exit(
            body       => $form->render( values => $values_href ),
            javascript => $JAVASCRIPT
        );
    }
    else {    # form is submitted
        if ( !$form->validate() or !validate_more($form) ) {    #User error(s)
            my $values_href = $form->field();
            $values_href->{bus_lines} = [ ( $cgi->param('bus_lines') ) ];
            my $announcement
                = '<p><b>We had problem(s) with your submission. '
                . 'Please see the message(s) highlighted below.</b></p>';

            $form->tmpl_param( $TOP_ERROR_MESSAGE_TAG, $announcement );

            add_areas_served_to_form(
                form       => $form,
                program_id => $program_id
            );
            add_transit_routes_to_form(
                form        => $form,
                program_id  => $program_id,
                values_href => $values_href
            );
            MVHub::Page::print_page_and_exit(
                body       => $form->render(),
                javascript => $JAVASCRIPT
            );
        }
        else {    # Successful submission
                  # submit it to the database
            my $values_href = $form->field();

            add_areas_served_to_record(
                form        => $form,
                values_href => $values_href
            );
            add_transit_routes_to_record(
                form        => $form,
                values_href => $values_href
            );

            prepare_record_for_db($values_href);

            # If the program id is NO_PROGRAM_ID, then we have new data.
            # INSERT a program record into the database
            if ( $program_id == $MVHub::Common::NO_PROGRAM_ID ) {

                $program_id = insert_program_record($values_href);

                if ( !$is_admin ) {    # an agency submitted a new record
                    send_admin_notification(
                        mode       => $MODE_PROGRAM_FORM,
                        program_id => $program_id,
                        agency_id  => $values_href->{agency_id},
                    );
                }

            # Redirect to the program category selection page.
            # Calling show_categories_selection() would be faster, but
            # it's cleaner to do a real redirection, so we'll do that for now.

                my $url = "program.pl?agency_id=$agency_id"
                    . "&program_id=$program_id&mode=$MODE_SHOW_NEW_CATEGORIES";

                print $cgi->redirect($url);
                exit(0);
            }

            # Else, since we do have a real program id, we must have gotten
            # updated data. UPDATE the program record in the database.
            else {
                update_program_record( $values_href, $agency_id,
                    $program_id );

                # Ordinarily, redirect to program category selection page:
                if ( !$is_admin ) {
                    my $url = "program.pl?agency_id=$agency_id"
                        . "&program_id=$program_id&mode=$MODE_SHOW_EXISTING_CATEGORIES";

                    print $cgi->redirect($url);
                    exit(0);
                }

                # But if we're admin, just redirect to the agency home page.
                MVHub::Common::display_agency_home_page_and_exit(
                    agency_id    => $agency_id,
                    init_session => 0,
                    message      => "Your program record has been updated."
                );
            }
        }
    }
}

sub insert_program_record {
    my ($values_href) = @_;
    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

    # Note: We first select the next available ID from program id
    # sequence, then we insert the row with the ID. Although we
    # could do this as single operation, we wouldn't know what the
    # ID actually was (and we need it to pass back to the caller).
    my $program_id = MVHub::Utils::DB::get_next_in_sequence( $dbh,
        'program_id_sequence' );
    $values_href->{program_id} = $program_id;

    MVHub::Utils::DB::insert_one_record(
        dbh         => $dbh,
        table       => 'program',
        values_href => $values_href
    );

    return $program_id;
}

# Returns true iff the given program name is not already in use
# by the agency with the given agency ID. If the given program ID is real
# (i.e not $MVHub::Common::NO_PROGRAM_ID), then any record with that ID will be
# excluded from the search. Returns false otherwise.
sub is_program_unique_for_agency {
    my %args         = @_;
    my $program_name = delete $args{'program_name'};
    my $agency_id    = delete $args{'agency_id'};
    my $program_id   = delete $args{'program_id'};
    MVHub::Utils::assert( defined $program_name && $program_name ne "",
        "Missing program_name argument" );
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );
    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

    my @bind_variables = ( lc($program_name), $agency_id );

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'PROGRAM_LOWER_PROGRAM_NAME_X_COUNT');

    # So they can update an existing program, and keep the same name
    if ( $program_id != $MVHub::Common::NO_PROGRAM_ID ) {
        $sql = MVHub::Utils::DB::get_sql_select_statement(
            'PROGRAM_LOWER_PROGRAM_NAME_ID_X_COUNT');
        @bind_variables = ( lc($program_name), $agency_id, $program_id );
    }

    my @row = $dbh->selectrow_array( $sql, { Slice => {} }, @bind_variables );

    return ( $row[0] == 0 );
}

sub prepare_record_for_db {
    my ( $values_href, $update_status ) = @_;

    foreach my $key ( keys %$values_href ) {
        my $value = $values_href->{$key};
        if ( $key =~ /_verification$/ ) {
            delete $values_href->{$key};
        }
        elsif ( !defined $value ) {
            next;
        }

        # ONLY if we get here are we certain value is defined (so we
        # don't have to constantly check for it)
        elsif ( ref($value) eq 'ARRAY' ) {
            $values_href->{$key} = MVHub::Common::delimit_field($value);

        }
        elsif ( $value =~ /^\s*$/ ) {
            $values_href->{$key} = undef;
        }
        elsif ( $key eq 'email' or $key eq 'contact_email' ) {
            $values_href->{$key} = lc($value);
        }
        elsif ( grep( /$key/, @MVHub::Common::PHONE_FIELDS ) ) {
            $values_href->{$key} = MVHub::Common::normalize_phone($value);
        }
    }

# If the user had the option of checking date_updating_option but didn't, it'll exist in values_href
# but be undefined. In that case we don't want to update the last updated date. In all other cases,
# we do:
    if (  !exists $values_href->{'date_updating_option'}
        || defined $values_href->{'date_updating_option'} )
    {
        $values_href->{'last_updated'} = DateTime->now()->ymd();
    }
    delete $values_href->{'date_updating_option'};

}

sub show_categories_selection {
    my %args     = @_;
    my $cgi      = delete $args{'cgi'};
    my $existing = delete $args{'existing'};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $agency_id  = $cgi->param('agency_id')  or die "missing agency_id";
    my $program_id = $cgi->param('program_id') or die "missing program_id";
    my $error = $cgi->param('error');    # optional

    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'CATEGORY_HEADING_CATEGORY_X_HEADING_CATEGORY_NAME');

    my $categories_aref = $dbh->selectall_arrayref( $sql, { Slice => {} } );

    # Fill a hash of: heading => { category_id => category_name, ... }
    my %categories_hash;
    foreach my $category_href (@$categories_aref) {
        $categories_hash{ $category_href->{heading_name} }
            { $category_href->{category_id} }
            = $category_href->{category_name};
    }

    # Get any category ID previously selected for this program, if any
    my $preselected_category_ids_aref;
    if ($existing) {
        my @bind_variables = ($program_id);
        $sql = MVHub::Utils::DB::get_sql_select_statement(
            'PROGRAM_CATEGORY_X_CATEGORY_ID');
        $preselected_category_ids_aref
            = $dbh->selectcol_arrayref( $sql, { Slice => {} },
            @bind_variables );
    }

    # Produce our page
    $cgi = new CGI("");

    my $code = $cgi->h2('Categorize Your Program');

    if ($error) {
        $code
            = '<p style="color: #ff0000">You must select at least one category.</p>';
    }

    $code
        .= $cgi->p(
        "Directory users search and browse for your program listing by way of the categories that you check off below. Use the following links to navigate to the different groups of categories. Please take the time to <strong>check off all that apply to your program</strong> and press the <strong>Submit</strong> button (at bottom) when done."
        );
    if ($existing) {
        $code
            .= $cgi->p(
            "Even if you are just returning to this page, please take the time to update your selections, as we periodically add new categories requested by MVHub users."
            )
            . $cgi->p(
            "<a href='/cgi-bin/mvhub/agency.pl?rm=display_agency_home&amp;agency_id=$agency_id'>Click here to return to the Agency Home Page</a>"
            );
    }
    $code .= $cgi->start_form(
        -name   => 'category_form',
        -method => 'POST',
        -action => 'program.pl'
        )
        . $cgi->hidden( 'program_id', $program_id )
        . $cgi->hidden( 'agency_id',  $agency_id )
        . $cgi->hidden( $MODE_PARAM,  $MODE_SET_PROGRAM_CATEGORIES )
        . $cgi->a( { name => "CATEGORY_LIST" } );

    # Generate an html table of navigational links for all the headings.

    # Start by extracting just the heading names in alphabetical order.
    my @nav_links = map { $cgi->a( { href => "#$_" }, "$_" ) }
        ( sort keys %categories_hash );

    $code .= MVHub::HTML::make_html_table(
        columns          => 3,
        content_aref     => \@nav_links,
        table_attributes => 'cellspacing="5"'
    );

    $code .= $cgi->hr;

    # Generate the headings and checkbox groups

    # See the stylesheet (main.css) for explanation on how the checkbox labels
    # are formatted.

    # Prevent autoescaping so we can use html in our checkbox labels
    $cgi->autoEscape(0);

    my $heading_index = 0;

    foreach my $heading ( sort keys %categories_hash ) {

        $heading_index++;

        $code .= $cgi->a( { name => $heading } );
        $code .= $cgi->h3($heading);
        $code .= $cgi->a( { href => "#CATEGORY_LIST" }, "Back to top" );

        # sort the keys (ids) by their values (category names)
        my $checkbox_values_aref = [
            sort {
                $categories_hash{$heading}{$a}
                    cmp $categories_hash{$heading}{$b}
                } ( keys %{ $categories_hash{$heading} } )
        ];

        my $labels_href;
        my $attr_href;
        foreach my $checkbox_value ( keys %{ $categories_hash{$heading} } ) {
            my $label = $categories_hash{$heading}->{$checkbox_value};

            # add space around forward slashes so a label like
            # "Translating/Interpreting" will wrap nicely
            $label =~ s|\s*/\s*| / |g;
            $label = CGI::escapeHTML($label);

   # The unique IDs are used by javascript in highlighting category labels and
   # checking off categories checkboxes for categories that appear in multiple
   # headings.

   # Sets a unique ID for every CheckboxLabel DIV, of the form, so they can be
   # referenced directly
            $labels_href->{$checkbox_value}
                = '<DIV id="lbl'
                . $checkbox_value . '_'
                . $heading_index
                . '" class="CheckboxLabel'
                . (
                grep( /^$checkbox_value$/, @$preselected_category_ids_aref )
                ? 'Selected'
                : ''
                )
                . '">'
                . $label
                . '</DIV>';

     # Sets a unique ID for every checkbox, so they can be referenced directly
            $attr_href->{$checkbox_value}
                = { id => 'chk' . $checkbox_value . '_' . $heading_index };
        }

        # Generate a checkbox group of categories sorted by category name
        $code .= '<DIV class="CheckTable">';
        $code .= $cgi->checkbox_group(
            -name    => 'category_id',
            -values  => $checkbox_values_aref,
            -labels  => $labels_href,
            -columns => $CATEGORY_COLUMNS,
            -onClick => 'highlightLabel(this)',

            # Check off any pre-selected values. Yeah,
            # $preselected_category_ids_aref is a list of all of the
            # program's chosen category values, not just those for
            # this heading, but cgi just ignores the excessive ones.
            -default    => $preselected_category_ids_aref,
            -attributes => $attr_href
        );
        $code .= '</DIV>';
    }

    # We're done, turn it back on
    $cgi->autoEscape(1);

    $code
        .= '<div style="width: 100%; text-align: center"><span class="nicebutton">'
        . $cgi->submit(
        -name  => 'submit',
        -value => 'Submit',
        -style => 'float: none'
        ) . '</span></div>';

    $code .= $cgi->endform();

    # TODO possibly: employ a template for the body
    # Set up template
    #    my $template = HTML::Template->new(
    #        filename => 'category_selection.tmpl' );
    #    $template->param( 'PROGRAM_ID', $program_id );
    #    $template->param( 'AGENCY_ID', $agency_id );
    #    $template->param( 'CHECKBOX_CODE', $code );
    #    my $body = $template->output();

    # Put body within the main template

    $code = '<DIV class="CheckboxGroup">' . $code . '</DIV>';

    MVHub::Page::print_page_and_exit(
        body       => $code,
        javascript => <<JS
<script type="text/javascript">
  function highlightLabel(checkbox) {
      /* This highlights all labels for the the same category, and      
       * checks the other checkboxes for categories that are duplicated 
       * across multiple headings. It is fast!
       */
      for ( var i = 1; i <= $heading_index; i++ ) {
          var this_checkbox = document.getElementById( 'chk' + checkbox.value + '_' + i );
          if ( this_checkbox != null ) {
              this_checkbox.checked = checkbox.checked;
          }
          var checkbox_label = document.getElementById( 'lbl' + checkbox.value + '_' + i );
          if ( checkbox_label != null ) {
              if ( checkbox.checked ) {
                  checkbox_label.className = "CheckboxLabelSelected";
              } else {
                  checkbox_label.className = "CheckboxLabel";
              }
          }
      }
  }
</script>
JS
    );
}

sub update_program_record {
    my ( $values_href, $agency_id, $program_id ) = @_;

    # Set # e-mail reminders to 0
    $values_href->{reminders_sent} = 0;
    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );
    MVHub::Utils::DB::update_one_record(
        dbh         => $dbh,
        table       => 'program',
        values_href => $values_href,
        where_href =>
            { 'agency_id' => $agency_id, 'program_id' => $program_id }
    );
}

sub validate_more {
    my $form       = shift;
    my $return_val = 1;

    # Program name must be unique for the given agency
    if (!is_program_unique_for_agency(
            program_name => $form->field('program_name'),
            agency_id    => $form->field('agency_id'),
            program_id   => $form->field('program_id')
        )
        )
    {
        my $msg
            = 'Your agency already has a program with the following name. '
            . 'Please either edit that record instead, or enter a different name: ';
        $form->tmpl_param( 'custom_error-program_name', $msg );

        $return_val = 0;
    }

# Since the e-mail field isn't required, the e-mail verfication field never
# gets validated by form-builder if it is left blank - thus we have to check
# here. We do the fancy 'or ''" to avoid warnings about doing equality operations
# on undefined values.
    if ( ( $form->field('email_verification') or '' ) ne
        ( $form->field('email') or '' ) )
    {
        $form->tmpl_param( 'custom_error-email_verification',
            "The e-mail and e-mail verification fields must match." );
        $return_val = 0;
    }

    if (   $form->field('mail_address1')
        || $form->field('mail_address2')
        || $form->field('mail_city')
        || $form->field('mail_state')
        || $form->field('mail_zip') )
    {
        if (!(     $form->field('mail_address1')
                && $form->field('mail_city')
                && $form->field('mail_state')
                && $form->field('mail_zip')
            )
            )
        {
            $form->tmpl_param( 'custom_error-mail_address1',
                "If anything in the Mailing Address section is filled out, "
                    . "then it must be filled out entirely." );
            $return_val = 0;
        }
    }

    my $area_type = $form->cgi_param('area_type');
    my @area_data
        = $form->cgi_param( $AREAS_SERVED_CBOX_PREFIX . $area_type );

    if ( ( grep ( ( $area_type eq $_ ), qw(COUNTY REGION TOWN) ) )
        and !@area_data )
    {
        $form->tmpl_param( 'custom_error-areas_served',
                  "Since you chose $area_type, below, you must check "
                . "off at least one $area_type." );
        $return_val = 0;

    }
    return $return_val;
}

sub set_program_categories {
    my %args = @_;
    my $cgi  = delete $args{'cgi'};

    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );
    my $agency_id  = $cgi->param('agency_id')  or die "missing agency ID";
    my $program_id = $cgi->param('program_id') or die "missing program ID";
    my @category_ids
        = Set::Array->new( $cgi->param('category_id') )->unique();

    if ( @category_ids < 1 ) {

        # If they didn't select any categories, redirect them back to the
        # category selection page with an error message.
        my $url = "program.pl?agency_id=$agency_id"
            . "&program_id=$program_id&mode=$MODE_SHOW_NEW_CATEGORIES&error=1";
        print $cgi->redirect($url);
        exit(0);
    }

    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );
    $dbh->{AutoCommit} = 0;

    db_delete_program_categories( dbh => $dbh, program_id => $program_id );

    db_insert_program_categories(
        dbh               => $dbh,
        program_id        => $program_id,
        category_ids_aref => \@category_ids
    );

    $dbh->commit();
    $dbh->{AutoCommit} = 1;

    MVHub::Common::display_agency_home_page_and_exit(
        agency_id    => $agency_id,
        init_session => 0,
        message      => 'Your changes have been saved.'
    );

}

sub db_insert_program_categories {
    my %args              = @_;
    my $dbh               = delete $args{'dbh'};
    my $program_id        = delete $args{'program_id'};
    my $category_ids_aref = delete $args{'category_ids_aref'};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $sql = MVHub::Utils::DB::get_sql_insert_statement(
        'PROGRAM_CATEGORY_X_REQUIRED_FIELDS');

    my $sth = $dbh->prepare($sql);

    foreach (@$category_ids_aref) {
        $sth->execute( $program_id, $_ );
    }
}

sub db_delete_program_categories {
    my %args       = @_;
    my $dbh        = delete $args{'dbh'};
    my $program_id = delete $args{'program_id'};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my @bind_variables = ($program_id);
    my $sql            = MVHub::Utils::DB::get_sql_delete_statement(
        'PROGRAM_CATEGORY_X_PROGRAM_ID');

    $dbh->do( $sql, undef, @bind_variables );
}

sub hide_program {
    my %args       = @_;
    my $program_id = delete $args{'program_id'};
    my $agency_id  = delete $args{'agency_id'};
    my $is_admin   = delete $args{'is_admin'};
    my $reason     = delete $args{'reason'};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    MVHub::Utils::assert( $reason, "Expecting a reason" );

    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );
    $reason
        = "Hidden on "
        . DateTime->now()->mdy() . " by: "
        . (
        $is_admin ? 'the MVHub Administrator' : 'the agency account holder' )
        . "<BR>Reason: <i>$reason</i>";

    my @bind_variables = ( $reason, $program_id );
    my $sql = MVHub::Utils::DB::get_sql_update_statement(
        'PROGRAM_X_HIDDEN_REASON');
    $dbh->do( $sql, undef, @bind_variables );

    if ( !$is_admin ) {
        send_admin_notification(
            mode       => $MODE_HIDE_PROGRAM,
            program_id => $program_id,
            agency_id  => $agency_id
        );
    }

    MVHub::Common::display_agency_home_page_and_exit(
        agency_id    => $agency_id,
        init_session => 0,
    );
}

sub unhide_program {
    my %args       = @_;
    my $program_id = delete $args{'program_id'};
    my $agency_id  = delete $args{'agency_id'};
    my $is_admin   = delete $args{'is_admin'};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

    my @bind_variables = ($program_id);
    my $sql            = MVHub::Utils::DB::get_sql_update_statement(
        'PROGRAM_X_HIDDEN_REASON_IS_NULL');

# Send the e-mail before executing the db query, so we can get the hidden_reason
    if ( !$is_admin ) {
        send_admin_notification(
            mode       => $MODE_UNHIDE_PROGRAM,
            program_id => $program_id,
            agency_id  => $agency_id
        );
    }

    $dbh->do( $sql, undef, @bind_variables );

    MVHub::Common::display_agency_home_page_and_exit(
        agency_id    => $agency_id,
        init_session => 0,
    );
}

# Returns the number programs deleted (0 or 1). Will die if more than 1 program
# is deleted.
sub remove_program {
    my %args         = @_;
    my $program_name = delete $args{'program_name'};
    my $program_id   = delete $args{'program_id'};
    my $agency_id    = delete $args{'agency_id'};
    my $is_admin     = delete $args{'is_admin'};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

# Make sure there is a program to be deleted (because someone might accidentally click
# the "Remove" link twice in a row)

    my $sql
        = MVHub::Utils::DB::get_sql_select_statement('PROGRAM_X_COUNT_ALL');
    my @bind_variables = ($program_id);

    if (scalar $dbh->selectrow_array( $sql, { Slice => {} }, @bind_variables )
        )
    {

        if ( !$is_admin ) {

 # We send the notification before doing the deletion since the following call
 # queries the database for that record.
            send_admin_notification(
                mode       => $MODE_REMOVE_PROGRAM,
                program_id => $program_id,
                agency_id  => $agency_id,
            );
        }

        my @bind_variables = ($program_id);
        my $sql            = MVHub::Utils::DB::get_sql_delete_statement(
            'PROGRAM_X_PROGRAM_ID');

        my $rows_affected = $dbh->do( $sql, undef, @bind_variables );

        MVHub::Utils::assert( $rows_affected == 1,
            "Expected 1 row removed; got $rows_affected" );
    }

    my $msg = "The \"$program_name\" program has been removed.";
    MVHub::Common::display_agency_home_page_and_exit(
        agency_id    => $agency_id,
        init_session => 0,
        message      => $msg
    );
}

sub send_admin_notification {
    my %args       = @_;
    my $mode       = delete $args{'mode'};
    my $program_id = delete $args{'program_id'};
    my $agency_id  = delete $args{'agency_id'};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_PROGRAM_X_AGENCY_NAME');
    my @bind_variables = ( $program_id, $agency_id );

    # Regardless of mode, get the program info and prepare it for display
    my $result_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, @bind_variables );

    MVHub::Utils::assert( ( @$result_aref == 1 ),
        "Expected 1 result, got " . scalar @$result_aref );

    my $values_href = $$result_aref[0];

    my $program_name = $values_href->{program_name};
    my $agency_name  = $values_href->{agency_name};

    my $subject = "Program [ $program_name ]";

    my $body = '';

    # Mode-specific actions
    if ( $mode eq $MODE_REMOVE_PROGRAM ) {
        $subject .= " removed";
        $body = "$subject by the agency\n\n";
    }
    elsif ( $mode eq $MODE_HIDE_PROGRAM ) {
        $subject .= " hidden";
        $body = "$subject by the agency\n\n";
        $body
            .= "Reason for hiding program: $values_href->{hidden_reason}\n\n";
    }
    elsif ( $mode eq $MODE_UNHIDE_PROGRAM ) {
        $subject .= " unhidden";
        $body = "$subject by the agency\n\n";
        $body
            .= "Original reason program was hidden: $values_href->{hidden_reason}\n\n";
    }
    elsif ( $mode eq $MODE_PROGRAM_FORM ) {
        $subject .= " added";
        $body = "$subject by ";
        $body .= "the agency \n\n";
    }
    else {
        MVHub::Utils::assert( 0,
            "Unknown mode ($mode) passed to send_admin_notification." );
    }

    $body .= "\n-------------------------------------\n\n";

    $body .= "Agency Name: $agency_name\n\n";

    # Prepare the fields for display
    foreach my $field (@MVHub::ProgramFieldDefs::FIELDS) {

        if ( defined $values_href->{$field} ) {
            $values_href->{$field} =~ s/$MVHub::Common::FIELD_DELIMITER/, /g;
        }

        if ( exists $values_href->{$field} )
        {    # Since we deleted the _verification fields
            $body 
                .= uc($field) . ": "
                . (
                ( defined $values_href->{$field} )
                ? $values_href->{$field}
                : ''
                ) . "\n\n";
        }

    }
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from(
        $ENV{MV_CONFIG_FILE} );
    my $to = $cfg->param('NOTIFICATION.admin_email');

    # Send the e-mail
    MVHub::Common::notification_email(
        to      => $to,
        subject => $subject,
        message => $body
    );
}

# Take the area_type and $AREAS_SERVED_CBOX_PREFIX<area_type-value> fields
# and put them into the record (values_href) that is being prepared for entry
# in the database
# The STATE area_type is a bit odd. It is designed to handle any
# number of states, but right now Massachusetts is the only one,
# so it is hardcoded in (as there is only a radio button)
sub add_areas_served_to_record {
    my %args        = @_;
    my $form        = delete $args{form};
    my $values_href = delete $args{values_href};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $area_type = $form->cgi_param('area_type');
    my $area_data_aref
        = [ $form->cgi_param( $AREAS_SERVED_CBOX_PREFIX . $area_type ) ];
    if ( $area_type eq 'STATE' ) { $area_data_aref = ['MA']; }

    $values_href->{area_type} = $area_type;
    $values_href->{area_data} = $area_data_aref;

}

# Add a radio button group to the form for the different ways (i.e.
# area_type)  the user can choose to define the areas served by the
# program. Some of those types may provide a checkbox group of choices
# (i.e. area_data). E.g. area_type of COUNTY provides a list of counties
# to check off.
# In the form, each set of checkboxes has its own name (i.e.
# $AREAS_SERVED_CBOX_PREFIX + [the area_type].
# The given $values_href can contain default value(s) for area_type
# and/or area data.
# If this sub is called as the result of a previous form submission that
# failed validation, this method will pull the sticky values out of the
# CGI environment.
sub add_areas_served_to_form {
    my %args        = @_;
    my $form        = delete $args{form};
    my $program_id  = delete $args{program_id};
    my $values_href = delete $args{values_href};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $AREA_TYPES_AREF = [ 'EVERYWHERE', 'STATE', 'COUNTY', 'TOWN' ];
    my $COUNTIES_AREF = [ 'Middlesex County', 'Essex County' ];

    my $cgi = new CGI;

    my $default_area_type;
    if ( defined $values_href ) {
        $default_area_type = $values_href->{area_type};
    }

    my @radio_code = $cgi->radio_group(
        -name => 'area_type',

        # TODO: move below definition to someplace more general
        -values    => $AREA_TYPES_AREF,
        -default   => $default_area_type,
        -linebreak => 'true',
        -labels    => {
            EVERYWHERE => 'Everywhere',
            STATE      => 'Massachusetts only',
            COUNTY     => 'By County:',
            TOWN       => 'By Town:'
        },
        -onClick => 'updateAreasServed(this)'
    );

    my $towns_aref = get_selectable_towns_from_db();

    my %default_checked;

    if (    defined $values_href->{area_type}
        and defined $values_href->{area_data} )
    {
        $default_checked{ $values_href->{area_type} }
            = $values_href->{area_data};
    }

    # $checkbox_code[N] and $radio_code[N] must refer to the same area_type
    # area_types that do not contain data should have a entry of 0 in
    # @checkbox_code
    my @checkbox_code = (
        0, 0,
        scalar $cgi->checkbox_group(    # By County
            -name    => $AREAS_SERVED_CBOX_PREFIX . 'COUNTY',
            -values  => $COUNTIES_AREF,
            -default => $default_checked{'COUNTY'},
            -onClick => 'updateAreasServed(this)'
        ),
        scalar $cgi->checkbox_group(    # By Town
            -name    => $AREAS_SERVED_CBOX_PREFIX . 'TOWN',
            -values  => $towns_aref,
            -default => $default_checked{'TOWN'},
            -onClick => 'updateAreasServed(this)',
            -columns => 3
        )
    );

    my @areas_served_loop_params;

    for ( my $i = 0; $i < @radio_code; $i++ ) {
        push @areas_served_loop_params,
            {
            $PF_AREAS_SERVED_RADIO_BUTTON_TAG => $radio_code[$i],
            $PF_AREAS_SERVED_CHECKBOXES_TAG   => $checkbox_code[$i]
            };
    }

    $form->tmpl_param(
        $PF_AREAS_SERVED_LOOP_TAG => \@areas_served_loop_params );

}

# Only a certain set of towns (of those in the db) is available to the user when they want to
# choose the areas served by town. These towns have a value of TRUE in the display column of
# the town table.
sub get_selectable_towns_from_db {
    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );
    my $sql = MVHub::Utils::DB::get_sql_select_statement('TOWN_X_TOWN_STATE');
    my $towns_aref = $dbh->selectcol_arrayref($sql);
    MVHub::Utils::assert( @$towns_aref > 0, "Expected at least one town." );
    return $towns_aref;
}

# Adds checkbox list of transit routes (bus, train, etc.) for each company listed below.
# The companies and routes really need to go into the database, but most of my (John's) time right
# now needs to be devoted to the data entry interface, so this is a quick 'n' dirty solution.
# CGI::FormBuilder doesn't have a good method (or I just don't know it well enough :-)
# to separate transit routes by company, hence this subroutine.
sub add_transit_routes_to_form {
    my %args        = @_;
    my $form        = delete $args{form};
    my $program_id  = delete $args{program_id};
    my $values_href = delete $args{values_href};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    # Eventually move this stuff to db, as it's already pretty unwieldy.
    # If any other companies are included, then it's essential that we use
    # a database to store transit info.
    # The hash ref is in the form desired by HTML::Template for nested loops.
    my $TRANSIT_COMPANIES_AREF = [
        {   company_name   => 'MVRTA',
            company_url    => 'http://www.mvrta.com/routes.shtml',
            company_routes => [
                { route_name => '01 Haverhill - Methuen - Lawrence' },
                { route_name => '13 Main St. / North Ave.' },
                { route_name => '14 Bradford / Ward Hill' },
                { route_name => '15 Hilldale Ave. / Westgate' },
                { route_name => '16 Washington St. / Westgate' },
                { route_name => '18 Riverside' },
                { route_name => '19 Summer St. / Westgate' },
                { route_name => '21 Andover Shuttle' },
                { route_name => '32 Andover' },
                { route_name => '33 North Andover' },
                { route_name => '34 Prospect Hill' },
                { route_name => '35 Water Street' },
                { route_name => '36 Lawrence Street / Holy Family Hospital' },
                { route_name => '37 Beacon St.' },
                { route_name => '38 Hampshire St.' },
                { route_name => '39A Colonial Heights / N. Andover Mall' },
                { route_name => '39B North Andover Mall / Phillips' },
                { route_name => '40 Methuen Square' },
                { route_name => '41 Lawrence-Lowell' },
                { route_name => '42 Ring & Ride Methuen' },
                { route_name => '51 Haverhill - Amesbury - Newburyport' },
                { route_name => '72 Haverhill' },
                { route_name => '73 Lawrence' },
                { route_name => '83 Salisbury Beach / Hampton Beach' }
            ],
        },

        {   company_name   => 'LRTA',
            company_url    => 'http://www.lrta.com/fixedroute.html',
            company_routes => [
                { route_name => '01 Christian Hill' },
                { route_name => '02 Belvidere' },
                { route_name => '03 South Lowell' },
                { route_name => '04 Shaw - Stevens' },
                { route_name => '05 Westford Street/Drum Hill' },
                { route_name => '06 Broadway / UMass Lowell' },
                { route_name => '07 Pawtucketville / UMass Lowell North' },
                { route_name => '08 Centralville' },
                { route_name => '09 Downtown Circulator' },
                { route_name => '10 Dracut / Tyngsboro' },
                { route_name => '11 IRS/Raytheon via Rte. 133' },
                { route_name => '12 Tewksbury via Rte. 38' },
                { route_name => '13 Billerica via Edson' },
                { route_name => '14 Burlington Mall / Lahey Clinic' },
                { route_name => '15 Chelmsford via Rte. 129' },
                { route_name => '16 Chelmsford via Chelmsford Street' },
                { route_name => '17 North Chelmsford via Middlesex Street' },
                { route_name => '18 Express Shuttle' }
            ],
        },
    ];

# Pre-populate transit route checkboxes
# Adds a template variable, route_is_selected, for each transit route
# used by the program.  UGLY UGLY since we're traversing through two
# separate nested data structures and have to break things up by company.
# Will be far cleaner when transit routes are put into the database.
# TODO: fix resubmission problem (bus lines don't get re-populated after submission errors)
    if ( defined $values_href->{bus_lines} ) {
        foreach my $transit_company_href (@$TRANSIT_COMPANIES_AREF) {
            foreach my $transit_route_href (
                @{ $transit_company_href->{company_routes} } )
            {

           # If there's one bus line, it's stored in $values_href as a string.
           # If more than one bus line, it's stored as an array reference.
           # This returns an array for both cases.
                my @selected_transit_routes
                    = ref $values_href->{bus_lines}
                    ? @{ $values_href->{bus_lines} }
                    : ( $values_href->{bus_lines} );
                foreach my $selected_transit_route (@selected_transit_routes)
                {
                    if ( $transit_route_href->{route_name} eq
                        $selected_transit_route )
                    {

                        # Creating a new template var on the fly.
                        $transit_route_href->{route_is_selected} = 1;
                    }
                }
            }
        }
    }

    $form->tmpl_param( transit_companies => $TRANSIT_COMPANIES_AREF );
}

# Takes bus/transit routes from CGI submission and puts them into the values_href hash
# so they can be properly inserted into the database.
sub add_transit_routes_to_record {
    my %args        = @_;
    my $form        = delete $args{form};
    my $values_href = delete $args{values_href};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $transit_routes_aref = [ $form->cgi_param('bus_lines') ];
    $values_href->{bus_lines} = $transit_routes_aref;
}

sub send_confirmation_email {
    my $values_href = shift;
    my $agency_name
        = MVHub::Utils::DB::get_full_name( $values_href->{agency_id} );

    # So these can be tweaked as necessary.  For example, we may want to
    # change the $from address according to who submits the form.
    my $from;
    my $to;
    my $subject;
    my $body;
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from(
        $ENV{MV_CONFIG_FILE} );
    my $team_name   = $cfg->param('NOTIFICATION.team_name');
    my $admin_email = $cfg->param('NOTIFICATION.admin_email');

    $from    = "$team_name <$admin_email>";
    $to      = $values_href->{contact_email};
    $subject = $agency_name . " has been added!";

    $body = << "BODY";
Hi $values_href->{contact_first_name}

$values_href->{agency_name} has just been added to our database.

You were listed as our contact for this agency. 

If you didn't enter this information and feel that 
you should not be the contact for this agency, 
please contact us. For future reference, your 
agency's password is: 

    $values_href->{password}

Thank you for registering with the Hub
$team_name

BODY

    my $top = MIME::Entity->build(
        Type    => "text/plain",
        From    => $from,
        To      => $to,
        Subject => $subject,
        Data    => $body
    );
    MVHub::Common::sendmail($top);
}
