#!/usr/bin/perl

# PURPOSE: create printable .pdf directory of all agnecies and programs

use strict;
use warnings;
use Carp;

use FindBin;
use HTML::Template;
use MVHub::GeneratePDF;
use MVHub::Utils::ConfigSimple;
use MVHub::Utils::DB;

# Needed to remove dependency with MVhub::Common.
# The road to hell is paved with lazy globals.
our $DISGUSTING_FIELD_DELIMITER = '!DELIM!';

{    # main
    my $config_file = $ARGV[0] || $ENV{MV_CONFIG_FILE};
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from($config_file);

    chdir $cfg->param('ABSOLUTE_PATH.tmp_dir');

    my $website = $cfg->param('SITE.website_name');

    my $stem = "${website}_agencies_and_all_programs";

    my $latex_file = "$stem.tex";
    my $index_file = "$stem.idx";
    my $pdf_file   = "$stem.pdf";

    open( LATEX, ">$latex_file" )
        || die "couldn't open temp file $latex_file for writing";

    print LATEX ${ latex_report_sref( $cfg, $config_file ) };

    my $latex_cmd
        = "/usr/bin/pdflatex --interaction batchmode $latex_file > /dev/null";
    my $index_cmd = "/usr/bin/makeindex $index_file 2> /dev/null > /dev/null";

    # if you do multiple indexes, or something else tricky
    # the number of passes required would change.
    foreach my $pass ( 1 .. 4 ) {
        ( system($latex_cmd) == 0 )
            or
            die( "'$latex_cmd' failed on pass '$pass' " . describe_error() );

        if ( $pass == 3 ) {
            ( system($index_cmd) == 0 )
                or die( "'$index_cmd' failed: " . describe_error() );
        }
    }

    my $reports_dir = $cfg->param('ABSOLUTE_PATH.reports_dir');
    ( system("cp $pdf_file $reports_dir") == 0 )
        or die "failed to copy $pdf_file to $reports_dir";

}    # main

# creates a more descriptive error message.
sub describe_error {
    'exit value: ' . ( $? >> 8 ) . ' signal: ' . ( $? & 127 );
}

# returns a reference to a scalar
# containing the latex report
sub latex_report_sref {
    my $cfg         = shift or croak 'missing Param: $cfg';
    my $config_file = shift or croak 'missing param: $config_file';

    my $dbh = MVHub::Utils::DB::get_dbh($config_file);

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'HEADING_X_HEADING_NAME');

    my $headings_aref = $dbh->selectcol_arrayref($sql);

    # keys are entry ids
    # values are headings
    # starts blank, and as entries are output, keys are added.
    my %already_seen_under;

    my $website             = $cfg->param('SITE.website_name');
    my $website_description = $cfg->param('SITE.website_description');
    my $website_location    = $cfg->param('SITE.location');

    my $latex = doc_head( $website, $website_description );
    $latex .= agency_list_section($dbh);
    $latex
        .= about_mvhub( $website, $website_description, $website_location );
    $latex .= see_alsos($dbh);
    $latex .= "\\mainmatter\n";

    for my $heading (@$headings_aref) {

        $heading = MVHub::GeneratePDF::clean($heading);

        $latex .= section_heading($heading);

        # We need four separate queries:
        #	multiple programs programs: (mult_prog_prog)
        #		programs under agencies with multiple programs
        #		display only program fields
        #	multiple programs agencies: (mult_prog_agen)
        #		agencies with multiple programs
        #		display agency fields
        #		categories are the union of program categories
        #	single program agencies: (sing_prog_agen)
        #		agencies with a single program
        #		fold together agency and program fields

        #	zero program agencies: (zero_prog_agen)
        #		agencies with no programs
        #               OR agencies with programs that aren't categorized
        #		(This might be a problem with the data model,
        #		 as agencies do not have headings. For the moment,
        #		 place under 'Other')

        my $mult_prog_prog_aref
            = get_programs_with_siblings( $dbh, $heading );

        my $mult_prog_agen_aref
            = get_agencies_with_multiple_programs( $dbh, $heading );

        my $sing_prog_agen_aref
            = get_single_program_agencies( $dbh, $heading );
        my $zero_prog_agen_aref
            = $heading eq 'Other' ? get_agencies_without_programs($dbh) : [];

        my $entries_aref = MVHub::GeneratePDF::clean_merge_sort(
            $mult_prog_prog_aref, $mult_prog_agen_aref,
            $sing_prog_agen_aref, $zero_prog_agen_aref
        );

        for my $prev_heading (@$headings_aref) {
            my @seen_entries = grep {
                ( $already_seen_under{ $_->{'id'} } || '' ) eq $prev_heading;
            } @{$entries_aref};
            if (@seen_entries) {
                $latex .= "Listed under $prev_heading:\n";
                $latex .= list_entries(@seen_entries);
            }
        }

        for my $r ( @{$entries_aref} ) {
            unless ( $already_seen_under{ $r->{'id'} } ) {
                my @categories = categories_of( $r, $dbh );
                $latex .= indexing( $r, @categories );
                $latex .= summary( $r, $dbh );
                $already_seen_under{ $r->{'id'} } = $heading;
            }
        }
    }

    $latex .= doc_foot();

    return \$latex;
}

sub doc_head {
    my $website = shift or croak 'missing Parameter: $website';
    my $website_description = shift
        or croak 'missing Parameter: $website_description';

    my $text = <<'END';
\documentclass{book}

\usepackage{times}
% a narrower font than the default.

\usepackage{makeidx}
\makeindex

\oddsidemargin=0.0in
\evensidemargin=-0.5in
\textwidth=6.5in
\textheight=9in


\renewcommand{\baselinestretch}{0.85}
\newenvironment{tightlist}
{\begin{description}\setlength{\itemsep}{-0.3\baselineskip}}
{\end{description}}

\renewcommand{\contentsname}{\textsf{Contents}}
\renewcommand{\chaptername}{}
\renewcommand{\thechapter}{}

\usepackage{fancyhdr}
\pagestyle{fancy}
\fancyhead{}% delete the default header
\fancyfoot{}% delete the default footer
\fancyhead[LE,RO]{\thepage}
\fancyhead[RE]{\textsf{http:// <TMPL_VAR NAME=website>: Community Resource Directory}}
\fancyhead[LO]{\nouppercase{\leftmark}}
% yes, we\'re putting the leftmark on the right-hand page.
\renewcommand{\footrulewidth}{0.5pt}
\fancyfoot[C]{\large{\textsf{\textbf{Printed on \today}\\\small{Supported by the Greater Lowell Workforce Investment Board and North Shore Community Action Programs}}}}

\begin{document}
\raggedright
\frontmatter
\title{\textsf{\textbf{\Huge  <TMPL_VAR NAME=website_description> }\\ Community Resource Directory}}
\author{Community Software Lab\\Lowell, MA}
\maketitle

\section*{\textsf{Copyright}}
The software to generate this document is Copyright \copyright{} 2007-2010 Community Software Lab Lowell, MA and licensed under the GNU AGPL v3 The data in document is in the public domain. Combining the data in this document with proprietary data will not remove it from the public domain.

\section*{\textsf{Acknowledgments}}

Thanks to the Greater Lowell Workforce Investment Board who paid to develop this report and to NorthShore Community Action Programs who paid to improve it.

\setcounter{tocdepth}{1}
\textsf{\tableofcontents}
\newpage
END

    my $t = HTML::Template->new( scalarref => \$text );
    $t->param( website             => $website );
    $t->param( website_description => $website_description );

    return $t->output();
}

sub about_mvhub {
    my $website = shift or croak 'missing Parameter: $website';
    my $website_description = shift
        or croak 'missing Parameter: $website_description';
    my $website_location = shift
        or croak 'missing Parameter: $website_location';

    my $text = <<END;
\\section*{\\textsf{About <TMPL_VAR NAME=website_description>}}
The <TMPL_VAR NAME=website_description> provides a database of community services for
<TMPL_VAR NAME=website_location>.
To find out more about the <TMPL_VAR NAME=website_description>, or to list your organization, please visit http://<TMPL_VAR NAME=website>
END

    my $t = HTML::Template->new( scalarref => \$text );
    $t->param( website             => $website );
    $t->param( website_description => $website_description );
    $t->param( website_location    => $website_location );

    return $t->output();
}

sub stop_word {
    my $word = shift;
    return 1 unless $word;
    return 1 if $word =~ /^\d*$/;
    return 1 if length $word < 2;
    return in_list( lc($word),
        qw/the of and a in to is was it for that with he be on I by s at you are had his not this have but from which she they or an were as we their been has that will would there/
    );
}

sub in_list {
    my $target = shift;
    return grep { $target eq $_ } @_;
}

sub get_programs_with_siblings {
    my $dbh     = shift or croak 'missing Parameter:$dbh';
    my $heading = shift or croak 'missing Parameter:$heading';
    $heading = MVHub::GeneratePDF::unclean($heading);

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'PROGRAM_AGENCY_X_PROGRAM_AGENCY_NAME');

    return $dbh->selectall_arrayref( $sql, { Slice => {} }, ($heading) );
}

sub get_agencies_with_multiple_programs {
    my $dbh     = shift or croak 'missing Parameter:$dbh';
    my $heading = shift or croak 'missing Parameter:$heading';
    $heading = MVHub::GeneratePDF::unclean($heading);

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_HEADING_PROGRAM_X_AGENCY_ID');

    return $dbh->selectall_arrayref( $sql, { Slice => {} }, ($heading) );

}

sub get_single_program_agencies {
    my $dbh     = shift or croak 'missing Parameter:$dbh';
    my $heading = shift or croak 'missing Parameter:$heading';
    $heading = MVHub::GeneratePDF::unclean($heading);

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_HEADING_PROGRAM_X_AGENCY_ID_EQUAL');

    my $result_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, ($heading) );
    return $result_aref;
}

# or agencies that have uncategorized programs
sub get_agencies_without_programs {
    my $dbh = shift or croak 'missing Parameter:$dbh';

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_PROGRAM_X_AGENCY_ID');

    return $dbh->selectall_arrayref( $sql, { Slice => {} } );

}

sub see_alsos {

    # given a db handle,
    # returns the "\index{foo|see{bar}}"s to be included at the top.

    my $dbh = shift;

    # get leads from non-first words in categories to the category
    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'CATEGORY_X_IN_CATEGORY_NAME');

    my @categories = @{ $dbh->selectcol_arrayref( $sql, { Slice => {} } ) };
    my $ret = join '', map {
        my $category = MVHub::GeneratePDF::clean($_);
        my @words = split( /\W/, $category );
        $category =~ s/,.*//;    # just point to the top level
        shift @words;            # no need for the first one.
        map { ( not stop_word($_) ) ? "\\index{$_|see{$category}}\n" : () }
            @words;
    } @categories;

    # get leads from aliases to categories
    # and get leads from non-first words in aliases to categories
    $sql = MVHub::Utils::DB::get_sql_select_statement(
        'ALIAS_CATEGORY_X_ALIAS_NAME_CATEGORY');

    my $aref = $dbh->selectall_arrayref( $sql, { Slice => {} } );
    $ret .= join '', map {
        my $alias = MVHub::GeneratePDF::clean( $_->{'alias_name'} );
        $alias = index_subheading($alias);
        my @alias_words = split( /\W/, $alias );
        my $category = MVHub::GeneratePDF::clean( $_->{'category_name'} );
        $category =~ s/,.*//;    # just point to the top level
        my ($first_word) = $category =~ m/^(\S*)/;
        shift @alias_words;      # no need for the first one.
        (   "\\index{$alias|see{$category}}\n",
            map {
                ( not stop_word($_) and ( $_ ne $first_word ) )
                    ? "\\index{$_|see{$category}}\n"
                    : ();
                } @alias_words
        );
    } @$aref;

    return $ret;
}

sub doc_foot {
    <<'END'
\addcontentsline{toc}{chapter}{\quad{}\enspace{}Index}
\printindex
\end{document}
END
}

sub section_heading {
    my $heading = shift;
    "\\chapter{\\textsf{$heading}}\n";
}

sub list_entries {

    # given a list of entry hash refs,
    # return the latex for a list of references.

    return unless @_;

    my $ret = "\\begin{tightlist}\n";
    for my $r (@_) {
        $ret .= "\\item $r->{'name'}: \\pageref{$r->{'id'}}\n";
    }
    $ret .= "\\end{tightlist}\n";
    return $ret;
}

sub categories_of {

    # given an entry hash ref and a db handle,
    # return the latex cleaned categories for the entry.

    my $r   = shift;
    my $dbh = shift;
    return () unless $r->{'program_id'};
    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'CATEGORY_PROGRAM_X_CATEGORY_NAME');
    my @bind_variables = ( $r->{'program_id'} );
    return
        map { MVHub::GeneratePDF::clean($_) }
        @{ $dbh->selectcol_arrayref( $sql, { Slice => {} }, @bind_variables )
        };
}

sub index_subheading {
    my $head = shift;
    $head =~ s/, */!/;
    $head =~ s/, */!/;

    # exactly twice, because latex limits subheadings to 3 levels.
    return $head;
}

sub indexing {

    # given an entry and a list of applicable categories,
    # returns the latex \label and \index-s

    my $r          = shift;
    my @categories = @_;

    my $ret = "\\label{$r->{'id'}}\n";
    for (@categories) {
        $_ = index_subheading($_);
        $ret .= "\\index{$_}\n";
    }
    return $ret;
}

sub optional {

    # for a value, label pair
    # return the latex for a list item

    my $val   = shift;
    my $label = shift;
    return ''              unless $val;
    return "\\item $val\n" unless $label;
    return "\\item[$label] \\hspace*{\\fill} \\\\ $val\n";
}

sub line_break {
    return "\\hspace*{\\fill} \\\\\n";
}

sub get_addr {

    # given an entry hashref, and a field prefix,
    # concatenate the fields which comprise the address
    # and return the latex.

    my $r   = shift;
    my $pre = shift;
    $pre .= $pre ? '_' : '';
    return '' unless $r->{"${pre}address1"};
    my $ret = qq|$r->{"${pre}address1"}\\\\\n|;
    $ret .= qq|$r->{"${pre}address2"}\\\\\n| if $r->{"${pre}address2"};
    $ret .= qq|$r->{"${pre}city"}, $r->{"${pre}state"} $r->{"${pre}zip"}\n|;
    return $ret;
}

sub split_and_join_field {
    my $field = shift;
    return '' unless $field;
    return ( join ', ', split( $DISGUSTING_FIELD_DELIMITER, $field ) );
}

sub list_programs {

    # given an entry, and a db handle,
    # returns a latex list of the programs, if it is an agency
    # TODO: the calling convention is inconsistent between this
    # and list_entries; change the name of one?

    my $r_href = shift;
    my $dbh    = shift;

    return if $r_href->{'program_id'};
    my $agency_id = $r_href->{'agency_id'};

    my $sql
        = MVHub::Utils::DB::get_sql_select_statement('AGENCY_PROGRAM_X_ALL');

    my $aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, ($agency_id) );
    return unless scalar(@$aref) > 1;

    $aref = MVHub::GeneratePDF::clean_merge_sort($aref);
    return list_entries(@$aref);

    # this break comes out a bit wider than I'd like:
    # the difference between this and the addresses is that they
    # are just a sequence of lines: this is a list, which
    # gets some padding at the top and the bottom by default.
}

sub agency_list_section {

    # given a db handle,
    # returns a latex section tag and a list of agencies.

    my $dbh = shift;

    my $sql = MVHub::Utils::DB::get_sql_select_statement('AGENCY_X_ALL');

    my $aref = $dbh->selectall_arrayref( $sql, { Slice => {} } );

    $aref = MVHub::GeneratePDF::clean_merge_sort( $aref, [] );

    # list ref forces it to treat it as a list of agencies,
    # rather than a list of programs. UGLY UGLY UGLY.

    my $ret = "\\section*{\\textsf{\\LARGE{Agencies}}}\n";
    foreach my $agency (@$aref) {
        $ret
            .= "\\contentsline{section}{$agency->{'name'}}{\\pageref{$agency->{'id'}}}\n";
    }
    $ret .= "\\newpage\n";
}

sub summary {

    # given an entry hash ref,
    # return the latex of a summary.

    my $r   = shift;
    my $dbh = shift;

    # synthesize a few fields.
    $r->{'street_address'}  = get_addr($r);
    $r->{'mailing_address'} = get_addr( $r, 'mail' );
    $r->{'languages'}       = split_and_join_field( $r->{'languages'} );
    $r->{'bus_lines'}       = split_and_join_field( $r->{'bus_lines'} );
    $r->{'programs'}        = list_programs( $r, $dbh );

    # these fields are paired, with "_description" implicit unless specified,
    # or else they should be suppressed for "no", "na", etc.
    my @optional_descriptions = (
        'languages',
        'wait_list',
        [ 'fees_charged'            => 'fees_explained' ],
        [ 'child_care_provided'     => 'child_care_description' ],
        [ 'volunteer_opportunities' => 'volunteer_information' ],
        'resource_library',
        'meeting_space',
        'news_letter',
        'public_net_access',
        'appointment',
        'eligibility',
    );

    # condense description fields and main fields
    # and suppress ones which don't say much.
    for (@optional_descriptions) {
        my ( $main, $desc ) = ref($_) ? @$_ : ( $_, $_ . '_description' );
        $r->{$main} .= " ($r->{$desc}) " if $r->{$desc};
        $r->{$main} = ''
            if in_list(
            lc( $r->{$main} ),
            'no', 'not applicable',
            'english', 'free'
            ) && not $r->{$desc};
    }

    # suppress fields likely to be redundant
    my @likely_redundancies = ( [ 'mailing_address' => 'street_address' ], );
    for (@likely_redundancies) {
        my ( $secondary, $primary ) = @$_;
        next unless $r->{$secondary} && $r->{$primary};
        $r->{$secondary} = '' if $r->{$secondary} eq $r->{$primary};
    }

    # list all the fields to be printed out,
    # (if there is data), and specify a label if
    # "_"->" " isn't good enough.
    # (note that these may be coming from program or agency records.
    #  if both an agency and a program have a field with the same
    #  name I'm not sure which takes priority, or which should take
    #  priority for that matter.)
    my @field_labels = (
        'programs',           # synthetic
        'street_address',     # synthetic
        'mailing_address',    # synthetic
        [ 'main_phone'      => 'phone' ],
        [ 'toll_free_phone' => 'toll free' ],
        [ 'hotline'         => 'hotline' ],
        [ 'tty_ttd'         => 'tty/ttd' ],
        'fax',
        'website',
        'email',
        [ 'program_schedule' => 'schedule' ],
        'hours',
        [ 'appointment' => 'appointment necessary' ],
        [ 'bring'       => 'items to bring' ],
        'bus_lines',
        'parking',
        'transportation',
        'handicapped_accessible',
        'other_information',
        'eligibility',
        'areas_served_description',
        'wait_list',
        [ 'fees_charged' => 'fees' ],
        'child_care_provided',
        'volunteer_opportunities',
        'resource_library',
        'meeting_space',
        'news_letter',
        [ 'public_net_access' => 'public internet access' ]
    );

    my $ret = "\\section*{\\textsf{$r->{'name'}}}\n";

    # Note double newlines so we get a paragraph break.
    $ret .= "$r->{'description'}\n\n";
    $ret .= "\\begin{tightlist}\n";
    for (@field_labels) {
        my ( $field, $label ) = ref($_) ? @$_ : ( $_, $_ );
        $label =~ s/_/ /g;
        $ret .= optional( $r->{$field}, $label );
    }
    $ret .= "\\end{tightlist}\n";

    return $ret;
}
