#!/usr/bin/perl 

# LICENSE: AGPLv3
# CONTACT: joe@thecsl.org
# PURPOSE: send emails to people asking people
#          to update records that are old

# goal is to send minimum number of emails to
# each contact. Several out of date records
# mean one email.

use warnings;
use strict;

use Getopt::Long;

use MVHub::Common;
use MVHub::Utils;
use MVHub::Utils::ConfigSimple;
use MVHub::Utils::DB;
use MVHub::Notifications;

# Command Line option handling
my $CFG_FILENAME = $ENV{MV_CONFIG_FILE};
my $EXECUTE      = 0;
Getopt::Long::GetOptions(
    'config=s' => \$CFG_FILENAME,
    'execute'  => \$EXECUTE
) or die "GetOptions failed";

# Config File Variables
my $CFG = MVHub::Utils::ConfigSimple::create_config_from($CFG_FILENAME);

my $MAX_NOTIFICATIONS = $CFG->param('NOTIFICATION.max_notifications');
my $EXPIRE_MONTHS     = $CFG->param('NOTIFICATION.expire_months');

my $DBH = MVHub::Utils::DB::get_dbh($CFG_FILENAME);

# Email Template Paths
my $TEMPLATE_PATH = $CFG->param('ABSOLUTE_PATH.template_text_dir');

my $FIRST_NOTIFICATION_TEMPLATE_FILE
    = "$TEMPLATE_PATH/reminder_email_01.tmpl";
my $SECOND_NOTIFICATION_TEMPLATE_FILE
    = "$TEMPLATE_PATH/reminder_email_02.tmpl";
my $THIRD_NOTIFICATION_TEMPLATE_FILE
    = "$TEMPLATE_PATH/reminder_email_03.tmpl";
my $FOURTH_NOTIFICATION_TEMPLATE_FILE
    = "$TEMPLATE_PATH/reminder_email_04.tmpl";
my $FINAL_NOTIFICATION_TEMPLATE_FILE
    = "$TEMPLATE_PATH/reminder_email_05.tmpl";

{    # Main
    my @date = Date::Calc::Add_Delta_YM( Date::Calc::Today(), 0,
        -($EXPIRE_MONTHS) );
    my $expire_date = join '-', @date;

    my $expired_agency_records_aref
        = MVHub::Notifications::get_expired_records(
        dbh               => $DBH,
        max_notifications => $MAX_NOTIFICATIONS,
        expire_date       => $expire_date,
        query_name        => 'AGENCY_X_EXPIRED_RECORDS'
        );

    my $expired_program_records_aref
        = MVHub::Notifications::get_expired_records(
        dbh               => $DBH,
        max_notifications => $MAX_NOTIFICATIONS,
        expire_date       => $expire_date,
        query_name        => 'PROGRAM_X_EXPIRED_RECORDS'
        );

    my %all_notifications = MVHub::Notifications::create_notifications_using(
        $expired_program_records_aref);
    %all_notifications = MVHub::Notifications::create_notifications_using(
        $expired_agency_records_aref, \%all_notifications );

    MVHub::Notifications::die_if_record_count_mismatch(
        'input_01' => $expired_agency_records_aref,
        'input_02' => $expired_program_records_aref,
        'output'   => \%all_notifications
    );

    my %final_notifications
        = MVHub::Notifications::shift_off_notifications_with_reminder_count(
        4, \%all_notifications );
    my %fourth_notifications
        = MVHub::Notifications::shift_off_notifications_with_reminder_count(
        3, \%all_notifications );
    my %third_notifications
        = MVHub::Notifications::shift_off_notifications_with_reminder_count(
        2, \%all_notifications );
    my %second_notifications
        = MVHub::Notifications::shift_off_notifications_with_reminder_count(
        1, \%all_notifications );
    my %first_notifications
        = MVHub::Notifications::shift_off_notifications_with_reminder_count(
        0, \%all_notifications );

    MVHub::Utils::assert( !scalar(%all_notifications),
        "Leftover notifications" );

    my $email_constants_href
        = MVHub::Notifications::get_email_constants_from($CFG);

    my @emails;
    push @emails,
        MVHub::Notifications::create_standard_notification_emails_with(
        $FIRST_NOTIFICATION_TEMPLATE_FILE,
        $email_constants_href, %first_notifications );
    push @emails,
        MVHub::Notifications::create_standard_notification_emails_with(
        $SECOND_NOTIFICATION_TEMPLATE_FILE,
        $email_constants_href, %second_notifications );
    push @emails,
        MVHub::Notifications::create_standard_notification_emails_with(
        $THIRD_NOTIFICATION_TEMPLATE_FILE,
        $email_constants_href, %third_notifications );
    push @emails,
        MVHub::Notifications::create_fourth_notification_emails_with(
        $FOURTH_NOTIFICATION_TEMPLATE_FILE,
        $email_constants_href, %fourth_notifications );
    push @emails,
        MVHub::Notifications::create_final_notification_emails_with(
        $FINAL_NOTIFICATION_TEMPLATE_FILE,
        $email_constants_href, %final_notifications );

    if ($EXECUTE) {
        MVHub::Notifications::send_emails_contained_in(@emails);
        %all_notifications = (
            %first_notifications, %second_notifications,
            %third_notifications, %fourth_notifications,
            %final_notifications
        );
        MVHub::Notifications::increment_reminder_levels_for(
            \%all_notifications, $MAX_NOTIFICATIONS );

        $DBH->begin_work();
        MVHub::Notifications::set_reminder_level_in_db( $DBH, $expire_date,
            \%all_notifications, 'AGENCY_X_REMINDERS_SENT' );
        MVHub::Notifications::set_reminder_level_in_db( $DBH, $expire_date,
            \%all_notifications, 'PROGRAM_X_REMINDERS_SENT' );
        $DBH->commit();
    }
    else {
        MVHub::Notifications::print_emails_contained_in(@emails);
    }
}
