#! /usr/bin/perl

# ---------------------------------------------------------------------
# This sub loads the $CATEGORIES_FILE (defined below) and inserts
# its categories into the database. See below for which tables are
# populated.
#
# The $CATEGORIES_FILE file should be in the format:
#
# [HEADING1]
# category1=
# category2=alias1
# category3=
# category4=alias1;alias2;...;aliasN
#
# [HEADING2]
# category5=
# category2=aliasB
#
# A category may be repeated under more than one heading
# Multiple categories can share the same alias
#
# All aliases defined for a category apply at all times, regardless
# of whether they were defined under different headings. In the example
# format shown above, both aliasB and alias1 will be defined as aliases
# for category2
#
# Since there are other tables whose data dependes on these tables,
# this script must NEVER be re-run on an existing site. Only run
# it once, after initial installation.
# ---------------------------------------------------------------------

use warnings;
use strict;

use Config::IniFiles;
use Time::HiRes;

use MVHub::Utils;

use MVHub::Common;

my $CATEGORIES_FILE = '../../config/guide/categories.conf';

# The populated tables:
my $HEADING_TABLE          = 'heading';
my $CATEGORY_TABLE         = 'category';
my $ALIAS_TABLE            = 'alias';
my $HEADING_CATEGORY_TABLE = 'heading_category';

if ( @ARGV == 1 ) {
    ( -f $ARGV[0] and -r $ARGV[0] ) or die "File $ARGV[0] not found/readable";
    $CATEGORIES_FILE = $ARGV[0];
}

# Tie the sections/parameters from any existing file to a hash,
# for retrieving existing values.
my %categories;

if ( !-f $CATEGORIES_FILE ) {
    die "$CATEGORIES_FILE does not exist";
}

tie %categories, 'Config::IniFiles',
    (
    -file                => $CATEGORIES_FILE,
    -allowedcommentchars => '#'
    );

# If parsing of config file failed, print error + die
if ( !scalar keys %categories ) {
    die "Encountered the following errors parsing $CATEGORIES_FILE:\n\n"
        . join( "\n", @Config::IniFiles::errors ) . "\n";
}

# Let's time our query for fun
my $t0 = Time::HiRes::gettimeofday();

my $dbh = MVHub::Utils::DB::get_dbh();
$dbh->{AutoCommit} = 0;

my %added_categories;    # category_name => category_id

my %added_aliases;       # category_name => alias_name

foreach my $heading ( sort( keys %categories ) ) {
    MVHub::Utils::DB::insert_one_record(
        dbh         => $dbh,
        table       => $HEADING_TABLE,
        values_href => { 'heading_name' => $heading }
    );

    foreach my $category ( keys %{ $categories{$heading} } ) {

        # We don't want to insert a category more than once if it is
        # listed under multiple headings
        if ( !exists $added_categories{ lc($category) } ) {

            # print "INSERT $category\n";
            my $next_category_id
                = MVHub::Utils::DB::get_next_in_sequence( $dbh,
                'category_id_sequence' );
            MVHub::Utils::DB::insert_one_record(
                dbh         => $dbh,
                table       => $CATEGORY_TABLE,
                values_href => {
                    'category_name' => $category,
                    'stems_name'    => MVHub::Common::get_stems($category),
                    'category_id'   => $next_category_id
                }
            );
            $added_categories{ lc($category) } = $next_category_id;
        }

        # print "INSERT $category FOR $heading\n";
        MVHub::Utils::DB::insert_one_record(
            dbh         => $dbh,
            table       => $HEADING_CATEGORY_TABLE,
            values_href => {
                'category_id'  => $added_categories{ lc($category) },
                'heading_name' => $heading
            }
        );

        my @aliases = split( /\s*;\s*/,
            MVHub::Utils::trim( $categories{$heading}{$category} ) );
        foreach my $alias (@aliases) {

           # If the alias isn't already inserted into the db for this category
            if ( !grep( /^$alias$/i, @{ $added_aliases{ lc($category) } } ) )
            {

                # print "INSERT $alias INTO $category\n";
                MVHub::Utils::DB::insert_one_record(
                    dbh         => $dbh,
                    table       => $ALIAS_TABLE,
                    values_href => {
                        'category_id' => $added_categories{ lc($category) },
                        'alias_name'  => $alias,
                        'stems_name'  => MVHub::Common::get_stems($alias)
                    }
                );

             # Add the alias to the list of inserted aliases for this category
                push @{ $added_aliases{ lc($category) } }, lc($alias);
            }
        }

    }
}

# All inserts are done, commit
$dbh->commit();

# Benchmarking
my $t1 = Time::HiRes::gettimeofday();
print "\n\n***Inserted in ", ( $t1 - $t0 ), " seconds.\n";

