#!/usr/bin/perl

use strict;
use warnings;

use Carp;
use File::Temp;
use Test::More;
use Params::Validate;

use TestHelper;
use TestHelper::History;

my $PERLTIDYRC = "lib-mvhub/t/conf/perltidyrc";

{    # main
    if ( !_have_right_perltidy_setup() ) {

        plan skip_all =>
            'Perl::Tidy version 20090616 and valid t/perltidyrc required for this test';
        exit 0;    # no error
    }

    my $perl_files_aref = _get_all_perl_files();

    my $history_href;
    $history_href = TestHelper::History::load_test_history();

    my $files_to_test_aref = TestHelper::History::files_needing_retest(
        raw_files_aref => $perl_files_aref,
        test_name      => 'perltidy',
        history_href   => $history_href,
    );

    my @files_tested_ok = _test_perltidy($files_to_test_aref);

    $history_href = TestHelper::History::set_time_tested(
        tested_files_aref => \@files_tested_ok,
        test_name         => 'perltidy',
        history_href      => $history_href,
    );

    TestHelper::History::store_test_history($history_href);
}    # main

sub _have_right_perltidy_setup {
    eval {
        require Perl::Tidy;
        $Perl::Tidy::VERSION eq '20090616'
            or die "need Perl::Tidy version 20090616";
    };

    return defined $@;
}

sub _get_all_perl_files {
    my @perl_files;
    my @all_files = get_files_from( '.', q{'*'} );
    foreach my $file (@all_files) {
        next if $file =~ m/\.bak$/;
        my $result = `/usr/bin/file -b $file`;
        push @perl_files, $file if $result =~ m/perl/i;
    }

    @perl_files = map { s#^.*lib-mvhub/t/\.\.#lib-mvhub#;  $_; } @perl_files;
    @perl_files = map { s#^.*app-mvhub/t/\.\.#app-mvhub/#; $_; } @perl_files;
    @perl_files = map { s#^\./##;                          $_; } @perl_files;

    return \@perl_files;
}

sub _test_perltidy {

    my $files_aref = shift;
    my @files_to_fix;
    my @files_tested_ok;

    my $test_count = scalar @$files_aref;
    if ( !$test_count ) {
        plan skip_all => 'files not modified skipping tests';
        return;
    }

    Test::More::plan tests => ($test_count);

    foreach my $file (@$files_aref) {

        my $destination = ( File::Temp::tempfile() )[1];
        Perl::Tidy::perltidy(
            source      => $file,
            destination => $destination,
            perltidyrc  => $PERLTIDYRC,

        );

        my $ok = ok( TestHelper::no_difference( $file, $destination ),
            "perltidy: $file" );

        if ($ok) {
            push @files_tested_ok, $file;
        }
        else {    # yes ok
            push @files_to_fix, $file;
        }

        unlink $destination

    }    # foreach

    diag( _build_fix_msg( $PERLTIDYRC, @files_to_fix ) )
        if scalar @files_to_fix;

    return @files_tested_ok;

}

sub _is_perl {

    Validate::Params::validate_pos( @_, 1 );
    my $file   = shift;
    my $result = `file $file`;
}

sub _build_fix_msg {
    my $perltidyrc = shift or croak 'missing pararm: $perltidyrc';
    my @files      = @_    or die 'missing parameter: @files';
    my $files = join ' ', @files;

    my $cmd = <<"END";
TO FIX: 

      cdw && perltidy  -b --profile=$perltidyrc $files
       
END
    return $cmd;
}

