
package MVHub::AgencyAccount;

our ($VERSION) = '$Revision: 1670 $' =~ /([.\d]+)/;

use strict;
use warnings;

use CGI::Session;
use MIME::Entity;
use URI::URL;
use DateTime;
use HTML::Template;

use base 'MVHub::CGIAppBase';

use MVHub::AuthAccount;
use MVHub::Common;
use MVHub::Utils;

my $AGENCY_HOME_TEMPLATE_FILE  = 'agency_home.tmpl';
my $AH_AGENCY_ID_TAG           = 'AGENCY_ID_TAG';
my $AH_AGENCY_NAME_TAG         = 'AGENCY_NAME_TAG';
my $AH_LAST_UPDATED_TAG        = 'LAST_UPDATED_TAG';
my $AH_MESSAGE_TAG             = 'MESSAGE_TAG';
my $AH_AGENCY_CONTACT_NAME_TAG = 'agency_contact_name';
my $AH_PROGRAM_LOOP_TAG        = 'PROGRAM_LOOP_TAG';

my $L_AGENCY_LIST_TAG   = 'AGENCY_LIST_TAG';
my $L_ERROR_MESSAGE_TAG = 'ERROR_MESSAGE_TAG';

my $NO_REPORT_TEMPLATE = 'report_not_ready.tmpl';

# TODO put in config file
my $TEST_COOKIE_NAME = 'guide_test_cookie';

# Available CGI::Application params
# caller_id   agency_id

#TODO: move this somewhere else
my $javascript1 = <<JS;
<script type="text/javascript">
    function confirmRemove() {
        return confirm("Are you sure you want to delete this program?");
    }
    function askWhyHiding ( agency_id, program_id ) {
        var form = document.hide_program;
        var hide_reason =
            prompt("Please give a reason for hiding this program.", "", "MVHub Community Services Directory");
        if ( hide_reason ) {
            form.reason.value = hide_reason;
            form.agency_id.value = agency_id;
            form.program_id.value = program_id;
            form.submit();
        }
    }
    function refreshTopFrame() {
        if ( top != self ) {
            top.admin_bar.location.reload(true);
        }
    }
</script>
JS

my %JAVASCRIPT_BY_MODE = ( 'display_agency_home' => $javascript1, );

sub setup {
    my $self = shift;

    $self->run_modes(
        [   'show_login_page',
            'request_password',
            'display_agency_home',
            'display_agency_report',
            'do_login',
            'remove_agency',
            'ql'    # quick log-in, purposely abbreviated for obfuscation
                    # you should UPDATE @CGIAppBase::LOGIN_ONLY_MODES
                    # when adding new modes that are LOGIN_ONLY.
        ]
    );
    $self->start_mode('show_login_page');
    $self->mode_param('rm');
}

sub cgiapp_prerun {
    my $self           = shift;
    my $cgi            = $self->query();
    my $requested_mode = $self->get_current_runmode();

    # Modes that require the user to first be logged in:
    my @LOGIN_ONLY_MODES = qw(display_agency_home remove_agency);

    MVHub::Utils::clean_cgi_params($cgi);

    $self->header_add( -Cache_Control => 'no-store' );

    my $cookie_name = $self->get_config_param('COOKIES.auth_cookie_name');
    my $tmp_dir     = $self->get_config_param('ABSOLUTE_PATH.tmp_dir');

    my $session_id = $cgi->cookie($cookie_name);

    my $cgi_agency_id = $cgi->param('agency_id');

    # TODO put Directory in config file [PATH].session_dir
    my $session
        = new CGI::Session( "driver:File", $session_id,
        { Directory => $tmp_dir } )
        or croak CGI::Session->erstr();

    if ( !defined $session_id || $session_id ne $session->id() ) {

        if ( grep /$requested_mode/, @LOGIN_ONLY_MODES ) {
            $self->param(
                login_error => 'You must log in to access this page.' );
            $self->prerun_mode('show_login_page');
        }
    }
    else {
        my $caller_id = $session->param('caller_id');
        MVHub::Utils::assert( $caller_id,
            "Session object does not have a caller_id value" );
        my $agency_id;

        # If the caller is not an admin, their caller_id is their agency_id.
        # TODO: better error message if a bad agency_id is passed in
        if ( $caller_id != $MVHub::Common::ADMIN_ID ) {

            $agency_id = $caller_id;

            if ( defined $cgi_agency_id && $cgi_agency_id != $agency_id ) {
                if ( grep /$requested_mode/, @LOGIN_ONLY_MODES ) {
                    $self->param( login_error =>
                            'You must log in to access this page.' );
                    $self->prerun_mode('show_login_page');
                }
            }
        }
        else {

            $agency_id = $cgi->param('agency_id');

            # If we don't get an agency id, it's either a programming error or
            # possibly the user is currently logged in as admin, goes back
            # to the home/search page, and then clicks on agency log-in.
            # We'll optimistically assume the latter, and just redirect the
            # user to the admin site.
            if ( !$agency_id ) {
                $self->redirect( absolute_path => '/html/admin/' );
            }
        }

        # If they requested the show_login_page, send them to their
        # agency home page (if they aren't an admin)
        if (   $requested_mode eq 'show_login_page'
            && $caller_id != $MVHub::Common::ADMIN_ID )
        {
            $self->prerun_mode('display_agency_home');
        }

        # They are logged in, set the caller_id and agency_id params
        $self->param( agency_id => $agency_id );
        $self->param( caller_id => $caller_id );
    }
}

sub cgiapp_postrun {
    my $self       = shift;
    my $output_ref = shift;
    my $cgi        = $self->query();
    my $dbh        = $self->dbh();

    my $calling_mode = $self->get_current_runmode();
    my $caller_id    = $self->param('caller_id');

    # If it's an admin, they probably don't want the pretty, space-wasting GUI
    # However, we neeFd some additional HTML tags, and the javascript
    if ( defined $caller_id && $caller_id == $MVHub::Common::ADMIN_ID ) {
        my $js = $JAVASCRIPT_BY_MODE{$calling_mode} || '';
        my $page_body = $$output_ref;
        $$output_ref = "<html><head>$js</head>";

        # We need to refresh the top frame when the agency home is displayed
        # in admin mode, because there is a good chance that the admin just
        # added that agency, and the drop-down box in the admin frame needs
        # to be refreshed to reflect the addition
        if ( $calling_mode eq 'display_agency_home' ) {
            $$output_ref .= '<body onLoad="refreshTopFrame()">';
        }
        else {
            $$output_ref .= '<body>';
        }

        $$output_ref .= $page_body . '</body></html>';
        return;
    }

    # Add the space wasting GUI:
    $$output_ref = MVHub::Page::prepare_page(
        body                          => $$output_ref,
        javascript                    => $JAVASCRIPT_BY_MODE{$calling_mode},
        transfer_internal_style_sheet => 1,
        show_left_menu                => 1,
        show_search_bar               => 0,
        site_code     => $self->get_config_param('SITE.website_code'),
        site_location => $self->get_config_param('SITE.location'),
        search_box_help_text =>
            $self->get_config_param('SITE.search_box_help_text'),
    );
}

# ************ Run Modes **************

=head1 METHODS

=head2 show_login_page()

  method to display page that shows list of agencies
  and allows user to choose agency and enter password

  uses the HTML::Template file logon.tmpl

=cut

sub show_login_page {
    my $self = shift;

    my $cgi           = $self->query();
    my $error_message = $self->param('login_error');
    if ( !defined $error_message ) {
        $error_message = $cgi->param('error_code');
    }
    my $dbh = $self->dbh();

    my $tmpl = $self->load_tmpl('login.tmpl')
        or die "Failed to open template file 'login.tmpl'";

    my $agency_list_src = MVHub::Common::get_agency_select_list( $dbh, 7 );

    $tmpl->param( $L_AGENCY_LIST_TAG   => $agency_list_src );
    $tmpl->param( $L_ERROR_MESSAGE_TAG => $error_message );

    # Set a test cookie, so we can tell the user if they have cookies disabled
    # and that the site won't work until they enable cookies
    my $cookie = CGI::cookie(
        -name  => $TEST_COOKIE_NAME,
        -value => 'COOKIE_SET',
        -path  => '/cgi-bin/mvhub',
    );
    $self->header_add( -cookie => $cookie );

    return $tmpl->output();
}

=head2 request_password

  If called without an cgi "email" param, displays a form for requesting
  a password reminder.

  Otherwise, send an email with password(s) for any agency for which the
  given email address is either an agency or program contact and display
  confirmation page.

=cut 

sub request_password {
    my $self      = shift;
    my $cgi       = $self->query();
    my $email     = $cgi->param('email');
    my $page_body = "<h3>Password Reminder</h3>";

    my $dbh = $self->dbh();

    if ($email) {

        # Process a password request
        my $sql = $self->retr_select_sql('AGENCY_X_AGENCY_NAME_PASSWORD');

        my $agency_passwords_aref
            = $dbh->selectall_arrayref( $sql, { Slice => {} }, $email,
            $email );
        if (@$agency_passwords_aref) {
            my $from_name = $self->get_config_param('NOTIFICATION.team_name')
                or die "failed to get team_name";
            my $from_email
                = $self->get_config_param('NOTIFICATION.admin_email')
                or die "failed to get email";
            _send_password_reminder( $email, $agency_passwords_aref,
                "$from_name <$from_email>" );

            # Success page
            return "We've sent you an email with your password. "
                . '<a href="/cgi-bin/mvhub/agency.pl?rm=show_login_page">Return to the log-in page<a>';
        }
        else {
            $page_body
                = "<p>Sorry, there is no agency or program contact listed in our"
                . " database with e-mail address <b>$email</b>. Please ensure "
                . "that it is entered correctly, or try a different address.</p>";
        }
    }

    # If we haven't returned yet, display the password request form
    $page_body
        .= "We can send a password reminder to any e-mail address already registered "
        . "for an agency or program contact. Please enter your registered "
        . "e-mail address below to have your password sent there.<br>";

    $page_body
        .= $cgi->start_form( -method => "POST", -action => "agency.pl" )
        . $cgi->hidden( -name => "rm", -default => "request_password" )
        . $cgi->textfield(
        -name    => "email",
        -default => $email,
        -size    => 60
        )
        . $cgi->submit( -name => "Get Password" )
        . $cgi->endform();

    return $page_body;
}

sub do_login {
    my $self = shift;
    my $cgi  = $self->query();
    my $dbh  = $self->dbh();

    my $agency_id   = $cgi->param('agency_id');
    my $password    = $cgi->param('agency_password');
    my $cookie_name = $cgi->cookie($TEST_COOKIE_NAME);
    my $error_code;

    if ( !defined $agency_id || $agency_id eq q{} ) {
        $error_code = "Please select an agency.";
    }
    elsif ( !defined $password || $password eq q{} ) {
        $error_code = "Please enter a password.";
    }
    elsif (
        !MVHub::AuthAccount::is_correct_password(
            $dbh, $agency_id, $password
        )
        )
    {
        $error_code = "Incorrect password for this agency. Please try again.";
    }
    elsif ( !defined $cookie_name || $cookie_name ne 'COOKIE_SET' ) {
        return
              "Note - we were unable to set a cookie in your browser. "
            . "In order for you to use the agency administration features, "
            . "you must have cookies enabled. Please enable them and try "
            . 'to <a href="/cgi-bin/mvhub/agency.pl">log in</a> again.';
    }

    if ( defined $error_code ) {
        $self->redirect(
            cgi_params_href => {
                rm         => 'show_login_page',
                error_code => $error_code
            }
        );
        return "Login failed. Redirecting back to login page.";
    }

    # Log in was successful
    # Initialize a session, then redirect to display_agency_home
    my $cookie = MVHub::AuthAccount::init_session( caller_id => $agency_id );
    $self->header_add( -cookie => $cookie );
    $self->redirect( cgi_params_href => { rm => 'display_agency_home' } );

    return "Login successful - redirecting to agency $agency_id home page";
}

sub display_agency_home {
    my $self      = shift;
    my $cgi       = $self->query();
    my $agency_id = $self->param('agency_id');
    my $message   = $cgi->param('message');      # optional
    my $dbh       = $self->dbh();

    my $updates_are_pending = 0;

    my $agency_href = MVHub::Utils::DB::select_agency_record(
        'agency_id' => $agency_id,
        'dbh'       => $dbh
    );
    my $caller_id = $self->param('caller_id');

    my $template = $self->load_tmpl(
        $AGENCY_HOME_TEMPLATE_FILE,
        global_vars       => 1,    # for is_admin
        loop_context_vars => 1     # for coloring every other program row
    ) or die("Unable to open $AGENCY_HOME_TEMPLATE_FILE");

    my $last_update = $agency_href->{last_updated};
    my $expire_months
        = int $self->get_config_param('NOTIFICATION.expire_months');
    my $agency_needs_updating = has_expired( $last_update, 6 );

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_X_QUICK_LOGIN_PASSWD');
    my $ql_id = ( $dbh->selectrow_array( $sql, {}, $agency_id ) )[0];

    $template->param(
        'QUICK_LOGIN_URL' => MVHub::Utils::generate_quick_login_url(
            $ENV{HTTP_HOST}, $agency_id, $ql_id
        ),
        $AH_AGENCY_ID_TAG   => $agency_id,
        $AH_MESSAGE_TAG     => $message,
        $AH_AGENCY_NAME_TAG => $agency_href->{agency_name},
        $AH_AGENCY_CONTACT_NAME_TAG =>
            "$agency_href->{contact_first_name} $agency_href->{contact_last_name}",
        $AH_LAST_UPDATED_TAG  => $agency_href->{last_updated},
        agency_needs_updating => $agency_needs_updating,
        expiration_duration   => "$expire_months months",
    );

    my @bind_variables = ($agency_id);
    $sql = MVHub::Utils::DB::get_sql_select_statement('PROGRAM_X_AGENCY_ID');

    my $programs_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, @bind_variables );

    foreach my $record_href (@$programs_aref) {
        $record_href->{'last_updated'}
            = MVHub::Utils::yyyy_mm_dd_to_mm_dd_yyyy(
            $record_href->{'last_updated'} );
    }

    if ( scalar @$programs_aref ) {

        # To each program record, add a flag indicating if it has expired
        foreach my $program_href (@$programs_aref) {
            $program_href->{program_needs_updating}
                = has_expired( $program_href->{last_updated},
                $expire_months );
        }
        $template->param( $AH_PROGRAM_LOOP_TAG => $programs_aref );
    }

    $template->param( updates_pending => 1 ) if $updates_are_pending;

    return $template->output();
}

# This displays a per-agency usage report that's generated daily.
# The reports are located in /reports/agency_usage_reports/
# and are readable only by the mvhub2 user & group.  This should prevent most
# people from viewing them.
sub display_agency_report {
    my $self = shift;
    my $cgi  = $self->query();

    # Only retrieve the agency_id parameter as
    # the report is the same regardless whether an agency
    # or admin user is viewing the report.
    my $agency_id = $cgi->param('agency_id');

    # Check who's calling the script.  The admin user should
    # be able to view all reports; agency users should only
    # be able to view their own reports.  Return an error
    # for anyone else.
    my $caller_id = $self->param('caller_id');
    MVHub::Utils::assert(
        ( $caller_id == $MVHub::Common::ADMIN_ID )
            || ( $caller_id == $agency_id ),
        "Error when displaying agency traffic reports--caller id ($caller_id) doesn't match agency id"
    );

   # Slurp the report file into a single variable,
   # add a few navigational features to the variable, then
   # return a reference to the variable.  We've already used HTML::Template to
   # to generate the report  see: generate_agency_usage_reports.pl;
   # there's no reason to use it again.
    my $report_file
        = $MVHub::Common::AGENCY_USAGE_REPORTS_PATH . "html/$agency_id.inc";
    if ( -f $report_file ) {
        open AGENCY_REPORT_HTML,
            $MVHub::Common::AGENCY_USAGE_REPORTS_PATH . "html/$agency_id.inc"
            or die "Could not open $report_file: $!";
        local $/;
        my $agency_report = <AGENCY_REPORT_HTML>;
        close AGENCY_REPORT_HTML;
        return \$agency_report;
    }
    else {
        my $template = $self->load_tmpl(
            filename    => $NO_REPORT_TEMPLATE,
            global_vars => 0
        );
        $template->param( agency_id => $agency_id );

        return $template->output();
    }
}

sub remove_agency {
    my $self      = shift;
    my $cgi       = $self->query();
    my $caller_id = $self->param('caller_id');
    my $agency_id = $self->param('agency_id');
    my $dbh       = $self->dbh();

    MVHub::Utils::DB::remove_agency( $dbh, $agency_id );

    if ( $caller_id == $MVHub::Common::ADMIN_ID ) {
        return "Agency removed.";
    }
    else {

        # dan 2008-02-07 not sure this code is actually ever reached

        # TODO: Meta-refresh to logout
        return "Your agency has been removed.";
    }
}

# ql stands for quick-login (purposely abbreviated to conceal its devious purpose).
# The ql runmode can be used from a hyper-link where we want to log-in a user in one click,
# e.g. from within an email.
sub ql {
    my $self               = shift;
    my $cgi                = $self->query();
    my $dbh                = $self->dbh();
    my $agency_id          = $cgi->param('aid');
    my $quick_login_id     = $cgi->param('qlid');
    my $old_quick_login_id = $cgi->param('id');

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_X_QUICK_LOGIN_MATCH');
    my @bind_values = ( $agency_id, $quick_login_id );
    my @is_valid = $dbh->selectrow_array( $sql, {}, @bind_values );

    if ( $is_valid[0] ) {
        my $cookie
            = MVHub::AuthAccount::init_session( caller_id => $agency_id );
        $self->header_add( -cookie => $cookie );
        $self->param( 'agency_id', $agency_id );
        return $self->display_agency_home();

    }
    elsif ($old_quick_login_id) {

        return $self->old_quick_login_message();
    }
    else {
        $self->param( 'login_error',
            'Invalid quick log-in ID, please log in manually' );
        return $self->show_login_page();
    }
}

################ non run-mode functions #################

# Has the given last-updated date (in mm-dd-yyyy format) expired?
sub has_expired {
    my ( $mdy_date, $expire_months ) = @_;

    my ( $month, $day, $year ) = split( '-', $mdy_date );
    my $dt_date
        = DateTime->new( year => $year, month => $month, day => $day );

    my $now = DateTime->now();
    die "last update date ($mdy_date) cannot be in the future"
        if ( $now < $dt_date );

    my $age_duration = DateTime->now() - $dt_date;
    return $age_duration->delta_months() >= $expire_months;
}

sub old_quick_login_message {
    my $self = shift;
    my $template_file
        = $self->get_config_param('ABSOLUTE_PATH.template_html_dir')
        . '/old_quick_login_message.tmpl';
    my $tmpl        = HTML::Template->new( filename => $template_file );
    my $admin_name  = $self->get_config_param('NOTIFICATION.admin_name');
    my $admin_email = $self->get_config_param('NOTIFICATION.admin_email');
    my $admin_phone = $self->get_config_param('NOTIFICATION.admin_phone');
    $tmpl->param(
        admin_name  => $admin_name,
        admin_email => $admin_email,
        admin_phone => $admin_phone,
    );
    return $tmpl->output();
}

sub redirect {
    my $self = shift;
    my %args = @_;

# The absolute path is the abs_path part of a URL, as show below (from the HTTP 1.1 RFC)
# http:" "//" host [ ":" port ] [ abs_path [ "?" query ]]
# TODO: Send an e-mail to CGI.pm maintainers to nitpick
    my $absolute_path   = delete $args{absolute_path};
    my $cgi_params_href = delete $args{cgi_params_href};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $cgi = $self->query();

    # This returns the "http:" "//" host [ ":" port ]" section of the URL
    my $base_url = $cgi->url( -base => 1 );

    # If an absolute_path is not given, get the current absolute_path
    if ( !defined $absolute_path ) {
        $absolute_path = $cgi->url( -absolute => 1 );
    }

    my $url = URI::URL->new_abs( $absolute_path, $base_url );
    if ($cgi_params_href) {
        $url->query_form($cgi_params_href);
    }

    $self->header_add( -location => $url->as_string );
}

=head2 _send_password_reminder

  Parameters: 
	$email  		# destination address

	$agency_passwords_aref 	# reference to array of hash refs
				# containing Agency and password 
				# key/value pairs

	$from			# the sender email 

   Takes paraemters, build message body and sends email

=cut 

sub _send_password_reminder {
    my ( $email, $agency_passwords_aref, $from ) = @_;

    my $body = "Here are the password(s) you requested:\n";

    foreach (@$agency_passwords_aref) {
        $body
            .= "\n\nAgency: "
            . $_->{agency_name}
            . "\nPassword: "
            . $_->{password};
    }

    $body
        .= "\n\nIf you did not request this information, please let us know by "
        . "replying to this e-mail.\n\n--The MVHub Team";

    my $top = MIME::Entity->build(
        Type    => "text/plain",
        From    => $from,
        To      => $email,
        Subject => "MVHub Password Reminder",
        Data    => $body
    );

    MVHub::Common::sendmail($top);
}
1;
__END__

=head1 NAME

MVHub::AgencyAccount - CGI::Application class containing methods for authentication, and account management

=head1 SYNOPSIS

    use MVHub::AgencyAccount;
    my $app = MVHub::AgencyAccount->new();
    $app->run();

=head1 DESCRIPTION

 Used by agency.pl 

=head1 LAST MODIFIED

Last modified on $Date: 2009-02-12 11:37:32 -0500 (Thu, 12 Feb 2009) $ by $Author: omacneil $

=head1 BUGS

This Class overlaps with subs in MVHub::AuthAccount

=head1 LICENSE

The GNU Affero General Public License, version 3.0 or later

=head1 COPYRIGHT

Community Software Lab, Inc ( http://thecsl.org )

=cut

