package MVHub::AgencyFieldDefs;
require Exporter;

use strict;
use warnings;

use MVHub::Common;

our @ISA    = qw(Exporter);
our @EXPORT = qw(get_agency_field_definitions_href);

our ($VERSION) = '$Revision: 1627 $' =~ /([.\d]+)/;

# A list of fields in the order that that you might want to display them
# (somewhere besides the form, whose display is dictated by templates),
# e.g. in an email.
# Excludes the date_updating_option, password, and password_verification fields.
our @FIELDS = (
    'last_updated',               'agency_id',
    'agency_name',                'agency_alias',
    'address1',                   'address2',
    'city',                       'state',
    'zip',                        'main_phone',
    'toll_free_phone',            'hotline',
    'fax',                        'tty_tdd',
    'website',                    'email',
    'email_verification',         'mission',
    'hours',                      'other_info',
    'mail_address1',              'mail_address2',
    'mail_city',                  'mail_state',
    'mail_zip',                   'chief',
    'title',                      'parent_organization',
    'resource_library',           'resource_library_description',
    'meeting_space',              'meeting_space_description',
    'staff_internet_access',      'contact_first_name',
    'contact_last_name',          'contact_phone',
    'contact_fax',                'contact_email',
    'contact_email_verification', 'contact_email_available',
    'admin_notes'
);

# The field definitions. These are divided into 3 hashes (representing the 3 main sections
# of the displayed form) just for organizational purposes. It would also make it easy
# if we ever wanted to break the form into multiple pages (in which we'd refactor
# get_agency_field_definitions() to take the name of a hash).

my %agency_field_defs = (
    agency_id    => { type => 'hidden' },
    last_updated => {
        type  => 'static',
        label => 'Record last updated on:'
    },
    password => {
        type     => 'password',
        label    => 'Password (must be at least 6 characters)',
        validate => '/^.{6,}/',
        required => 1
    },
    password_verification => {
        type     => 'password',
        label    => 'Password (please verify)',
        required => 1,
        validate => {
            perl       => 'eq $form->field("password")',
            javascript => '== form.password.value'
        }
    },
    date_updating_option => {
        type => 'checkbox',
        options =>
            [ { 'yes' => 'Check here to update this date upon submission' } ]
    },
    agency_name  => { required => 1 },
    agency_alias => { label    => 'Also Known As' },
    address1     => { label    => 'Address 1', required => 1 },
    address2     => { label    => 'Address 2' },
    city         => { required => 1 },
    state => { type => 'select', options => [ 'MA', 'NH' ], required => 1 },
    zip =>
        { size => 5, maxlength => 5, validate => 'ZIPCODE', required => 1 },

    mail_address1 => { label => 'Address 1' },
    mail_address2 => { label => 'Address 2' },
    mail_city     => { label => 'City' },
    mail_state =>
        { type => 'select', options => [ 'MA', 'NH' ], label => 'State' },
    mail_zip =>
        { size => 5, maxlength => 5, validate => 'ZIPCODE', label => 'Zip' },

    main_phone => {
        size      => '24',
        maxlength => '24',
        validate  => $MVHub::Common::PHONE_REG_EX,
        comment   => '(ex: 123-456-7890)',
        required  => 1
    },

    toll_free_phone => {
        size      => '24',
        maxlength => '24',
        comment   => '(ex: 800-456-7890)',
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    hotline => {
        size      => '24',
        maxlength => '24',
        label     => 'Hotline/Emergency/After Hours #',
        comment   => '(ex: 123-456-7890)',
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    fax => {
        size      => '24',
        maxlength => '24',
        comment   => '(ex: 123-456-7890)',
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    tty_tdd => {
        size      => '24',
        maxlength => '24',
        label     => 'TTY/TDD',
        validate  => $MVHub::Common::PHONE_REG_EX,
        comment   => '(ex: 123-456-7890)'
    },
    website => {},
    email   => {
        size     => 30,
        validate => 'EMAIL'
    },
    email_verification => {
        size     => 30,
        validate => {
            perl       => 'eq $form->field("email")',
            javascript => '== form.email.value'
        }
    },
    mission => {
        type     => 'textarea',
        rows     => 10,
        label    => 'Agency Mission',
        required => 1
    },
    hours => {
        type     => 'textarea',
        rows     => 3,
        label    => 'Days/Hours',
        required => 1
    },
    other_info => {
        type  => 'textarea',
        rows  => 3,
        label => 'Other Information'
    },
    chief => {
        size     => 30,
        label    => 'Person in Charge',
        required => 1
    },
    title => {
        size     => 30,
        required => 1
    },
    parent_organization => {},
    admin_notes         => {
        type => 'textarea',
        label =>
            'Administrative Notes (only shown here, and only to administrator)',
        comment => '<br>ex: 12/15/2005: Left phone message. (FS)<br>'
    }
);

my %contact_field_defs = (
    contact_first_name => {
        label    => 'First Name',
        size     => 24,
        required => 1
    },
    contact_last_name => {
        label    => 'Last Name',
        size     => 24,
        required => 1
    },
    contact_phone => {
        label     => 'Phone',
        size      => 24,
        maxlength => 24,
        comment   => '(ex: 123-456-7890)',
        required  => 1,
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    contact_fax => {
        label     => 'Fax',
        size      => 24,
        maxlength => 24,
        comment   => '(ex: 123-456-7890)',
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    contact_email => {
        label    => 'Email Address',
        size     => 30,
        required => 1,
        validate => $MVHub::Common::ALLOW_CONTACT_EMAIL_REGEX
    },
    contact_email_verification => {
        label    => 'E-mail Address (please verify)',
        size     => 30,
        required => 1,
        validate => {
            perl       => 'eq $form->field("contact_email")',
            javascript => '== form.contact_email.value'
        }
    },
    contact_email_available => {
        label => 'Please make my e-mail available to other non-profits '
            . 'so I can be informed of events &amp;  programs.',
        options  => [ 'Yes', 'No' ],
        type     => 'radio',
        required => 1,
    }
);

my %other_info_field_defs = (
    resource_library => {
        type     => 'radio',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    resource_library_description => {
        type => 'textarea',
        rows => 3
    },
    meeting_space => {
        type     => 'radio',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    meeting_space_description => {
        type => 'textarea',
        rows => 3
    },
    staff_internet_access => {
        type     => 'radio',
        options  => [ 'Limited', 'Yes', 'No', 'Not Applicable' ],
        required => 1
    }
);

# Get the field definitions as a hash ref, keyed by field name.
sub get_agency_field_definitions_href {
    return { %agency_field_defs, %contact_field_defs,
        %other_info_field_defs };
}

# ALTERNATE IMPLEMENTATION BELOW. Instead of the caller deleting fields it doesn't want,
# we define a hash of discretionary fields that have to specifically be requested.
# This makes the caller's logic more straightforward, since it's more natural to say, "include
# these fields under circumstance x" rather than "exclude these fields in all but circumstance
# x".

# my %discretionary_field_defs = { password => {...}, password_verification => {...}, date_updating_option => {...}};

# sub get_field_definitions_href_alternative {
#
#    my %args = @_;
#    my $discretionary_fieldnames_aref = $args{discretionary_fieldnames_aref};
#    my %fields = (%agency_field_defs, %contact_field_defs, %other_info_field_defs);
#    if (defined $discretionary_fieldnames_aref) {
#        foreach (@$discretionary_fieldnames_aref) {
#            $fields{$_} = $discretionary_field_defs{$_};
#        }
#    }
#    return \%fields;
#}

1;
