package MVHub::ArgChecker;

=head1 NAME

ArgChecker - Validate named arguments passed to your subroutine

=head1 SYNOPSIS

use MVHUb::ArgChecker;

update_database(address => '123 Main', email => 'jdoe123@thecsl.org', name => 'John Doe');
...
sub update_database {
    my ($name, $email, $address) = MVHub::ArgChecker::get_args(
       ['name', '?email', 'address'], @_  );
    ...    
}


=head1 DESCRIPTION

ArgChecker lets you specify a set of named arguments for your 
subroutine, which ones are optional.

There are many modules on CPAN that already do this plus much more (e.g. type
validation). However, they are almost all more complicated. I wanted something
that was very simple to use, so that it would be used. You might find it
ironic that get_args does not itself take named arguments. Again, the reason
is to make its signature as compact as possible to promote its usage.

There may be times where you do want something more full featured;
take a look at Params::Validate.

=head2 sub get_args

...takes the following arguments:

     - first argument: an array reference of argument names. Any name with a 
leading or tailing ? means the argument is optional.

     - second argument: the @_ array passed into your subroutine.
    
...dies with a helpful error message and stack dump if the caller:
     - fails to pass in a required argument
     - passes in an argument name you didn't specify

.. and otherwise:
     - returns the values of the arguments in the order you specified them.

=cut

require Exporter;
our @ISA       = qw/Exporter/;
our @EXPORT_OK = qw/get_args/;

use Carp;

sub get_args {

    my ( $arg_names_aref, %args ) = @_;
    my @return_values;

    # Potential error of the parent caller:
    assert_this(
        $arg_names_aref && ref($arg_names_aref) eq 'ARRAY',
        "First argument to get_args must be a reference to an "
            . "array of argument names"
    );

    foreach my $arg_name (@$arg_names_aref) {
        my $is_optional = ( $arg_name =~ s/\?// );
        my $arg_value = delete $args{$arg_name};
        if ( not $is_optional ) {
            assert_this( defined $arg_value,
                "$arg_name is a required argument" );    #grandparent error
        }
        push @return_values, $arg_value;
    }
    my @excess_arg_names = keys %args;

    assert_this( @excess_arg_names == 0,
        "Unknown argument(s): @excess_arg_names" );      #grandparent error
    return @return_values;
}

sub assert_this ($;$$) {
    my ( $expression, $error_msg ) = @_;
    return if ($expression);
    my ( $package, $filename, $line ) = caller(1);

    my $msg = "\n\nFailed at line: '$line' of filename: '$filename'";
    $msg .= ": $error_msg";
    die("$msg\n");
}

1;
