package MVHub::AuthAccount;
require Exporter;

use strict;
use warnings;

use MVHub::Utils;
use MVHub::Utils::ConfigSimple;
use Params::Validate;
use Carp;

our ($VERSION) = '$Revision: 1146 $' =~ /([.\d]+)/;

# TODO: Look into pre-written authentication/authorization
# framework rather than rolling our own.

my $CFG
    = MVHub::Utils::ConfigSimple::create_config_from( $ENV{MV_CONFIG_FILE} );
my $TMP_DIR          = $CFG->param('ABSOLUTE_PATH.tmp_dir');
my $AUTH_COOKIE_NAME = $CFG->param('COOKIES.auth_cookie_name');
my $ADMIN_ID         = $CFG->param('CONSTANTS.admin_id');
my $SITE_CODE        = $CFG->param('SITE.website_code');
our @EXPORT_OK = qw(
    authenticate_session
    _die_loudly_if_dir_not_writable
    init_session
);

sub authenticate_session {
    my $cgi = shift;

    my $agency_id = $cgi->param('agency_id');
    MVHub::Utils::assert( $agency_id,
        "\$agency_id is undefined, empty or zero" );

    my $is_admin  = 0;
    my $is_agency = 0;
    my $error_message;

    # get session id passed from client (cookie), if any
    my $session_id = $cgi->cookie($AUTH_COOKIE_NAME);

    my $caller_id;

    if ( defined $session_id && $session_id ne '' ) {

        # Someone is logged in.  Figure out who.
        my $session = _produce_session($session_id);
        if ( !$session ) {
            _show_login_and_exit(
                      "For your security, your session has timed out. "
                    . " Please log in again." );
        }

        # we have a valid session object
        $caller_id = $session->param('caller_id');
    }

   # People are who they say they are
   # The first if conditions are numeric regexes to avoid throwing errors with
   # text usernames.
    if (   defined $caller_id
        && $caller_id =~ /^-?\d+$/
        && $caller_id == $ADMIN_ID )
    {
        $is_admin = 1;
    }
    elsif (defined $caller_id
        && $caller_id =~ /^\d{6,7}$/
        && $caller_id == $agency_id )
    {
        $is_agency = 1;
    }

    # Prevent logged-in agencies from masquerading as other agencies
    if (   defined $caller_id
        && $caller_id =~ /^\d{6,7}$/
        && $caller_id != $agency_id )
    {
        $error_message = " caller_id != agency_id";
    }

    return ( $is_admin, $is_agency, $error_message );
}

sub init_session {
    my %args      = @_;
    my $caller_id = $args{caller_id};
    croak("caller id undefined/empty/0") if !$caller_id;
    my $session = _produce_session();
    $session->param( 'caller_id', $caller_id );

    my $cookie = CGI::cookie(
        -name  => $AUTH_COOKIE_NAME,
        -value => $session->id(),
        -path  => '/cgi-bin/mvhub'
    );
    return $cookie;
}

sub is_correct_password {
    my $dbh       = shift;
    my $agency_id = shift;
    my $password  = shift;

    my $sql
        = MVHub::Utils::DB::get_sql_select_statement('AGENCY_X_AUTH_AGENCY');

    my $rows_aref
        = $dbh->selectall_arrayref( $sql, undef, $agency_id, lc($password) );
    return ( scalar @$rows_aref == 1 );
}

sub logout {
    my $cgi        = new CGI;
    my $session_id = $cgi->cookie($AUTH_COOKIE_NAME);
    my $cookie;

    # If they're actually logged in and have a server-side session, expire it
    if ($session_id) {
        _delete_session($session_id);
        $cookie = CGI::cookie(
            -name    => $AUTH_COOKIE_NAME,
            -value   => 'EXPIRED',
            -path    => '/cgi-bin/mvhub',
            -expires => '-1d'
        );
    }
    my $body = <<BODY;
<p>You have been logged out.<a href='/'>Home</a></p>
BODY

    MVHub::Page::print_page_and_exit(
        cookie    => $cookie,
        body      => $body,
        site_code => $SITE_CODE
    );

    # TODO: separate the above body into a file

}

####### Private Subroutines, used by only this module: ################
sub _delete_session {
    my $session_id = shift;
    my $session    = _produce_session($session_id);
    if ($session) {
        $session->delete();
    }
}

sub _die_loudly_if_dir_not_writable {
    Params::Validate::validate_pos( @_, 1 );
    my $dir = shift;

    my $writeable = -d $dir && -W $dir;
    use English '-no_match_vars';
    my $process_ownername = getpwuid($EFFECTIVE_USER_ID);

    if ( !$writeable ) {

        # extra carp  to avoid msg being truncated
        carp "not dir or writable: $dir \n";
        carp "apache runs as $process_ownername\n";
        croak "$process_ownername can't write to: $dir\n";
    }
}

# Returns a session object as follows:
# If you pass in no session id, returns a new session object.
# If you pass in a valid session id, returns its corresponding session object.
# If you pass in an invalid session id (e.g. an expired one), returns undef.
sub _produce_session {
    my $session_id = shift;

    _die_loudly_if_dir_not_writable($TMP_DIR);

    my $session = new CGI::Session( "driver:File", $session_id,
        { Directory => $TMP_DIR } );

    # It would appear that the only way to know that a session id was
    # invalid is to create a session object using it (as we did
    # above). If it was invalid, the created session object will be
    # new, and have a different session id.
    if ( defined $session_id and $session->id() ne $session_id ) {
        $session->delete();
        return;
    }
    else {
        return $session;
    }

}

# TODO: Deprecated, remove
# If an error message is provided, display the log-in page
# with the message.
# Otherwise:
#   if the caller is already logged in, show the agency home page.
#   Otherwise:
#       if proper log-in credential are provided, logs-in the user and
#          shows the agency_home page
#       if improper log-in credentials are provided, displays the log-in page
#          with an error message.
#       if no log-in credentials are provided, displays the log-in page.
# Always exits
sub _show_login_and_exit {
    my $error_message = shift;

    #    die "show_login_and_exit should not be getting called anymore.";
    my $cgi = new CGI("");

    # This returns the "http:" "//" host [ ":" port ]" section of the URL
    my $base_url = $cgi->url( -base => 1 );
    my $absolute_path = '/cgi-bin/mvhub/agency.pl';

    my $cgi_params_href = {
        rm         => 'show_login_page',
        error_code => $error_message
    };

    use URI::URL;
    my $url = URI::URL->new_abs( $absolute_path, $base_url );
    $url->query_form($cgi_params_href);
    print $cgi->header( -location => $url->as_string );

    exit(0);

}

# all good modules return 1
return 1;
__END__

=head1 NAME

MVHub::AuthAccount authentification, authorization, session and account information subs
 
=head1 VERSION

This documentation refers to MVHub::AuthAccount version $Revision$
 
=head1 SYNOPSIS
    FIXME
    use <Module::Name>;
    # Brief but working code example(s) here showing the most common usage(s)
 
    # This section will be as far as many users bother reading
    # so make it as educational and exemplary as possible.
   
=head1 DESCRIPTION
 
A full description of the module and its features.
May include numerous subsections (i.e. =head2, =head3, etc.) 
 
=head1 SUBROUTINES/METHODS 
 
=head2 authenticate_session 

Determine if a session is for any or all of an agency, admin or data operator.

Authorization (username/password login stuff) must be done before this sub. 

  Parameters: $cgi, a CGI query 

  Return: a list of three booleans $is_admin, $is_agency, and $is_data_entry_operator,

=head2 init_session 

Create a CGI::Session session

Parameters: $caller_id / $agency_id (can be $ADMIN_ID, special value for admin user)

Returns:  $cookie, a CGI::Cookie, containing a session id for 
          that agency or for ADMIN user

=head1 DIAGNOSTICS
 
A list of every error and warning message that the module can generate
(even the ones that will "never happen"), with a full explanation of each 
problem, one or more likely causes, and any suggested remedies.
(See also  QUOTE \" " INCLUDETEXT "13_ErrorHandling" "XREF83683_Documenting_Errors_"\! Documenting Errors QUOTE \" " QUOTE " in Chapter "  in Chapter  INCLUDETEXT "13_ErrorHandling" "XREF40477__"\! 13.)
 
 
=head1 CONFIGURATION AND ENVIRONMENT
 
A full explanation of any configuration system(s) used by the module,
including the names and locations of any configuration files, and the
meaning of any environment variables or properties that can be set. These
descriptions must also include details of any configuration language used.
(also see  QUOTE \" " INCLUDETEXT "19_Miscellanea" "XREF40334_Configuration_Files_"\! Configuration Files QUOTE \" " QUOTE " in Chapter "  in Chapter  INCLUDETEXT "19_Miscellanea" "XREF55683__"\! 19.)
 
 
=head1 DEPENDENCIES
 
A list of all the other modules that this module relies upon, including any
restrictions on versions, and an indication whether these required modules are
part of the standard Perl distribution, part of the module's distribution,
or must be installed separately.
 
=head1 INCOMPATIBILITIES
 
There are no known incompatibilites
 
=head1 BUGS AND LIMITATIONS
 
Funtionalty overlaps with MVHub::AgencyAccount

Please report problems to help@thecsl.org
Patches are welcome.
 
=head1 AUTHOR

Community Software Lab help@thecsl.org 
 
=head1 LICENCE AND COPYRIGHT

Copyright (c) 2004-2008 Community Software Lab, Inc ( http://thecsl.org )

The GNU Affero General Public License, version 3.0 or later

=cut
