our ($VERSION) = '$Revision: 1730 $' =~ /([.\d]+)/;

package MVHub::CGIAppBase;

use strict;
use warnings;

use base 'CGI::Application';

use Carp;
use CGI::Application::Plugin::DBH qw/dbh/;
use DBI;
use MVHub::Utils::ConfigSimple;
use MVHub::Utils::DB;
use SQL::Library;

sub cgiapp_init {
    my $self = shift;
    MVHub::Utils::die_if_missing_env_vars();

    $self->init_config( $ENV{MV_CONFIG_FILE} );

    my $dbh = MVHub::Utils::DB::get_dbh( $ENV{MV_CONFIG_FILE} );
    $self->CGI::Application::Plugin::DBH::dbh_config($dbh);

    my $sql_lib_file = $self->get_config_param('ABSOLUTE_PATH.conf_dir')
        . "/sql_select.lib";
    $self->{_sql_lib} = new SQL::Library { lib => $sql_lib_file };
}

sub retr_select_sql {
    my $self           = shift;
    my $statement_name = shift
        or die "missing param: statement_name";
    my $sql_lib = $self->{_sql_lib};
    return $sql_lib->retr($statement_name);
}

sub get_config_param {
    my $self = shift;
    my $config_param = shift or croak "Missing Parameter: config_param";

    my $result = $self->{cfg}->param($config_param)
        or croak "Couldn't find $config_param in file:"
        . $self->{_config_filename};
    return $result;
}

sub init_config {
    my $self = shift;
    my $config_file = shift or croak "Missing Parameter: config_file";
    $self->{cfg}
        = MVHub::Utils::ConfigSimple::create_config_from($config_file);
    $self->{_config_filename} = $config_file;
    return;
}

# end all modules with 1
1;
__END__


=head1 NAME

CGIAppBase - Base class for our CGI::Application classes, contains initialization common to all our classes 

=head1 SYNOPSIS

    use base CGIAppBase;
    
    sub a_run_method {
      my $self = shift;
    
      my $dbh = $self->dbh();
      my $sql = $self->retr_select_sql('AGENCY_X_ALL');
      my $tmp_dir =$self->get_config_param('ABSOULTE_PATH.tmp_dir');
      my $stuff_aref = $dbh->selectcol_arrayref($sql);
      #...                  
   }
   
=head1 DESCRIPTION

 MVHub::CGIAppBase is the class all our other classes should inherit 
 from. It contains methods for getting the object's database handle, 
 the variables contained in the configuration file and the SQL 
 statements stored in app-mvhub/conf/sql_select.lib

=head1 METHODS
  
  dbh()
  
   returns a database handle 


  get_config_param($PARAM_NAME)
  
    Given the name of a paramter, returns the value stored in the 
    application config file. 
    
    The application configuration file
    is currently either /var/www/mvhub/$USER/conf/mvh.conf or 
    /var/www/mvhub/$USER/conf/nsp.conf (though you should not
    depend on these files remaining in these places. 
        
  retr_select_sql($SQL_STATEMENT_NAME)

    Given the name of a sql SELECT statement, returns the SQL for that
    statement. Currently the file the SQL statements are stored in
    is: app-mvhub/conf/sql_select.lib, though you should not depend
    on this location. ---And you should not need to depend on a hard
    coded path as this is handled internally for you by CGIAppBase. 

=head1 SEE ALSO

 #  Particularly, cgiapp_init
 man CGI::Application 

 man CGI::Application::Plugin::DBD
 perldoc MVHub::Utils::ConfigSimple
 man SQL::Library
 
=head1 LAST MODIFIED

Last modified on $Date: 2010-06-02 22:11:05 $ by $Author: omacneil $

=head1 BUGS

CGIAppBase is used by AgencyAccount.pm and Guide.pm but not other 
classes. Much of the code, agency_form.pl and program.pl in particular 
don't use CGI::Application as they are entirely procedural.

=head1 LICENSE

The GNU Affero General Public License, version 3.0 or later

=head1 COPYRIGHT

Community Software Lab, Inc ( http://thecsl.org )

=cut


