
package MVHub::Guide;
use strict;
use warnings;

use Carp;

use List::Util;
use URI;
use URI::Escape;
use URI::URL;

use MVHub::Utils 'assert';
use base 'MVHub::CGIAppBase';
use MVHub::GuideSearch;
use MVHub::HTML;
use MVHub::Page;

my $ZIP_CODE_TAG           = 'zip_code';
my $CR_CATEGORIES_LOOP_TAG = 'category_results';
my $CR_CATEGORY_COUNT_TAG  = 'category_count';

my $PR_CRITERIA_TAG      = 'criteria';
my $PR_PROGRAM_COUNT_TAG = 'program_count';
my $PR_PROGRAMS_LOOP_TAG = 'program_results';

my $SA_PROGRAMS_LOOP_TAG = 'programs_loop';
my $AR_AGENCY_COUNT_TAG  = 'agency_count';
my $AR_AGENCIES_LOOP_TAG = 'agency_results';

my @AGENCY_RELATED_RUN_MODES = qw[show_agency_results show_agency];

sub setup {
    my $self = shift;
    $self->run_modes(
        'browse_headings'      => 'browse_headings',
        'browse_one_heading'   => 'browse_one_heading',
        'redirect'             => 'redirect',
        'show_search_results'  => 'show_search_results',
        'show_program_results' => 'show_program_results',
        'show_program'         => 'show_program',
        'show_agency_results'  => 'show_agency_results',
        'show_agency'          => 'show_agency',

# If the CGI's rm param resolves to something other than any of the above, call
# this method by default (see method for more explanation):
        'AUTOLOAD' => 'catch_bad_run_mode'

    );
    $self->start_mode('show_search_results');
    $self->mode_param('rm');
}

sub cgiapp_postrun {
    my $self          = shift;
    my $output_ref    = shift;
    my $search_phrase = $self->query()->param('search_phrase');

    $$output_ref = MVHub::Page::prepare_page(
        body          => $$output_ref,
        title         => $self->param('title'),
        site_code     => $self->get_config_param('SITE.website_code'),
        site_location => $self->get_config_param('SITE.location'),
        search_box_help_text =>
            $self->get_config_param('SITE.search_box_help_text'),
        show_left_menu  => 1,
        show_search_bar => 1,
        search_phrase   => $search_phrase,
        agency_search   => scalar grep( $_ eq $self->get_current_runmode(),
            @AGENCY_RELATED_RUN_MODES )
    );
}

# Run Modes

sub browse_headings {
    my $self = shift;

    my $dbh = $self->dbh();
    my $sql = $self->retr_select_sql(
        'PROGRAM_CATEGORY_HEADING_CATEGORY_X_HEADING_NAME');

    my $headings_aref = $dbh->selectcol_arrayref($sql);
    my $html          = << "HTML";
<!-- start generated by MVHub::Guide::browse_headings() --> 
HTML

    # Generate an html table of links for all the headings.
    my @links;
    my $url = "/cgi-bin/mvhub/guide.pl?rm=browse_one_heading&heading_name=";
    foreach (@$headings_aref) {

  # fully URL-escape the heading name, then partially URL-escape the whole url
  # (Why 2 steps? So ampersands in the heading name get escaped,
  #  while ampersands as parameter delimiters don't)
        my $link
            = new URI::URL( $url . URI::Escape::uri_escape($_) )->as_string();
        my $label = CGI::escapeHTML($_);
        push( @links, CGI::a( { href => $link }, $label ) );
    }
    $html .= MVHub::HTML::make_html_table(
        columns          => 3,
        content_aref     => \@links,
        table_attributes => 'cellspacing="5"'
    );
    $html .= "<!-- stop  generated by MVHub::Guide::browse_headings() -->\n";

    return $html;
}

sub browse_one_heading {
    my $self         = shift;
    my $cgi          = $self->query();
    my $heading_name = $cgi->param('heading_name');

    # Damn query-string truncating bots are triggering assertions:
    #  assert($heading_name, 'Missing heading name');
    # Just finish immediately instead:
    if ( !$heading_name ) {
        exit(1);
    }

    my $dbh = $self->dbh();

    my $sql = $self->retr_select_sql(
        "CATEGORY_PROGRAM_VIEW_HEADING_CATEGORY_X_ID_NAME_COUNT");
    my $categories_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, $heading_name );

    # 0 rows returned might or might not be a bug. So we don't test for it.

    my $html
        = "<!-- start generated by MVHub::Guide::browse_one_heading() -->";
    $html
        .= "<h1>$heading_name: Program Categories</h1>\n<p><strong>Click on a category to see related programs.</strong> (The number of related programs is in parenthesis.)<p>\n";

    # Generate an html table of category links
    my @links;
    my $url = "/cgi-bin/mvhub/guide.pl?rm=show_program_results&category_id=";
    foreach (@$categories_aref) {
        my $link  = $url . $_->{category_id};
        my $label = $_->{category_name} . " ($_->{program_count})";
        $label = CGI::escapeHTML($label);

        # add space around forward slashes so a label like
        # "Translating/Interpreting" will wrap nicely
        $label =~ s|\s*/\s*| / |g;
        push( @links, CGI::a( { href => $link }, $label ) );
    }
    $html .= MVHub::HTML::make_html_table(
        columns          => 4,
        content_aref     => \@links,
        table_attributes => 'cellspacing="5"'
    );

 # Before returning our page body, let's add some custom content to its title:
    $self->param( 'title', $heading_name );

    $html
        .= "<!-- stop generated by MVHub::Guide::browse_one_heading() -->\n";
    return $html;
}

# Redirects to an agency's website.  We're doing this so that we can
# log when users click through to an agency website.  If both an
# # agency and a program id are specified, redirect to the program's
# website since it's potentially the more pertinent site.
sub redirect {
    my $self = shift;
    my $cgi  = $self->query();

    my $agency_id  = $cgi->param('agency_id');
    my $program_id = $cgi->param('program_id');
    assert(
        ( defined $agency_id || defined $program_id ),
        "No program or agency id specified for redirect."
    );

    my $error_template
        = $self->load_tmpl( 'error_page.tmpl', global_vars => 1 );

    my $sql;
    my $dbh = $self->dbh();
    my $url_aref;

    # return an error unless one of these is defined.
    # use program_id if both are defined.
    if ( defined $program_id ) {
        $sql = $self->retr_select_sql("PROGRAM_X_WEBSITE");
        $url_aref = $dbh->selectrow_arrayref( $sql, undef, $program_id );
    }
    elsif ( defined $agency_id ) {
        $sql = $self->retr_select_sql("AGENCY_X_WEBSITE");
        $url_aref = $dbh->selectrow_arrayref( $sql, undef, $agency_id );
    }
    else {
        $self->header_type('header');
        $self->header_props( -status => '404 Not Found' );
        my $error_message = "<h2>404 Not Found</h2>\n";

        return $error_template->output()
            ;   # So that the error message is incorporated into the template.
    }

    # If $url_aref isn't defined here, this means that
    # someone specified a non-existent agency or program id.
    # We should return a 404 error because this resource doesn't exist.
    unless ( defined $url_aref ) {
        $self->header_type('header');
        $self->header_props( -status => '404 Not Found' );
        return $error_template->output();
    }

    # Create URI object and check for properly formed URLs:

    my $url = URI->new( $url_aref->[0] );

# Only accept URLs of the form http://www.thecsl.org or www.thecsl.org.
# Technically, this allows URLs of the form http:////////www.thecsl.org (obviously bogus),
# so I'll need to chop off the leading slashes.
    assert(
        ( !defined $url->scheme() || $url->scheme eq 'http' ),
        "Malformed URL--possibly an ftp: or mailto: url."
    );

    # Adds http:// to the front of the URL if it's not already there.
    $url->scheme('http') unless defined $url->scheme();

    # Strips off non-word characters at the start of the path
    # Otherwise, stuff like http://++www.thecsl.org would work....
    my $url_path = $url->opaque();

 # Get rid of non-word and escaped characters at the beginning of the url path
    $url_path =~ s/^(\%\d{2,3}|\W)+//;
    $url->opaque($url_path);

    # The CGI::Application way to do redirects.
    my $redirect_url = $url->canonical();
    $redirect_url =~ s/:/:\/\//
        ;    # since the slashes are stripped away by $url->canonical
    $self->header_type('redirect');
    $self->header_props( -url => $redirect_url );
    return "Redirecting to $redirect_url\n";
}

sub show_agency_results {
    my $self          = shift;
    my $agencies_aref = [];
    my $dbh           = $self->dbh();
    my $search_phrase = $self->query()->param('search_phrase');

    if ( defined $search_phrase && $search_phrase !~ /^\s*$/ ) {
        $agencies_aref
            = MVHub::GuideSearch::agency_search( $dbh, $search_phrase );
    }

    my $tmpl = $self->load_tmpl(
        'agency_results.tmpl',

        # set global vars so that the category result
        # links can all get the search phrase, which we
        # set as a tmpl var (down a few lines)
        global_vars => 1
    ) or die("Unable to open 'agency_results.tmpl");

    $tmpl->param( $AR_AGENCIES_LOOP_TAG => $agencies_aref );
    $tmpl->param( $AR_AGENCY_COUNT_TAG  => scalar @$agencies_aref );

    # This will put the search phrase in the agency links, for populating
    # the search box in the search bar of the ensuing page:
    $tmpl->param( $MVHub::Page::SEARCH_PHRASE_TAG, $search_phrase );

    return $tmpl->output();
}

sub show_search_results {
    my $self = shift;
    my $dbh;
    $dbh = $self->dbh();
    my $search_phrase = $self->query()->param('search_phrase');

    my $agencies_aref   = [];
    my $categories_aref = [];
    if ( defined $search_phrase && $search_phrase !~ /^\s*$/ ) {
        $agencies_aref
            = MVHub::GuideSearch::agency_search( $dbh, $search_phrase );
        $categories_aref
            = MVHub::GuideSearch::category_search( $dbh, $search_phrase );
    }

    my $tmpl = $self->load_tmpl(
        'search_results.tmpl',
        global_vars       => 1,
        die_on_bad_params => 0,
    ) or die("Unable to open HTML::Template file: 'search_results.tmpl'");

    my $agency_names_urls_aref
        = _get_agency_names_urls( $agencies_aref, $search_phrase );

    $tmpl->param( $AR_AGENCIES_LOOP_TAG   => $agency_names_urls_aref );
    $tmpl->param( $AR_AGENCY_COUNT_TAG    => scalar @$agencies_aref );
    $tmpl->param( $CR_CATEGORY_COUNT_TAG  => scalar @$categories_aref );
    $tmpl->param( $CR_CATEGORIES_LOOP_TAG => $categories_aref );

    # This will put the search phrase in the category links, for populating
    # the search box in the search bar of the ensuing page:
    $tmpl->param( $MVHub::Page::SEARCH_PHRASE_TAG, $search_phrase );

    return $tmpl->output();
}

sub _get_agency_names_urls {
    my $agency_aref = shift or croak 'missing parameter: agency_aref';
    my $search_phrase = shift || q//;    # optional
    my @agency_names_urls;

    foreach my $agency (@$agency_aref) {
        my $url = "/cgi-bin/mvhub/guide.pl?rm=show_agency";
        $url .= "&amp;agency_id=$$agency{agency_id}";
        $url .= $search_phrase ? "&amp;search_phrase=$search_phrase" : q//;

        my $name = $$agency{agency_name};
        $name .= $$agency{agency_alias} ? "($$agency{agency_alias})" : q//;

        push @agency_names_urls, { agency_name => $name, agency_url => $url };
    }
    return \@agency_names_urls

}

# TODO: deal with stale programs
# TODO: add "Serving: " to output
sub show_program_results {
    my $self          = shift;
    my $category_id   = $self->query()->param('category_id');
    my $search_phrase = $self->query()->param('search_phrase');

    assert(
        $category_id,
        'Missing category_id parameter from HTTP request. This could be a bug in the 
code that formed the referring link. But more than likely it is the result of an automated user agent 
(bot) truncating the URL it is crawling. The HTTP_FROM and/or HTTP_USER_AGENT values below should 
indicate whether the source is a bot.'
    );

    my $dbh = $self->dbh();
    my $sql = $self->retr_select_sql(
        "PROGRAM_AGENCY_PROGRAM_CATEGORY_X_AGENCY_NAME_ID_PROGRAM_NAME");
    my $programs_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, $category_id );
    $programs_aref = _add_program_urls_to($programs_aref);

    # We should never have been called if there aren't any programs for this
    # category. (There is the chance of a problem if a user does a search,
    # then an admin removes the only program in this category, then the user
    # clicks on the category name. A very small chance.)
    assert(
        ( @$programs_aref > 0 ),
        "Got no programs for category "
            . "$category_id; expected at least one."
    );

    my $PROGRAM_RESULTS_TEMPLATE = 'program_results.tmpl';
    my $tmpl                     = $self->load_tmpl(
        $PROGRAM_RESULTS_TEMPLATE,

        # Don't die on bad params--just ignore them.  This is how HTML
        # behaves, and is essential for template customization.
        die_on_bad_params => 0,

        # set global vars so that the category result
        # links can all get the search phrase, which we
        # set as a tmpl var (down a few lines)
        global_vars => 1

    ) or die("Unable to open $PROGRAM_RESULTS_TEMPLATE");

    $tmpl->param( $PR_PROGRAM_COUNT_TAG => scalar @$programs_aref );
    $tmpl->param( $PR_PROGRAMS_LOOP_TAG => $programs_aref );

    # Get the category name to show off in the result page.

    my $sql2 = $self->retr_select_sql("CATEGORY_X_CATEGORY_NAME");
    my $category_name
        = @{ $dbh->selectrow_arrayref( $sql2, undef, $category_id ) }[0];

    $tmpl->param( $PR_PROGRAMS_LOOP_TAG => $programs_aref );
    $tmpl->param( $PR_CRITERIA_TAG      => $category_name );

    # This will put the search phrase in the program links, for populating
    # the search box in the search bar of the ensuing page:
    $tmpl->param( $MVHub::Page::SEARCH_PHRASE_TAG, $search_phrase );

 # Before returning our page body, let's add some custom content to its title:
    $self->param( 'title', $category_name );

    # needed by search_bar.tmpl
    $self->param( 'site_location', $self->get_config_param('SITE.location') );

    $self->param( 'search_box_help_text',
        $self->get_config_param('SITE.search_box_help_text') );

    return $tmpl->output();

}

sub _add_program_urls_to {
    my $programs_aref = shift or croak 'missing parameter: $programs_ahref';
    my $search_phrase = shift || q//;    # optional
    my @programs = @$programs_aref;
    foreach my $program (@programs) {
        my $url = "/cgi-bin/mvhub/guide.pl?rm=show_program";
        $url .= "&amp;program_id=$$program{program_id}";
        $url .= $search_phrase ? "&amp;search_phrase=$search_phrase" : q//;

        $$program{program_url} = $url;
    }
    return \@programs;
}

sub show_agency {
    my $self = shift;
    my $cgi  = $self->query();
    my $dbh  = $self->dbh();

    my $agency_id = $cgi->param('agency_id');

    my $agency_href = MVHub::Utils::DB::select_agency_record(
        'agency_id'          => $agency_id,
        'allow_empty_result' => 1,
        'dbh'                => $dbh
    );

    exit_if_entity_count_not_one( 'agency', defined $agency_href );

    my $SHOW_AGENCY_TEMPLATE = 'show_agency.tmpl';
    my $tmpl
        = $self->load_tmpl( $SHOW_AGENCY_TEMPLATE, die_on_bad_params => 0 )
        or die "Unable to open template: $SHOW_AGENCY_TEMPLATE";

    # Loop through the keys and add them as parameters to the template object
    # Also the place to parse/modify values
    foreach ( keys %$agency_href ) {
        next unless $agency_href->{$_};

        # We want any newlines (in textareas) to display as intended:
        $agency_href->{$_} =~ s/\n/\n<br>/g;
        if (/website/) {

# The url for the website link.  The 'website' tmpl_var will be used as the link's title (in template).
            $tmpl->param( 'website_url',
                "/cgi-bin/mvhub/guide.pl?rm=redirect&agency_id=$agency_id" );
        }

        # TODO: Any more special parsing of specific fields

        $tmpl->param( $_, $agency_href->{$_} );
    }

    my @bind_variables = ($agency_id);
    my $sql            = MVHub::Utils::DB::get_sql_select_statement(
        'PROGRAM_X_AGENCY_ID_HIDDEN_REASON_IS_NULL');

    my $programs_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, @bind_variables );

    foreach my $record_href (@$programs_aref) {
        $record_href->{'last_updated'}
            = MVHub::Utils::yyyy_mm_dd_to_mm_dd_yyyy(
            $record_href->{'last_updated'} );
    }

    if ( scalar @$programs_aref ) {
        $tmpl->param( $SA_PROGRAMS_LOOP_TAG, $programs_aref );
    }

    return $tmpl->output();
}

sub show_program {
    my $self          = shift;
    my $search_phrase = $self->query()->param('search_phrase');
    my $program_id    = $self->query()->param('program_id');

    my $dbh = $self->dbh();
    my $sql
        = $self->retr_select_sql("AGENCY_PROGRAM_X_AGENCY_NAME_AGENCY_ALIAS");
    my $programs_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, $program_id );

    exit_if_entity_count_not_one( 'program', scalar @$programs_aref );

    my $SHOW_PROGRAM_TEMPLATE = 'show_program.tmpl';
    my $program_href          = @$programs_aref[0];
    my $tmpl
        = $self->load_tmpl( $SHOW_PROGRAM_TEMPLATE, die_on_bad_params => 0 )
        or die "Unable to open template: $SHOW_PROGRAM_TEMPLATE";

    # Loop through the keys and add them as parameters to the template object
    # Also the place to parse/modify values
    foreach ( keys %$program_href ) {
        next unless $program_href->{$_};

        # We want any newlines (in textareas) to display as intended:
        $program_href->{$_} =~ s/\n/\n<br>/g;
        if (/languages|area_data/) {
            $program_href->{$_} =~ s/$MVHub::Common::FIELD_DELIMITER/, /g;
        }
        if (/bus_lines/) {
            $program_href->{$_} =~ s/$MVHub::Common::FIELD_DELIMITER/<BR>/g;
        }
        if (/website/) {
            $tmpl->param( 'website_url',
                "/cgi-bin/mvhub/guide.pl?rm=redirect&program_id=$program_id"
            );
        }
        if (/last_updated/) {
            $program_href->{$_} = MVHub::Utils::yyyy_mm_dd_to_mm_dd_yyyy(
                $program_href->{$_} );
        }

        # TODO: Any more special parsing of specific categories

        $tmpl->param( $_, $program_href->{$_} );
    }

    return $tmpl->output();
}

# Given a program id, category id or agency id, etc. in the HTTP request,
# we would expect the database lookup we perform with it to produce exactly 1 record.
# We used to just assert this, thinking that any other result was a development
# error. However,a if anyone saves a link to a program/agency/etc. that is later removed,
# using the link will trigger this error. MSN's bots were doing this all the time (and
# we were getting a lot of annoying automatic notification emails). So now
# we assert only if the referer is internal; otherwise we give a useful
# error message to the user.
# TODO: possibly refactor this into an error handling module somewhere
sub exit_if_entity_count_not_one {
    my $entity = shift;
    my $count = shift || 0;
    return if $count == 1;
    if ( $ENV{HTTP_REFERER} && $ENV{HTTP_REFERER} =~ /$ENV{HTTP_HOST}/ ) {
        assert( 0, "Expected 1 $entity, got $count" );
    }
    else {
        my $message
            = "The $entity you were trying to look up unfortunately no longer exists.";
        MVHub::Common::display_error_page_and_exit($message);

        # TODO: Q: by doing this instead of returning some sort of 400+ error
        # code, will MSN ever learn that it should stop caching this URL?
    }
}

# We keep getting bots abusing our script by passing bogus email addresses in the rm
# param, causing CGI::Application to die, resulting in my getting a lot of automated
# alert emails. Intercept them here:
sub catch_bad_run_mode {
    my $self             = shift;
    my $intended_runmode = shift;
    my $message
        = "The web address of the page you requested was somewhat mangled, which "
        . "in our experience, leads us to believe you're a SPAM bot or other mischevious program. "
        . "If you're actually a well-intentioned, living, pulsing, breathing person, we apologize, "
        . "and could use your help to "
        . "correct our misjudgement. Would you please make a copy of the web address in your "
        . "browser's address bar above, and send it to us with a brief explanation, using the "
        . "contact link below? Thanks!<br><br>"
        . "On the other hand, if you are a bot... GOTCHA!";
    MVHub::Common::display_error_page_and_exit($message);
}

##
#  Non-run-mode subroutines
##

1;

