
package MVHub::GuideSearch;

our ($VERSION) = '$Revision: 1552 $' =~ /([.\d]+)/;

use strict;
use warnings;

use DBI;
use MVHub::Utils;
use MVHub::Utils::DB;
use MVHub::Common;

sub agency_search {
    my $dbh    = shift;
    my $phrase = shift;
    MVHub::Utils::assert( defined $phrase && $phrase !~ /^\s*$/,
        'missing phrase' );

    # Split into a list of words.
    my @phrase_words = split( /\/|\s+/, $phrase );

    # Remove bad characters and filler words.
    MVHub::Common::remove_filler( \@phrase_words );

    # Search was pure filler words/bad chars
    if ( !@phrase_words ) { return []; }

    my $matches_aref = get_matching_agencies( $dbh, \@phrase_words );

    log_agency_search_results( $dbh, $phrase, $matches_aref );

    return $matches_aref;
}

# Returns all agencies whose name or alias matches ANY words in the
# array referenced by search_terms_aref. Partial word matches are valid
# if the words begin the same
sub get_matching_agencies {
    my $dbh               = shift;
    my $search_terms_aref = shift;

    # Search like agency_name and agency_alias
    my $regex1 = '[[:<:]](' . join( '|', @$search_terms_aref ) . ')[[:>:]]';
    my $regex2 = '[[:<:]](' . join( '|', @$search_terms_aref ) . ')[[:>:]]';
    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_X_SEARCH_WITH_TERMS');
    my $matches_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, $regex1, $regex2 );

    return $matches_aref;
}

# Given a search phrase, returns an arrayref of hashrefs, representing the
# category hits, e.g.:
# [
#    { category_id => 800000, category_name => 'this is a category name', score => 2}
#     , ...
# ]
# ordered first by score (i.e. relevance), and second alphabetically,
# where all the category_ids are unique
sub category_search {
    my $dbh    = shift;
    my $phrase = shift;
    MVHub::Utils::assert( defined $phrase && $phrase !~ /^\s*$/,
        'missing phrase' );

    # Get a string containing the lowercase stems of the words in the
    # search phrase. get_stems removes bad characters and filler words
    my $phrase_stems_string = MVHub::Common::get_stems($phrase);

    my @phrase_stems = split( /\s+/, $phrase_stems_string );

    # Discontinue if phrase was entirely bad chars &/or filler words
    if ( !@phrase_stems ) { return []; }

    my @expanded_phrase_stems = @phrase_stems;

    # Expand list to include synonym stems of the phrase stems
    expand_synonyms( $dbh, \@expanded_phrase_stems );

    # Get all of the category ids and names where any stem in the
    # expanded phrase matches any names in the all_names_view
    # (i.e. any category or alias names)
    my $matches_aref
        = get_matching_categories( $dbh, \@expanded_phrase_stems );

    ## Uncomment to do benchmarking
    ## use Time::HiRes;
    ## my $t0 = Time::HiRes::gettimeofday();

    log_category_search_results( $dbh, $phrase, \@phrase_stems,
        $matches_aref );

    ## my $t1 = Time::HiRes::gettimeofday();
    ## my $elapsed = $t1-$t0;  #floating point number of seconds
    ## warn "TEST: elapsed time: $elapsed.";

    return $matches_aref;
}

# Make a record of the given program category search and its result, unless:
#   - this user has already performed more or less same search during the same session
#   - the user's IP address is one we want to ignore (e.g. us)
# This will allow us to see what searches people are conducting, and how successful
# those searches were.
# Arguments:
#  $phrase - the exact phrase submitted by the user
#  $phrase_stems_aref - a cleaned-up list of the stems of the words in the phrase
#  $matching_categories_aref - same as what category_search returns (see its comment)
# THIS IS A WORK IN PROGRESS. It'll probably need tweaks.
# Other TODO thoughts:
# If GuideSearch.pm returned a search result object (hash)
#    that included the filtered stems list, then we could rightly move this
#    subroutine into a separate module.
#
# Once you improve scoring, include the name of the highest scoring category
#
# To improve performance (which isn't currently an issue), you could do
# this in a low priority thread.
sub log_category_search_results {

    my ( $dbh, $phrase, $phrase_stems_aref, $matching_categories_aref ) = @_;

    my $user_ip = $ENV{REMOTE_ADDR};

# TODO: check into more sophisticated ways to usually get the clients IP address,
# eg. including looking at HTTP_CLIENT_IP and HTTP_X_FORWARDED_FOR
# Saw a discussion about it at http://ca3.php.net/getenv

 # commented out because dan broke sub
 #    return
 #        if ( MVHub::Common::is_ip_address_excluded_from_logging($user_ip) );

    my $user_session_id = get_user_session_id_or_makeshift_equivalent();

    my $phrase_stems_string = join( ' ', sort @$phrase_stems_aref );
    my $match_count = @$matching_categories_aref;
    my $top_score
        = $match_count ? $matching_categories_aref->[0]->{score} : 0;

    # Scale the score as a function of how many stems matched.
    $top_score /= scalar @$phrase_stems_aref;

    # Find out if we've recently recorded this search for this user.
    # FYI: The following SQL is relatively straightforward...
    #    my $sql = 'SELECT count(*) FROM category_search_log ' .
    #              'WHERE session_id = ? AND phrase_stems = ?';
    # ...but probably not as fast as the following, which will stop as soon
    # it finds one matching row:
    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'CATEGORY_SEARCH_LOG_X_IF_RECORDED_RECENT_CATEGORY_SEARCH_BY_USER');

    my $rows_aref = $dbh->selectcol_arrayref( $sql, undef, $user_session_id,
        $phrase_stems_string );
    return if ( $rows_aref->[0] );

    MVHub::Utils::DB::insert_one_record(
        dbh         => $dbh,
        table       => 'category_search_log',
        values_href => {
            session_id    => $user_session_id,
            ip_address    => $user_ip,
            referer       => $ENV{HTTP_REFERER},
            search_phrase => $phrase,
            phrase_stems  => $phrase_stems_string,
            top_score     => $top_score
        }
    );
}

# Make a record of the given agency search and its result, unless:
#   - this user has already performed more or less same search during the same session
#   - the user's IP address is one we want to ignore (e.g. us)
# This will allow us to see what searches people are conducting, and how successful
# those searches were.
# Arguments:
#  $phrase - the exact phrase submitted by the user
#  $matching_agencies_aref - list of matching agencies
sub log_agency_search_results {
    my $dbh = shift;
    my ( $phrase, $matching_agencies_aref ) = @_;

    my $user_ip = $ENV{REMOTE_ADDR};

# TODO: check into more sophisticated ways to usually get the clients IP address,
# eg. including looking at HTTP_CLIENT_IP and HTTP_X_FORWARDED_FOR
# Saw a discussion about it at http://ca3.php.net/getenv

  # commented out because dan broke the sub
  #    return
  #       if ( MVHub::Common::is_ip_address_excluded_from_logging($user_ip) );

    my $user_session_id = get_user_session_id_or_makeshift_equivalent();

    my $match_count = @$matching_agencies_aref;

    # Find out if we've recently recorded this agency search for this user.
    # FYI: The following SQL is relatively straightforward...
    #    my $sql = 'SELECT count(*) FROM agency_search_log ' .
    #              'WHERE session_id = ? AND phrase_stems = ?';
    # ...but probably not as fast as the following, which will stop as soon
    # it finds one matching row:

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_SEARCH_LOG_X_IF_RECORDED_RECENT_AGENCY_SEARCH_BY_USER');
    my $rows_aref
        = $dbh->selectcol_arrayref( $sql, undef, $user_session_id, $phrase );
    return if ( $rows_aref->[0] );

    MVHub::Utils::DB::insert_one_record(
        dbh         => $dbh,
        table       => 'agency_search_log',
        values_href => {
            session_id    => $user_session_id,
            ip_address    => $user_ip,
            referer       => $ENV{HTTP_REFERER},
            search_phrase => $phrase,
            hit_count     => $match_count
        }
    );
}

# Get the session ID from the cookie that every user of the site gets (assuming
# the webserver has been configured to create/dole out such a cookie.) If not,
# or if the current user has rejected the cookie,
# then this returns a makeshift session id contructed from the user's IP address,
# plus the hour, day, month (almost), and year (sort of).
# You can pass in a CGI object, or let this method create its own.
sub get_user_session_id_or_makeshift_equivalent {
    my $cgi = $_[0] || new CGI;

    #TODO fix hard coded cookie name should be from config file
    my $user_session_tracking_cookie_name = 'apache_usertrack_session_id';
    my $user_session_id = $cgi->cookie($user_session_tracking_cookie_name)
        || $ENV{REMOTE_ADDR} . '.' . join( '', (localtime)[ 2, 3, 4, 5 ] );
    return $user_session_id;
}

# Like category_search() but returns an array_ref only of the category names.
sub test_category_search {
    my $search_phrase = shift;
    my $results_aref  = category_search($search_phrase);
    my @ordered_names = map { $_->{category_name} } @$results_aref;
    return \@ordered_names;
}

# Expand the given list of words to include their synonyms, if any
sub expand_synonyms {
    my $dbh               = shift;
    my $phrase_words_aref = shift;

    my $word_list = '';
    foreach (@$phrase_words_aref) { $word_list .= $dbh->quote($_) . ',' }
    chop $word_list;

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'SYNONYM_STEM_X_WORD_STEM');

    my @synonyms = @{ $dbh->selectcol_arrayref( $sql, undef, $word_list ) };

    @$phrase_words_aref
        = Set::Array->new( @synonyms, @$phrase_words_aref )->unique();

}

# Takes an array ref of terms to search on.
#
# Returns an arrayref of hashrefs, e.g.
# [
#    { category_id => 800000, category_name => 'this is a category name', score=> 2}
#     , ...
# ]
# ordered first by score (ie.relevance) and second alphabetically
# where each category_id is unique
sub get_matching_categories {
    my $dbh               = shift;
    my $search_terms_aref = shift;

 # Get all the rows in the all_names_view that case-insensitively
 # match any of our synonyms on word boundaries. Our query uses a POSIX regex.
    my $regex = '[[:<:]](' . join( '|', @$search_terms_aref ) . ')[[:>:]]';

#   my $sql = "SELECT category_id, stems_name, category_name FROM all_names_view " .
#       "WHERE stems_name ~* '$regex'";

    # TODO: option of LEFT JOIN vs JOIN
    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'ALL_NAMES_VIEW_CATS_WITH_VISIBLE_PROGRAMS_VIEW_X_ALL_FOR_SYNONYMS');

    my $matches_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, $regex );

    order_matches( $matches_aref, $search_terms_aref );

    return $matches_aref;

}

# Takes
#   - an arrayref of hashrefs:
#    { category_id => 800000, stems_name => 'category or alias stems',
#      category_name => 'a category name'}
#   - the synonym-expanded search list
#
# Adds scores to each of the hasrefs above, and eliminates any non-
# highest scoring duplicates categories, then orders the encapsulating
# arrayref as mentioned several times in this script.
sub order_matches {
    my $matches_aref      = shift;
    my $phrase_words_aref = shift;

    # Adds a score=>N keyvalue pair to each hash referenced in the array
    score( $matches_aref, $phrase_words_aref );

    my %match_by_id;

    # Discards the non-highest scoring duplicate category match
    foreach my $match_href (@$matches_aref) {
        my $category_id = $match_href->{category_id};
        if ( !exists $match_by_id{$category_id}
            or $match_href->{score} > $match_by_id{$category_id}->{score} )
        {
            $match_by_id{$category_id} = $match_href;
        }
    }

    # Puts the data into the structure that is expected by callers
    my @sorted_matches = sort {
               ( $b->{score} <=> $a->{score} )
            or ( $a->{category_name} cmp $b->{category_name} )
    } ( values %match_by_id );

    @$matches_aref = @sorted_matches;
}

# Adds a score => nnn to each hashref (a match) in $matches_aref.
# For now, anyway, the score is the number of words in common
# between a match and the given phrase words.
sub score {
    my $matches_aref      = shift;
    my $phrase_words_aref = shift;

    foreach my $match (@$matches_aref) {
        my @words_in_name = split( /[\s\/]+/, $match->{stems_name} );
        my @words_in_both = Set::Array->new(@words_in_name)
            ->intersection($phrase_words_aref);
        $match->{score} = scalar @words_in_both;
    }

}

1;

