package MVHub::Notifications;
use strict;
use warnings;

use Carp;
use Date::Calc;

use MVHub::Common;
use MVHub::Utils;
use MVHub::Utils::DB;

use HTML::Template;
use Text::Wrap;

our ($VERSION) = '$Revision: 1334 $' =~ /([.\d]+)/;

use base 'Exporter';
our @EXPORT_OK = qw(
    _push_and_return_programs
    create_notifications_using
    create_standard_notification_emails_with
    create_fourth_notification_emails_with
    create_final_notification_emails_with
    die_if_record_count_mismatch
    get_email_constants_from
    get_expired_records
    increment_reminder_levels_for
    print_emails_contained_in
    send_emails_contained_in
    set_reminder_level_in_db
    shift_off_notifications_with_reminder_count
);

sub _format_expired_record_list {
    my $expired_records_aref = shift;
    my $result = join "\n   ", @$expired_records_aref;
    return "   $result";
}

sub _prepare_standard_body {
    my $template_file        = shift;
    my $email_constants_href = shift;
    my $notice_href          = shift;

    my $tmpl = HTML::Template->new( filename => $template_file );
    my $quick_login_url = MVHub::Utils::generate_quick_login_url(
        $email_constants_href->{website_name},
        $notice_href->{agency_id},
        $notice_href->{quick_login_passwd}
    );

    $tmpl->param(
        first_name      => $notice_href->{contact_first_name},
        agency_name     => $notice_href->{agency_name},
        contact_info    => '<!--TMPL_VAR name="contact_info"-->',
        signature       => '<!--TMPL_VAR name="signature"-->',
        expired_records => '<!--TMPL_VAR name="expired_records"-->',
        website_name    => $email_constants_href->{website_name},
        quick_login_url => $quick_login_url,
    );

    # wrap text to 75 columns
    $Text::Wrap::columns = 75;
    $Text::Wrap::huge    = 'overflow';
    my @paragraphs = split /\n/, $tmpl->output();
    my @text = Text::Wrap::fill( '', '', @paragraphs );
    my $text = join "\n", @text;

    # we jump through these hoop because we
    # we don't want to Text::Wrap various things
    $tmpl = HTML::Template->new( scalarref => \$text );
    my $expired_records
        = _format_expired_record_list( $notice_href->{expired_record_list} );

    my $contact_info
        = "   $email_constants_href->{admin_email}\n   $email_constants_href->{admin_phone}";
    my $signature
        = "\n$email_constants_href->{admin_name}\n$email_constants_href->{team_name}";

    $tmpl->param(
        expired_records => $expired_records,
        contact_info    => $contact_info,
        signature       => $signature,
    );

    return $tmpl->output();
}

sub _prepare_standard_subject {
    my $record_href = shift;

    my %nth = (
        0 => '',                   # This is supposed to be empty
        1 => '2nd Reminder: ',
        2 => '3rd Reminder: ',
        3 => '4th Reminder: ',
        4 => 'FINAL REMINDER: ',
    );

    my $subject = $nth{ $record_href->{reminders_sent} }
        . "Time to update your $record_href->{agency_name} listing";

    return $subject;
}

sub _push_and_return_programs {
    scalar @_ == 2 or croak "Bad number of parameters";
    my $programs_aref = shift || [];
    croak "Parameter not an array reference" if ref $programs_aref ne 'ARRAY';

    my $new_program = shift or croak 'Parameter $new_program not defined';
    push @$programs_aref, $new_program;

    return $programs_aref;
}

sub create_notifications_using {
    ( scalar @_ == 1 || scalar @_ == 2 ) or croak "Bad number of parameters";

    my $expired_records_aref = shift;
    croak "First Parameter not an array reference"
        if ref $expired_records_aref ne 'ARRAY';

    my $notifications_href = shift || {};
    croak "Second Parameter not a hash reference"
        if ref $notifications_href ne 'HASH';

    my %notifications_for = %$notifications_href;

    foreach my $expired_record_href (@$expired_records_aref) {
        my $unique_id = $expired_record_href->{'agency_id'}
            . $expired_record_href->{'contact_email'};

        $notifications_for{$unique_id}{'agency_id'}
            = $expired_record_href->{'agency_id'};

        $notifications_for{$unique_id}{'agency_name'}
            = $expired_record_href->{'agency_name'};

        $notifications_for{$unique_id}{'quick_login_passwd'}
            = $expired_record_href->{'quick_login_passwd'};

        $notifications_for{$unique_id}{'contact_email'}
            = $expired_record_href->{'contact_email'};

        $notifications_for{$unique_id}{'agency_contact_email'}
            = $expired_record_href->{'agency_contact_email'};

        $notifications_for{$unique_id}{'public_email'}
            = $expired_record_href->{'email'};

        $notifications_for{$unique_id}{'contact_first_name'}
            = $expired_record_href->{'contact_first_name'};

        $notifications_for{$unique_id}{'contact_last_name'}
            = $expired_record_href->{'contact_last_name'};

        $notifications_for{$unique_id}{expired_record_list}
            = _push_and_return_programs(
            $notifications_for{$unique_id}{expired_record_list},
            $expired_record_href->{'record_name'} );

        # To be sure we send only one email per contact person
        # We set the reminder count for all records to the
        # highest count found in any record.
        if ( !( defined $notifications_for{$unique_id}{'reminders_sent'} )
            || $expired_record_href->{'reminders_sent'}
            > $notifications_for{$unique_id}{'reminders_sent'} )
        {
            $notifications_for{$unique_id}{reminders_sent}
                = $expired_record_href->{'reminders_sent'};
        }
    }
    return %notifications_for;
}

sub create_standard_notification_emails_with {
    my $template_file        = shift;
    my $email_constants_href = shift;
    croak "Second parameter must be a hash reference"
        if ref $email_constants_href ne 'HASH';
    my %notifications = @_;
    my @emails;

    foreach my $unique_id ( keys %notifications ) {
        my $to
            = qq{$notifications{$unique_id}{contact_first_name} $notifications{$unique_id}{contact_last_name} <$notifications{$unique_id}{contact_email}>};

        my $email = MIME::Entity->build(
            Type => 'text/plain',
            From => $email_constants_href->{from},
            To   => $to,
            Subject =>
                _prepare_standard_subject( $notifications{$unique_id} ),
            Data => _prepare_standard_body(
                $template_file, $email_constants_href,
                $notifications{$unique_id}
            ),
        );
        push @emails, $email;
    }
    return @emails;
}

sub create_fourth_notification_emails_with {
    my $template_file        = shift;
    my $email_constants_href = shift;
    croak "Second parameter must be a hash reference"
        if ref $email_constants_href ne 'HASH';
    my %notifications = @_;
    my @emails;

    foreach my $unique_id ( keys %notifications ) {
        next
            if $notifications{$unique_id}{'contact_email'} eq
                $notifications{$unique_id}{'agency_contact_email'};

        my $to
            = qq{$notifications{$unique_id}{contact_first_name} $notifications{$unique_id}{contact_last_name} <$notifications{$unique_id}{contact_email}>};

        my $email = MIME::Entity->build(
            Type => 'text/plain',
            From => $email_constants_href->{from},
            To   => $to,
            CC   => $notifications{$unique_id}{'agency_contact_email'},
            Subject =>
                _prepare_standard_subject( $notifications{$unique_id} ),
            Data => _prepare_standard_body(
                $template_file, $email_constants_href,
                $notifications{$unique_id}
            ),
        );
        push @emails, $email;
    }
    return @emails;
}

sub create_final_notification_emails_with {
    my $template_file        = shift;
    my $email_constants_href = shift;
    croak "Second parameter must be a hash reference"
        if ref $email_constants_href ne 'HASH';
    my %notifications = @_;
    my @emails;

    foreach my $unique_id ( keys %notifications ) {
        next
            if !( defined $notifications{$unique_id}{'public_email'} )
                || $notifications{$unique_id}{'contact_email'} eq
                $notifications{$unique_id}{'public_email'};

        my $contact_email
            = qq{$notifications{$unique_id}{contact_first_name} $notifications{$unique_id}{contact_last_name} <$notifications{$unique_id}{contact_email}>};

        my $email = MIME::Entity->build(
            Type => 'text/plain',
            From => $email_constants_href->{from},
            To   => $notifications{$unique_id}{'public_email'},
            CC   => $contact_email,
            Subject =>
                _prepare_standard_subject( $notifications{$unique_id} ),
            Data => _prepare_standard_body(
                $template_file, $email_constants_href,
                $notifications{$unique_id}
            ),
        );
        push @emails, $email;
    }
    return @emails;
}

sub die_if_record_count_mismatch {
    my %param         = @_;
    my $input_01_aref = $param{input_01};
    my $input_02_aref = $param{input_02};
    my $in_count      = scalar @$input_01_aref + scalar @$input_02_aref;

    my $output_href = $param{output};
    my $out_count   = 0;
    foreach my $k ( keys %$output_href ) {
        my $expired_record_list = $$output_href{$k}{expired_record_list};
        $out_count += scalar @$expired_record_list;
    }
    MVHub::Utils::assert(
        $out_count == $in_count,
        "\n****Record count mismatch****\n input records = $in_count\n output records = $out_count"
    );
}

sub get_email_constants_from {
    my $cfg = shift;
    croak "Parameter must be a Config::Simple object"
        if ref $cfg ne 'Config::Simple';

    my %email_constants;
    my $error_msg = '';

    $email_constants{admin_name} = $cfg->param('NOTIFICATION.admin_name')
        or $error_msg .= "\tNOTIFICATION.admin_name\n";
    $email_constants{admin_email} = $cfg->param('NOTIFICATION.admin_email')
        or $error_msg .= "\tNOTIFICATION.admin_email\n";
    $email_constants{admin_phone} = $cfg->param('NOTIFICATION.admin_phone')
        or $error_msg .= "\tNOTIFICATION.admin_phone\n";
    $email_constants{team_name} = $cfg->param('NOTIFICATION.team_name')
        or $error_msg .= "\tNOTIFICATION.team_name\n";
    $email_constants{website_name} = $cfg->param('SITE.website_name')
        or $error_msg .= "\tSITE.website_name\n";

    my $config_filename = $cfg->param('META.this_config_file');

    if ($error_msg) {
        $error_msg
            = "config file: ${config_filename}\nmissing keys: \n$error_msg";
        croak $error_msg;
    }

    $email_constants{from}
        = "$email_constants{admin_name} <$email_constants{admin_email}>";

    return \%email_constants;
}

sub get_expired_records {
    scalar @_ == 8
        or croak "Bad number of parameters";
    my %param = @_;

    my $sql
        = MVHub::Utils::DB::get_sql_select_statement( $param{'query_name'} );

    my @bind_variables
        = ( $param{'expire_date'}, $param{'max_notifications'} );

    return $param{'dbh'}
        ->selectall_arrayref( $sql, { Slice => {} }, @bind_variables );
}

sub increment_reminder_levels_for {
    scalar @_ == 2 or croak "Bad number of parameters";
    my $reminders_href    = shift;
    my $max_notifications = shift;
    foreach my $unique_id ( keys %$reminders_href ) {
        warn
            "SHOULDN'T SEE THIS: Reminder count greater than MAX_REMINDER for $unique_id\n"
            if ( $reminders_href->{$unique_id}{'reminders_sent'}
            >= $max_notifications );
        $reminders_href->{$unique_id}{'reminders_sent'} += 1;
    }
}

sub print_emails_contained_in {
    my @emails = @_;
    foreach my $email (@emails) {
        print $email->as_string();
    }
}

sub send_emails_contained_in {
    my @emails = @_;
    foreach my $email (@emails) {
        MVHub::Common::sendmail($email);
    }
}

sub set_reminder_level_in_db {
    scalar @_ == 4
        or croak "Bad number of parameters";
    my $dbh            = shift;
    my $expire_date    = shift;
    my $reminders_href = shift;
    my $statement      = shift;
    my @bind_variables;

    my $sql = MVHub::Utils::DB::get_sql_update_statement($statement);

    foreach my $unique_id ( keys %$reminders_href ) {
        @bind_variables = (
            $reminders_href->{$unique_id}{'reminders_sent'},
            $reminders_href->{$unique_id}{'agency_id'}, $expire_date
        );
        $dbh->do( $sql, undef, @bind_variables );
    }
}

sub shift_off_notifications_with_reminder_count {
    scalar @_ == 2
        or croak "Bad number of parameters";

    my $reminder_count     = shift;
    my $notifications_href = shift;
    croak "Second Parameter not a hash reference"
        if ref $notifications_href ne 'HASH';

    my %shifted_reminders;

    foreach my $unique_id ( keys %$notifications_href ) {
        if ( $$notifications_href{$unique_id}{reminders_sent}
            == $reminder_count )
        {
            $shifted_reminders{$unique_id}
                = delete $$notifications_href{$unique_id};
        }
    }
    return %shifted_reminders;
}

1;
__END__

=head1 NAME
 
MVHub::Notifications - subs that build lists of emails and related info needed to send notifications of out-of-date records.
 
 
=head1 VERSION
 
This documentation refers to MVHub::Notifications version $Revision: 1334 $
 
=head1 SYNOPSIS
 
    use MVHub::Notifications;
    # Brief but working code example(s) here showing the most common usage(s)
 
    # This section will be as far as many users bother reading
    # so make it as educational and exemplary as possible.
  
  
=head1 DESCRIPTION
 
subs used by notification_email.pl and tested in notification_email.t
 
 
=head1 SUBROUTINES/METHODS 
 
A separate section listing the public components of the module's interface. 
These normally consist of either subroutines that may be exported, or methods
that may be called on objects belonging to the classes that the module provides.
Name the section accordingly.
 
In an object-oriented module, this section should begin with a sentence of the 
form "An object of this class represents...", to give the reader a high-level
context to help them understand the methods that are subsequently described.
 
 
=head1 DIAGNOSTICS
 
A list of every error and warning message that the module can generate
(even the ones that will "never happen"), with a full explanation of each 
problem, one or more likely causes, and any suggested remedies.
(See also  QUOTE \" " INCLUDETEXT "13_ErrorHandling" "XREF83683_Documenting_Errors_"\! Documenting Errors QUOTE \" " QUOTE " in Chapter "  in Chapter  INCLUDETEXT "13_ErrorHandling" "XREF40477__"\! 13.)
 
 
=head1 CONFIGURATION AND ENVIRONMENT
 
A full explanation of any configuration system(s) used by the module,
including the names and locations of any configuration files, and the
meaning of any environment variables or properties that can be set. These
descriptions must also include details of any configuration language used.
(also see  QUOTE \" " INCLUDETEXT "19_Miscellanea" "XREF40334_Configuration_Files_"\! Configuration Files QUOTE \" " QUOTE " in Chapter "  in Chapter  INCLUDETEXT "19_Miscellanea" "XREF55683__"\! 19.)
 
 
=head1 DEPENDENCIES
 
A list of all the other modules that this module relies upon, including any
restrictions on versions, and an indication whether these required modules are
part of the standard Perl distribution, part of the module's distribution,
or must be installed separately.
 
 
=head1 INCOMPATIBILITIES
 
There are no known incompatibilites
 
=head1 BUGS AND LIMITATIONS
 
There are no known bugs in this module. 
Please report problems to help@thecsl.org
Patches are welcome.
 
=head1 AUTHOR

Community Software Lab help@thecsl.org 
 
=head1 LICENCE AND COPYRIGHT

Copyright (c) 2004-2008 Community Software Lab, Inc ( http://thecsl.org )

The GNU Affero General Public License, version 3.0 or later


=cut
 

