
package MVHub::Page;

# This module's subroutines provides a common enclosing HTML page in which to
# place and optionally print your page content. Provides common banner,
# navigation, etc., with options to include or omit various elements.

use CGI;
use HTML::Template;

use MVHub::Utils;

####
# IMPORTANT DEVELOPER'S NOTE: prepare_page() & print_page_and_exit()
# have many arguments in common. If you change one's argument list,
# see if the other needs changing too
####

# This tag is used in a variety of templates, wherever we have to show
# the search phrase the user previously used:
our $SEARCH_PHRASE_TAG = 'last_search_phrase';

my $M_BODY_TAG               = 'mode_body';
my $M_SHOW_SEARCH_BAR_TAG    = 'show_search_bar';
my $M_OMIT_LEFT_MENU_TAG     = 'omit_left_menu';
my $M_SHOW_NSP_LEFT_MENU_TAG = 'show_nsp_left_menu';
my $M_JAVASCRIPT_TAG         = 'javascript_tag';
my $M_STYLE_TAG              = 'style_tag';
my $M_TITLE_TAG              = 'title_tag';

# TMPL VAR for the search bar template:
my $SB_CHECK_BY_AGENCY_TAG = 'search_bar_check_by_agency';

my $DEFAULT_TITLE = "Nonprofit Services Directory";
my $PARTIAL_TITLE = "";

# sub prepare_page
# Returns a page, ready for printing.
# Named Arguments:
#    body: the content for your page. Required. Omit <body> tags.
#    title: optional text to be added to the default page title (more or less). Omit <title> tags.
#    javascript: optional javascript code. Omit <script> tags.
#    transfer_internal_style_sheet: boolean, optional, default is false.
#       If true, you can include a <style>...</style> tag in the text of
#       your body argument, which will then be extracted and incorporated into
#       the resulting page header's internal style sheet.
#    show_left_menu: boolean, optional, default is false. Whether to show the left navigational menu.
#    show_search_bar: boolean, optional, default is false. Whether to show the search bar. If true, then
#       you may be interested in these arguments:
#        search_phrase: optional text with which to pre-populate the search bar's text field.
#        agency_search: boolean, optional, default is false. If true, the search bar's search type
#           is pre-set to an agency search, otherwise it's a program search.
sub prepare_page {
    my %args       = @_;
    my $body       = delete $args{body};
    my $title      = delete $args{title};
    my $javascript = delete $args{javascript};
    my $transfer_internal_style_sheet
        = delete $args{transfer_internal_style_sheet};
    my $show_left_menu       = delete $args{show_left_menu};
    my $show_search_bar      = delete $args{show_search_bar};
    my $search_phrase        = delete $args{search_phrase};
    my $site_code            = delete $args{site_code};
    my $site_location        = delete $args{site_location};
    my $search_box_help_text = delete $args{search_box_help_text};
    my $agency_search        = delete
        $args{agency_search};  # 1: agency search, 0 (default): program search
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    MVHub::Utils::assert( $body, "Missing required 'body' argument" );

    my $tmpl = HTML::Template->new(
        filename          => 'main.tmpl',
        die_on_bad_params => 0
    ) or die "Failed to initialize new template with file $MAIN_TEMPLATE";

    if ($transfer_internal_style_sheet) {

        # Note--the leading newline means that <style> sections
        # cannot be at the top of a document.
        $body =~ s|\n<style.*>(.*)</style>||s;
        my $styles = $1;
        $tmpl->param( $M_STYLE_TAG, $styles ) if $styles;
    }

    my $show_nsp_left_menu = 0;
    if ( $site_code eq 'nsp' ) {
        $show_nsp_left_menu = 1;
    }

    $tmpl->param( $M_BODY_TAG,               $body );
    $tmpl->param( $M_TITLE_TAG,              _make_title($title) );
    $tmpl->param( $M_SHOW_SEARCH_BAR_TAG,    $show_search_bar );
    $tmpl->param( $M_JAVASCRIPT_TAG,         $javascript );
    $tmpl->param( $M_OMIT_LEFT_MENU_TAG,     !$show_left_menu );
    $tmpl->param( $M_SHOW_NSP_LEFT_MENU_TAG, $show_nsp_left_menu );

    # The above template does a tmpl_include of the search bar.
    # Initialize the search bar, as appropriate:
    if ($show_search_bar) {
        if ($agency_search) {
            $tmpl->param( $SB_CHECK_BY_AGENCY_TAG, $agency_search );
        }

        $tmpl->param( $SEARCH_PHRASE_TAG,     $search_phrase );
        $tmpl->param( 'site_location',        $site_location );
        $tmpl->param( 'search_box_help_text', $search_box_help_text );
    }

    return $tmpl->output();
}

# sub print_page_and_exit
# Prints a page. Includes a no-cache directive in its header. Exits when done.
# Named Arguments (in addition to the arguments provided by prepare_page()):
#    cookie: optional
sub print_page_and_exit {
    my %args       = @_;
    my $body       = delete $args{body};
    my $title      = delete $args{title};
    my $cookie     = delete $args{cookie};
    my $javascript = delete $args{javascript};
    my $transfer_internal_style_sheet
        = delete $args{transfer_internal_style_sheet};
    my $show_left_menu  = delete $args{show_left_menu};
    my $show_search_bar = delete $args{show_search_bar};
    my $search_phrase   = delete $args{search_phrase};
    my $agency_search   = delete
        $args{agency_search};  # 1: agency search, 0 (default): program search
    my $site_code = delete $args{site_code};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my $content = prepare_page(
        body                          => $body,
        javascript                    => $javascript,
        transfer_internal_style_sheet => $transfer_internal_style_sheet,
        show_left_menu                => $show_left_menu,
        show_search_bar               => $show_search_bar,
        search_phrase                 => $search_phrase,
        agency_search                 => $agency_search,
        site_code                     => $site_code
    );

    my %header_options = ( -Cache_Control => 'no-store' );

    if ( defined $cookie ) {
        $header_options{-cookie} = $cookie;
    }

    # Print header with no-cache options
    print CGI::header(%header_options);

    print $content;

    exit(0);

}

######## PRIVATE ############

# Return an appropriate page title. If text is passed in, it
# will be included in the title.
sub _make_title {
    my $text = shift;
    return $DEFAULT_TITLE if ( !$text );
    if ( $text =~ /(program|service)s?/i ) {
        return "$text $PARTIAL_TITLE";
    }
    else {
        return "$text Services $PARTIAL_TITLE";
    }
}

1;
