package MVHub::ProgramFieldDefs;
require Exporter;

use strict;
use warnings;

use MVHub::Common;

our @ISA    = qw(Exporter);
our @EXPORT = qw(get_program_field_definitions_href);

our ($VERSION) = '$Revision: 1582 $' =~ /([.\d]+)/;

# A list of fields in the order that that you might want to display them
# (somewhere besides the form, whose display is dictated by templates),
# e.g. in an email.
# Excludes the date_updating_option field.
our @FIELDS = (
    'agency_id',                     'program_id',
    'program_name',                  'last_updated',
    'program_alias',                 'address1',
    'address2',                      'city',
    'state',                         'zip',
    'mail_address1',                 'mail_address2',
    'mail_city',                     'mail_state',
    'mail_zip',                      'main_phone',
    'toll_free_phone',               'hotline',
    'fax',                           'tty_tdd',
    'website',                       'email',
    'email_verification',            'short_program_description',
    'long_program_description',      'eligibility',
    'area_type',                     'area_data',
    'areas_served_description',      'fees_charged',
    'fees_explained',                'program_schedule',
    'hours',                         'languages',
    'languages_description',         'appointment',
    'bring',                         'online_registration',
    'bus_lines',                     'parking',
    'transportation',                'handicapped_accessible',
    'child_care_provided',           'child_care_description',
    'wait_list',                     'wait_list_description',
    'other_information',             'volunteer_opportunities',
    'volunteer_information',         'resource_library',
    'resource_library_description',  'meeting_space',
    'meeting_space_description',     'news_letter',
    'news_letter_description',       'public_net_access',
    'public_net_access_description', 'contact_first_name',
    'contact_last_name',             'contact_phone',
    'contact_fax',                   'contact_email',
    'contact_email_verification',    'contact_email_available',
    'admin_notes'
);

# The field definitions. These are divided into 3 hashes (representing the 3 main sections
# of the displayed form) just for organizational purposes. It would also make it easy
# if we ever wanted to break the form into multiple pages (in which we'd refactor
# get_program_field_definitions() to take the name of a hash).
my %program_field_defs = (
    agency_id    => { type => 'hidden' },
    program_id   => { type => 'hidden' },
    last_updated => {
        type  => 'static',
        label => 'Record last updated on:'
    },
    date_updating_option => {
        type => 'checkbox',
        options =>
            [ { 'yes' => 'Check here to update this date upon submission' } ]
    },
    program_name  => { required => 1 },
    program_alias => { label    => 'Also Known As' },
    address1      => { label    => 'Address 1', required => 1 },
    address2      => { label    => 'Address 2' },
    city          => { required => 1 },
    state => { type => 'select', options => [ 'MA', 'NH' ], required => 1 },
    zip =>
        { size => 5, maxlength => 5, validate => 'ZIPCODE', required => 1 },

    mail_address1 => { label => 'Address 1' },
    mail_address2 => { label => 'Address 2' },
    mail_city     => { label => 'City' },
    mail_state =>
        { type => 'select', options => [ 'MA', 'NH' ], label => 'State' },
    mail_zip =>
        { size => 5, maxlength => 5, validate => 'ZIPCODE', label => 'Zip' },

    main_phone => {
        size      => '24',
        maxlength => '24',
        validate  => $MVHub::Common::PHONE_REG_EX,
        comment   => '(ex: 123-456-7890)',
        required  => 1
    },

    toll_free_phone => {
        size      => '24',
        maxlength => '24',
        comment   => '(ex: 800-456-7890)',
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    hotline => {
        size      => '24',
        maxlength => '24',
        label     => 'Hotline/Emergency/After Hours #',
        comment   => '(ex: 123-456-7890)',
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    fax => {
        size      => '24',
        maxlength => '24',
        comment   => '(ex: 123-456-7890)',
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    tty_tdd => {
        size      => '24',
        maxlength => '24',
        label     => 'TTY/TDD',
        validate  => $MVHub::Common::PHONE_REG_EX,
        comment   => '(ex: 123-456-7890)'
    },
    website            => {},
    email              => { size => 30, validate => 'EMAIL' },
    email_verification => {
        size     => 30,
        validate => {
            perl       => 'eq $form->field("email")',
            javascript => '== form.email.value'
        }
    },

    short_program_description => { required => 1 },
    long_program_description =>
        { type => 'textarea', rows => 6, required => 1 },
    eligibility => {
        type     => 'textarea',
        rows     => 2,
        required => 1,
        comment  => '<br>(ex: income, town, state, age, gender, etc)<br>'
    },
    areas_served_description => {
        label => 'Areas Served - Extra Info',
        type  => 'textarea'
    },
    fees_charged => {
        type     => 'radio',
        options  => [ 'Free', 'Sliding Scale', 'Yes' ],
        required => 1
    },
    fees_explained   => { type => 'textarea' },
    program_schedule => {
        type     => 'textarea',
        required => 1,
        comment =>
            '<br>(ex: "Open year round", "Closed in the summer", etc.)<br>'
    },
    hours     => { type => 'textarea', required => 1, label => 'Days/Hours' },
    languages => {
        options => [
            'English', 'Portuguese', 'French', 'Spanish',
            'Khmer',   'Vietnamese'
        ],
        type     => 'checkbox',
        columns  => 3,
        required => 1
    },
    languages_description => {
        type  => 'textarea',
        label => 'Languages - Extra Info'
    },
    appointment => {
        label    => 'Appointment required?',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        type     => 'radio',
        required => 1
    },
    bring => {
        type  => 'textarea',
        label => 'What to bring when registering',
        comment =>
            '<br>(ex: proof of residency, proof of income, insurance info, etc)<br>'
    },
    online_registration => {
        label    => 'On-line registration?',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        type     => 'radio',
        required => 1
    },
    parking        => { type => 'textarea' },
    transportation => {
        type => 'textarea',
        comment =>
            '<br>(if your program provides transportation to participants, explain here)<br>'
    },
    handicapped_accessible => {
        type     => 'radio',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    child_care_provided => {
        type     => 'radio',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    child_care_description => { type => 'textarea' },
    wait_list              => {
        type     => 'radio',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    wait_list_description => { type => 'textarea' },
    other_information     => { type => 'textarea', rows => 3 },
    admin_notes => {
        type => 'textarea',
        label =>
            'Administrative Notes (only shown here, and only to administrator)',
        comment => '<br>ex: 12/15/2005: Left phone message. (FS)<br>'
    }
);

my %contact_field_defs = (

    #    use_agency_contact => {
    #       type=>'radio',
    #       options=>['Yes, use the agency contact info above',
    #                 'No, use the new contact info below'],
    #       required=> 1
    #    },
    contact_first_name =>
        { label => 'First Name', size => 24, required => 1 },
    contact_last_name => { label => 'Last Name', size => 24, required => 1 },
    contact_phone     => {
        label     => 'Phone',
        size      => 24,
        maxlength => 24,
        comment   => '(ex: 123-456-7890)',
        required  => 1,
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    contact_fax => {
        label     => 'Fax',
        size      => 24,
        maxlength => 24,
        comment   => '(ex: 123-456-7890)',
        validate  => $MVHub::Common::PHONE_REG_EX
    },
    contact_email => {
        label    => 'Email Address',
        size     => 30,
        required => 1,
        validate => $MVHub::Common::ALLOW_CONTACT_EMAIL_REGEX
    },
    contact_email_verification => {
        label    => 'E-mail Address (please verify)',
        size     => 30,
        required => 1,
        validate => {
            perl       => 'eq $form->field("contact_email")',
            javascript => '== form.contact_email.value'
        }
    },
    contact_email_available => {
        label => 'Please make my e-mail available to other non-profits '
            . 'so I can be informed of events &amp; programs.',
        options  => [ 'Yes', 'No' ],
        type     => 'radio',
        required => 1,
    }
);

my %other_info_field_defs = (
    volunteer_opportunities => {
        type     => 'radio',
        label    => 'Volunteer Opportunities Available',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    volunteer_information => {
        type => 'textarea',
        rows => 3
    },
    resource_library => {
        type     => 'radio',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    resource_library_description => {
        type => 'textarea',
        rows => 3
    },
    meeting_space => {
        type     => 'radio',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    meeting_space_description => {
        type => 'textarea',
        rows => 3
    },
    news_letter => {
        type     => 'radio',
        options  => [ 'Yes', 'No', 'Not Applicable' ],
        required => 1
    },
    news_letter_description => {
        type => 'textarea',
        rows => 3
    },
    public_net_access => {
        type     => 'radio',
        options  => [ 'Limited', 'Yes', 'No' ],
        required => 1,
        label    => 'Do you offer Internet Access to the Public?'
    },
    public_net_access_description => {
        type  => 'textarea',
        rows  => 3,
        label => 'Public Internet Access Description'
    }
);

# Get the field definitions as a hash ref, keyed by field name.
sub get_program_field_definitions_href {
    return { %program_field_defs, %contact_field_defs,
        %other_info_field_defs };
}

1;
