package MVHub::Synonyms;

use strict;
use warnings;

# ---------------------------------------------------------------
# Loads synonyms from the $SYNONYM_FILE (defined below) into the
# database.
#
# This script must be run from the installed site (i.e. not from
# your cvs projects path).
#
# Each line in the synonym file is a space-delimited list of one-word
# synonyms
#
# Since the $SYNONYM_TABLE (below) has no dependents (i.e.
# other tables that reference its data), this script may be re-run on
# on an existing site.
# The $SYNONYM_SEQUENCE (below) is reset to its initial value.
#
# ---------------------------------------------------------------

use MVHub::Common;

my $SYNONYM_TABLE = 'synonym_stem';

# The auto-incrementing sequence for the above table:
my $SYNONYM_SEQUENCE = 'synonym_id_sequence';

# The intial value for the sequence. Note: this is also hardcoded
# into database/create.sql.
my $SYNONYM_SEQUENCE_INIT_VAL = '900000';

# Returns a list of lines, where each line is a space delimited list of
# one word synonyms of each other.
sub fetch_synonyms_from_db {
    my $dbh = MVHub::Utils::DB::get_dbh();

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'SYNONYM_STEM_X_SUBMITTED_WORDS');

    my $rows_aref = $dbh->selectcol_arrayref($sql);
    foreach ( @{$rows_aref} ) {
        s/,//g;
    }
    return @{$rows_aref};
}

# Takes a list of lines, where each line is space delimited list of
# one word synonyms of each other.
# Submits them into the database table, $SYNONYM_TABLE.
# Dies on any database error.
sub submit_synonyms_to_db {
    my @synonym_lines = @_;
    my $dbh           = MVHub::Utils::DB::get_dbh();
    $dbh->{AutoCommit} = 0;

    my @bind_variables = ($SYNONYM_TABLE);
    my $sql
        = MVHub::Utils::DB::get_sql_delete_statement('SYNONYM_STEM_X_ALL');
    $dbh->do( $sql, undef, @bind_variables );

    @bind_variables = ( $SYNONYM_SEQUENCE, $SYNONYM_SEQUENCE_INIT_VAL );
    $sql = MVHub::Utils::DB::get_sql_select_statement('SETVAL_X_SYNONYM_SEQ');
    $dbh->do( $sql, undef, @bind_variables );

    foreach my $synonym_line (@synonym_lines) {
        $synonym_line = MVHub::Utils::trim($synonym_line);
        next unless $synonym_line;    # skip empty lines

        my @synonyms_in_line = split( /\s+/, $synonym_line );

        my $synonym_id = MVHub::Utils::DB::get_next_in_sequence( $dbh,
            $SYNONYM_SEQUENCE );
        foreach my $synonym (@synonyms_in_line) {
            MVHub::Utils::DB::insert_one_record(
                dbh         => $dbh,
                table       => $SYNONYM_TABLE,
                values_href => {
                    synonym_id     => $synonym_id,
                    word_stem      => MVHub::Common::get_stems($synonym),
                    submitted_word => $synonym
                }
            );
        }
    }

    $dbh->commit();
    $dbh->{AutoCommit} = 1;
}
