package MVHub::Utils;
use strict;
use warnings;

our ($VERSION) = '$Revision: 1740 $' =~ /([.\d]+)/xm;

use Carp;
use HTML::Strip;

use base 'Exporter';
our @EXPORT_OK = qw(
    assert
    choose_default_or_custom_file
    clean_cgi_params
    die_if_missing_env_vars
    generate_quick_login_url
    mm_dd_yyyy_to_yyyy_mm_dd
    parse_zip_code
    trim
    undelimit_field
    yyyy_mm_dd_to_mm_dd_yyyy
);

sub assert {
    my ( $expression, $string ) = @_;
    return if ($expression);
    my ( $package, $filename, $line ) = caller;
    my $msg = "\n\nAssertion failed at line $line of $filename : ";
    if ( defined $string ) { $msg .= "$string"; }

    # We don't want the filename and line number of this file:
    $msg .= "\n\n";
    confess $msg;
}

sub clean_cgi_params {
    my $cgi = shift;
    foreach my $param ( $cgi->param() ) {
        my @values = $cgi->param($param);

        @values = map { _clean_cgi_param($_) } @values;
        $cgi->param( $param, @values );
    }
    return $cgi;
}

sub _clean_cgi_param {
    my $clean_me = shift;
    $clean_me =~ s/^\s+//xm;
    $clean_me =~ s/\s+$//xm;

    $clean_me = HTML::Strip->new()->parse($clean_me);

    return $clean_me;
}

sub die_if_missing_env_vars {
    my @expected_env_vars
        = qw/MV_CONFIG_FILE HTML_TEMPLATE_ROOT PERL5LIB MV_TEST_CONFIG_FILE/;

    my $missing_vars;
    foreach my $var ( sort @expected_env_vars ) {
        if ( not defined $ENV{$var} ) {
            $missing_vars .= "\t$var\n";
        }
    }

    my $extra_vars;
    my @mv_vars = sort grep {/^MV_/xm} keys %ENV;

    foreach my $var (@mv_vars) {
        my $exists = grep { /$var/xm; } @expected_env_vars;
        if ( !$exists ) {
            $extra_vars .= "\t$var\n";
        }
        $exists = 0;
    }
    carp
        "unexpected enviroment vars in die_if_missing_env_vars(): $extra_vars\n;"
        if $extra_vars;

    if ($missing_vars) {
        my $msg = <<"MSG";
  Undefined environment variables 
  $missing_vars
  check apache config or source env_vars_setup.sh 
MSG
        croak $msg;
    }
}

sub generate_quick_login_url {
    my $website_name = shift or croak "Missing Param: website_name\n";
    my $agency_id    = shift or croak "Missing Param: agency_id\n";
    my $ql_id        = shift or croak "Missing Param: ql_id\n";
    return "http://${website_name}/${agency_id}/${ql_id}";
}

sub parse_zip_code {
    my $zip_code = shift;
    if ( defined $zip_code && $zip_code =~ /(\d{5})/xm ) {
        return $1;
    }
    else {
        return;
    }
}

sub mm_dd_yyyy_to_yyyy_mm_dd {
    my $date = shift;
    return if ( !$date );
    my $pattern = '^(\d{1,2})-(\d{1,2})-(\d\d\d\d)$';
    if ( $date !~ $pattern ) {
        croak("Date argument ($date) must be in mm-dd-yyyy format.");
    }
    $date =~ s/$pattern/$3-$1-$2/xm;
    return $date;
}

sub trim {
    my @out = @_;

    foreach (@out) {
        if ( defined $_ ) {
            s/^\s+//xm;
            s/\s+$//xm;
        }
    }
    return wantarray ? @out : $out[0];
}

sub undelimit_field {
    my $value = shift;
    return $value if !defined $value;

    # fix  problems caused by trailing DELEM
    $value =~ s/(.*)!DELIM!$/$1/;

    my $FIELD_DELIMITER = '!DELIM!';

    my @split = split( /$FIELD_DELIMITER/, $value );
    return @split > 1 ? \@split : $value;
}

sub yyyy_mm_dd_to_mm_dd_yyyy {
    my $date = shift;
    return if ( !$date );
    my $pattern = '^(\d\d\d\d)-(\d{1,2})-(\d{1,2})$';
    if ( $date !~ $pattern ) {
        croak("Date argument ($date) must be in yyyy-dd-mm format.");
    }
    $date =~ s/$pattern/$2-$3-$1/xm;
    return $date;
}

# return true as all modules should do
1;
__END__

=head1 NAME

MVHub::Utils - misc utility functions


MVHub::Utils  A collection of utility sub routines used by MVHub

=head1 VERSION

This documentation refers to MVHub::Utils version $Revision: 1740 $

=head1 SYNOPSIS

    use MVHub::Utils;

=head1 DESCRIPTION

subs used by more than one script 

=head1 SEE ALSO

 MVHub::Common

=head1 LAST MODIFIED

Last modified on $Date: 2009-02-27 19:01:34 -0500 (Fri, 27 Feb 2009) $ by $Author: omacneil $

=head1 BUGS

none known 

=head1 LICENSE AND COPYRIGHT

The GNU Affero General Public License, version 3.0 or later

Copyright (c) 2004-2008 Community Software Lab, Inc ( http://thecsl.org )

=cut

=head1 SUBROUTINES/METHODS 


=head2 assert($boolean,$message)

assert will croak with $message if $boolean is not true.

 assert(1=2,'2 does not equal 1');
 assert(@lines, 'there are some lines');



=head2 clean_cgi_params($cgi)

Takes a CGI object;

Returns a CGI object with all params having 
leading and trailing spaces stripped and
all html markup stripped 

Handles multi value CGI params

my $cleaned_cgi=clean_cgi_params($cgi);

=head2 die_if_missing_env_vars ()

  If any MV_* enviroment variable is missing, die
  If there are unexpected  MV_* enviroment variables, warn

  undef $ENV{MV_CONFIG_FILE};
  eval{die_if_missing_env_vars();};
  print 'died' if $@;

  $ENV{MV_FOO}  
  eval{die_if_missing_env_vars();};
  print 'died' if $@;

=head2 parse_zip_code($zip)

  Accept a string containing 5 digits, (a zip code)
  return only those 5 digits, or undef if the string
  doesn't contain a zip code

  my $zip=parse_zip_code('adsf 01852 ');
  print $zip == 01852;
  $zip=parse_zip_code('adsf  ');
 print $zip eq undef;

=head2 mm_dd_yyyy_to_yyyy_mm_dd($date)

 Returns converted date. Date parts must be delimited by dashes.
 day and month may be single digits.
 Takes a string or array of strings

  my $date=mm_dd_yyyy_to_yyyy_mm_dd('12-31-1999');
  print $date eq '1999-12-31';

=head2 trim($string)

 Takes a string or  array of strings

 Returns a copy of the given string or array of strings with
 leading and trailing whitespace stripped off. Undefined
 strings are left untouched.

  my $trimmed_str=trim('  asdf ');
  my @trimmed_strs =trim(' a ', 'b ');

=head2 yyyy_mm_dd_to_mm_dd_yyyy($date)
 
 Returns converted date. Date parts must be delimited by dashes.
 day and month may be single digits.
 Takes a string or  array of strings

  my $date=yyyy_mm_dd_to_mm_dd_yyyy('1999-12-31');
  print $date eq '12-31-1999';

=head1 DIAGNOSTICS

Error messages are listed with documentation for sub routines.

=head1 CONFIGURATION AND ENVIRONMENT

MVHub::Utils does not require configuration beyond that used by MVHub itself

=head1 DEPENDENCIES

Carp, HTML::Strip, Config::Simple;

=head1 BUGS AND LIMITATIONS

There are no known bugs in this module.
Please report problems to help@thecsl.org
Patches are welcome.

=head1 AUTHOR

Community Software Lab help@thecsl.org
=head1 INCOMPATIBILITIES

No known incompatibilities


