package MVHub::Utils::ConfigSimple;
use strict;
use warnings;

use Config::Simple;

our ($VERSION) = '$Revision: 1740 $' =~ /([.\d]+)/xm;

use Carp;

use base 'Exporter';
our @EXPORT_OK = qw(
    create_config_from
    _calculate_extra_cfg_params_from
);

sub create_config_from {
    my $config_file = shift or croak "No config filename parameter";
    my $user = shift;
    if ( !( defined $user ) ) {
        $user = defined $ENV{SUDO_USER} ? $ENV{SUDO_USER} : $ENV{USER};
    }
    my $cfg;
    $cfg = new Config::Simple($config_file)
        or croak
        "Error opening config file: '$config_file' Config::Simple Error:"
        . Config::Simple->error();
    return _calculate_extra_cfg_params_from( $cfg, $config_file, $user );
}

sub _calculate_extra_cfg_params_from {
    my $cfg = shift;
    croak "Missing Config::Simple parameter"
        if ref $cfg ne "Config::Simple";
    my $config_file = shift or croak "No config filename parameter";
    my $user        = shift or croak "No user parameter";
    $cfg->param( "META.this_config_file", $config_file );

    my $error_msg;

    my $relative_paths_href  = $cfg->get_block('RELATIVE_PATH');
    my $count_relative_paths = scalar( keys %$relative_paths_href );

    if ( $count_relative_paths == 0 ) {
        $error_msg
            .= "no RELATIVE_PATH keys found in config file: '$config_file'\n";
    }

    my $base_dir = $cfg->param('BASE.dir')
        or $error_msg
        .= "param: 'BASE.dir' not found in config file: '$config_file'\n";
    croak $error_msg if $error_msg;

    foreach my $path_name ( keys %$relative_paths_href ) {
        my $absolute_path
            = "$base_dir/$user/$$relative_paths_href{$path_name}";
        $path_name =~ s|^RELATIVE_PATH\.||;
        $cfg->param( "ABSOLUTE_PATH.$path_name", $absolute_path );
    }
    return $cfg;
}

# return true as all modules should do
1;
__END__

=head1 NAME

MVHub::Utils::ConfigSimple - read constants from our config file  

MVHub::Utils::ConfigSimple  

=head1 VERSION

This documentation refers to MVHub::Utils version $Revision: 1740 $

=head1 SYNOPSIS

    use MVHub::Utils::ConfigSimple qw/create_config_from /;

    my $cfg=create_config_from($ENV{MV_CONFIG_FILE});
    
    my $allow_robots=$cfg->param('SITE.allow_robots');
 
    my $tmp_dir=$cfg->param('ABSOLUTE_PATH.tmp_dir');

=head1 DESCRIPTION

MVHub::Utils::ConfigSimple is a wrapper to Config::Simple 
that exports single function create_config_from , that takes 
a config filename and returns a Config::Simple object

You can then use the various Config::Simple methods, in 
particular: param(); to get configuration parameters/
CONSTANTS.

[BASE].dir and [RELATIVE_PATH] are stored in the 
config file. 

Calculated paths which include $ENV{USER} can be retrieved with the [ABSOLUTE_PATH] 
section.

In most cases, you can create the Config Files and 
Config objects as you would a regular Config::Simple
object. 

# in the config file:

[BASE]
dir=/var/www/mvhub/

[RELATVE_PATH] 
tmp_dir=tmp/

# in Perl

print  $cfg->param('ABSOLUTE_PATH.tmp_dir') 
  eq  '/var/www/mvhub/$ENV{USER}/tmp/'

The application configuration file
is currently either /var/www/mvhub/$USER/conf/mvh.conf or 
/var/www/mvhub/$USER/conf/nsp.conf (though you should NOT
depend on these files remaining in these places. 

A better practice for getting the name of the config file
is the enviroment variable $ENV{MV_CONFIG_FILE}. which is set
in the development shell enviroment with the 'cdw' alias 
and in the web server in the apache configuration file.

Our wrapper allows us to calculate paths and store them
as unique pieces in the config file. 

=head1 SEE ALSO

 Config::Simple
 MVHub::CGIAppBase
 
=head1 LAST MODIFIED

Last modified on $Date: 2009-02-27 19:01:34 -0500 (Fri, 27 Feb 2009) $ by $Author: omacneil $

=head1 BUGS

none known 

=head1 LICENSE AND COPYRIGHT

The GNU Affero General Public License, version 3.0 or later

Copyright (c) 2004-2008 Community Software Lab, Inc ( http://thecsl.org )

=cut

=head1 SUBROUTINES/METHODS 


=head1 DIAGNOSTICS

Error messages are listed with documentation for sub routines.

=head1 CONFIGURATION AND ENVIRONMENT

MVHub::Utils does not require configuration beyond that used by MVHub itself

=head1 DEPENDENCIES

Carp, Config::Simple;

=head1 BUGS AND LIMITATIONS

There are no known bugs in this module.
Please report problems to help@thecsl.org
Patches are welcome.

=head1 AUTHOR

Community Software Lab help@thecsl.org
=head1 INCOMPATIBILITIES

No known incompatibilities
