package MVHub::Utils::DB;
use strict;
use warnings;

our ($VERSION) = '$Revision: 1740 $' =~ /([.\d]+)/xm;

use Carp;
use DBI;

use MVHub::Utils::ConfigSimple;
use MVHub::Utils;

use SQL::Library;

use base 'Exporter';
use SQL::Library;

our @EXPORT_OK = qw(
    get_data_source
    get_dbh
    get_full_name
    get_hash_of_agencies
    get_next_in_sequence
    get_program_name
    get_sql_insert_statement
    get_sql_select_statement
    get_sql_update_statement
    get_sql_delete_statement
    insert_one_record
    remove_agency
    select_agency_record
    select_program_record
    update_one_record
);

sub get_data_source {
    my $file = shift or croak 'missing parameter: config_file';

    my $cfg = MVHub::Utils::ConfigSimple::create_config_from($file);

    my $host          = $cfg->param('DATABASE.database_host') || q{};
    my $db_name       = $cfg->param('DATABASE.database_name') || q{};
    my $database_type = $cfg->param('DATABASE.database_type') || q{};
    my $port          = $cfg->param('DATABASE.database_port') || q{};

    # if port not configured, don't set it
    my $port_setting = $port ? ";port=$port" : q{};

    my $data_source;
    if ( $database_type eq 'Pg' ) {
        $data_source .= "dbi:Pg:dbname=$db_name;host=$host$port_setting";
    }
    elsif ( $database_type eq 'SQLite' ) {
        $data_source .= "dbi:SQLite:dbname=$db_name";
    }
    else {
        croak "'$database_type': unknown database_type in '$file'";
    }

    return $data_source;
}

sub get_dbh {
    my $config_file = shift or croak "missing param: config_file";
    croak "couldn't read config_file: '$config_file'"
        if not -r $config_file;

    my $data_source = get_data_source($config_file);

    my $cfg               = new Config::Simple($config_file);
    my $database_username = $cfg->param('DATABASE.database_user');
    my $auth              = $cfg->param('DATABASE.database_magic_word');

    my %attr = (
        'PrintError'         => 0,
        'RaiseError'         => 1,
        'AutoCommit'         => 1,
        'ShowErrorStatement' => 1,
    );

    my $dbh = DBI->connect( $data_source, $database_username, $auth, \%attr )
        or die "couldn't connect to database";
    return $dbh;
}

# Given a username, user_id, or agency id, return the corresponding full name.
# For example, an agency id of 101342 would return "Community Software Lab,"
# while a username of jmiller would return "John Miller."
sub get_full_name {
    my $dbh                    = shift;
    my $username               = shift;
    my $is_data_entry_operator = shift;

    my $sql
        = MVHub::Utils::DB::get_sql_select_statement('AGENCY_X_AGENCY_NAME');

    my ( $first_name, $last_name )
        = $dbh->selectrow_array( $sql, undef, $username );

    return $last_name ? "$first_name $last_name" : $first_name;
}

# Get a hash of all agencies in the form : agency_id => agency_name
# sorted by agency name
sub get_hash_of_agencies {
    my $dbh = shift;

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'AGENCY_X_AGENCY_ID_NAME');

    my $aref = $dbh->selectcol_arrayref( $sql, { Columns => [ 1, 2 ] } );
    my %agency_hash = @$aref;
    return %agency_hash;
}

# Advance a database sequence (Postgres-speak for auto incrementing counter),
# and return its new value.
# Why would you ever call this? If you needed to insert a row AND find
# out its new ID, call this first and explicitly use the value for the ID column
# in the INSERT statement.
# This is the only way to do this in Postgres/DBI (currently) without having to
# worry about race conditions.
sub get_next_in_sequence {
    my ( $dbh, $sequence_name ) = @_;
    my $sql
        = MVHub::Utils::DB::get_sql_select_statement('NEXTVAL_X_SEQUENCE');
    my @row = $dbh->selectrow_array( $sql, undef, $sequence_name );
    MVHub::Utils::assert(
        @row == 1,
        "Expected one row from $sequence_name sequence selection; "
            . "got @row"
    );
    return $row[0];
}

sub get_program_name {
    my $dbh        = shift;
    my $program_id = shift;

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'PROGRAM_ID_X_PROGRAM_NAME');

    my ($program_name) = $dbh->selectrow_array( $sql, undef, $program_id );
    return $program_name;
}

sub get_sql_insert_statement {
    my $statement_name = shift
        or croak "missing parameter: \$statement_name\n";
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from(
        $ENV{MV_CONFIG_FILE} );
    my $conf_path = $cfg->param('ABSOLUTE_PATH.conf_dir');
    my $file      = "$conf_path/sql_insert.lib";
    my $sql_lib   = new SQL::Library { lib => $file };
    my $result    = $sql_lib->retr($statement_name);
    croak "sql_insert.lib: '$statement_name' query does not exist\n"
        if !( defined $result );

    return $result;
}

sub get_sql_select_statement {
    my $statement_name = shift
        or croak "missing parameter: \$statement_name\n";
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from(
        $ENV{MV_CONFIG_FILE} );
    my $conf_path = $cfg->param('ABSOLUTE_PATH.conf_dir');
    my $file      = "$conf_path/sql_select.lib";
    my $sql_lib   = new SQL::Library { lib => $file };
    my $result    = $sql_lib->retr($statement_name);
    croak "sql_select.lib: '$statement_name' query does not exist\n"
        if !( defined $result );

    return $result;
}

sub get_sql_update_statement {
    my $statement_name = shift
        or croak "missing parameter: \$statement_name\n";
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from(
        $ENV{MV_CONFIG_FILE} );
    my $conf_path = $cfg->param('ABSOLUTE_PATH.conf_dir');
    my $file      = "$conf_path/sql_update.lib";
    my $sql_lib   = new SQL::Library { lib => $file };
    my $result    = $sql_lib->retr($statement_name);
    croak "sql_update.lib: '$statement_name' query does not exist\n"
        if !( defined $result );

    return $result;
}

sub get_sql_delete_statement {
    my $statement_name = shift
        or croak "missing parameter: \$statement_name\n";
    my $cfg = MVHub::Utils::ConfigSimple::create_config_from(
        $ENV{MV_CONFIG_FILE} );
    my $conf_path = $cfg->param('ABSOLUTE_PATH.conf_dir');
    my $file      = "$conf_path/sql_delete.lib";
    my $sql_lib   = new SQL::Library { lib => $file };
    my $result    = $sql_lib->retr($statement_name);
    croak "sql_delete.lib: '$statement_name' query does not exist\n"
        if !( defined $result );

    return $result;
}
#################################

# Insert a record in the given database table.
# Arguments:
#    dbh - a database handle.
#
#    table - surface with 3 or more legs.
#
#    values_href - a hashref of column names and their new values. Missing column
#        names as well as undefined values will produce NULLs in their corresponding
#        columns.
#
# Dies if exactly one row was not inserted.
sub insert_one_record {
    my %args        = @_;
    my $dbh         = delete $args{dbh};
    my $table       = delete $args{table};
    my $values_href = delete $args{values_href};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    my @fields = sort keys %{$values_href};
    my @bind_values;
    foreach my $field_name (@fields) {
        push @bind_values, $values_href->{$field_name};
    }
    my $qmarks = "?," x @fields;
    chop($qmarks);
    my $fields = join( ",", @fields );
    my $sql = "INSERT INTO $table ($fields) VALUES ($qmarks)";

    my $row_count = $dbh->do( $sql, undef, @bind_values );
    MVHub::Utils::assert( $row_count == 1,
        "Expected one $table row inserted; " . "got $row_count" );
}

sub remove_agency {
    my $dbh            = shift;
    my $agency_id      = shift;
    my @bind_variables = ($agency_id);
    my $sql
        = MVHub::Utils::DB::get_sql_delete_statement('AGENCY_X_AGENCY_ID');
    $dbh->do( $sql, undef, @bind_variables );
}

# Return a hashref of agency data for the given agency id.
# If the given agency id does not exist in the database, and
# the allow_empty_result argument is:
#    - false or undef: we die. This is the typical desired behavior.
#    - true: we return undef. This gives the caller more control in
#            error handling.
sub select_agency_record {
    my %args               = @_;
    my $dbh                = $args{dbh};
    my $agency_id          = $args{agency_id};
    my $allow_empty_result = $args{allow_empty_result};

    my @bind_variables = ($agency_id);
    my $sql
        = MVHub::Utils::DB::get_sql_select_statement('AGENCY_X_AGENCY_ID');

    my $records_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, @bind_variables );

    return if ( @$records_aref == 0 && $allow_empty_result );
    MVHub::Utils::assert(
        @$records_aref == 1,
        "Expected one agency row selected with agency_id=$agency_id; "
            . "got "
            . scalar @$records_aref
    );

    my $record_href = $records_aref->[0];

    # Make the date more readable

    $record_href->{'last_updated'} = MVHub::Utils::yyyy_mm_dd_to_mm_dd_yyyy(
        $record_href->{'last_updated'} );

    # Set email verification fields to value of email fields, since the
    # verification fields are not stored in the db
    $record_href->{email_verification} = $record_href->{email};
    $record_href->{contact_email_verification}
        = $record_href->{contact_email};

    return $record_href;
}

sub select_program_record {
    my $dbh        = shift;
    my $agency_id  = shift;
    my $program_id = shift;

    my @bind_values = ( $agency_id, $program_id );
    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'PROGRAM_X_AGENCY_PROGRAM_ID');

    my @result
        = @{ $dbh->selectall_arrayref( $sql, { Slice => {} }, @bind_values )
        };
    my $record_href = $result[0];

    # Make the date more readable

    $record_href->{'last_updated'} = MVHub::Utils::yyyy_mm_dd_to_mm_dd_yyyy(
        $record_href->{'last_updated'} );

    # Set email verification fields to value of email fields, since the
    # verification fields are not stored in the db
    $record_href->{email_verification} = $record_href->{email};
    $record_href->{contact_email_verification}
        = $record_href->{contact_email};

   # turn any delimited strings back into array refs (for multi-valued fields)
    foreach my $field_name ( keys %$record_href ) {
        $record_href->{$field_name}
            = MVHub::Utils::undelimit_field( $record_href->{$field_name} );
    }

    return $record_href;
}

# Update a unique record in the given database table.
# Arguments:
#    dbh - a database handle.
#
#    table - surface with 3 or more legs.
#
#    values_href - a hashref of column names and their new values. Missing column
#        names will leave their corresponding table columns untouched. Undef
#        value(s) will update the corresponding table column(s) with NULL(s).
#
#    where_href - a hashref of column(s) to values that will be tested for equality
#        in order to select the one row. The column(s) must represent a unique key.
#
# Dies if more or less than one row is updated.

sub update_one_record {
    my %args        = @_;
    my $dbh         = delete $args{dbh};
    my $table       = delete $args{table};
    my $values_href = delete $args{values_href};
    my $where_href  = delete $args{where_href};
    MVHub::Utils::assert( ( scalar keys %args ) == 0,
        "Unknown arguments: @{[%args]}" );

    if ( !defined $dbh ) {
    }

    my @field_names  = keys %{$values_href};
    my @field_values = values %{$values_href};

    my $sql = "UPDATE $table SET";

    foreach my $field_name (@field_names) {
        $sql .= " $field_name = ?,";
    }
    chop($sql);

    my @where_names  = keys %{$where_href};
    my @where_values = values %{$where_href};
    $sql .= " WHERE ";
    $sql .= join( ' = ? AND ', @where_names );
    $sql .= ' = ? ';

    my @bind_values = ( @field_values, @where_values );

    my $row_count = $dbh->do( $sql, undef, @bind_values );

    MVHub::Utils::assert(
        $row_count == 1,
        "Expected one $table row where: @{[%$where_href]}; "
            . "got $row_count"
    );

    # The above trick displays the contents of $where_href
}

#return true like all good modules should
1;

__END__

=head1 NAME

MVHub::Utils::DB -functions to return info from database

=head1 VERSION

This documentation refers to MVHub::Utils::DB version $Revision: 1740 $

=head1 SYNOPSIS

    use MVHub::Utils::DB;

=head1 DESCRIPTION

subs that return a datbase handle ($dbh) or take a database handle and return 
data from the Database

=head1 SEE ALSO

 MVHub::Common MVHub::Utils

=head1 LAST MODIFIED

Last modified on $Date: 2009-02-27 19:01:34 -0500 (Fri, 27 Feb 2009) $ by $Author: omacneil $

=head1 BUGS

none known 

=head1 LICENSE AND COPYRIGHT

The GNU Affero General Public License, version 3.0 or later

Copyright (c) 2004-2008 Community Software Lab, Inc ( http://thecsl.org )

=cut

=head1 SUBROUTINES/METHODS 

=head2 get_data_source($config_file)

  Given a configuration file return the right DBI
  data source , depending on what database_type
  is in the config file
      
     "dbi:SQLite:dbname=$db_name"
   
   or:
      "dbi:Pg:dbname=$db_name;host=$host";
                
   croak if the database_type in the config file is neither
   SQLite or Pg


=head2 get_dbh
 
 TODO
 
=head2 get_full_name
 
 TODO
 
=head2 get_hash_of_agencies
 
 TODO
 
=head2 get_next_in_sequence
 
 TODO
 
=head2 get_program_name
 
 TODO
 
=head2 get_sql_insert_statement($SQL_STATEMENT_NAME)

   Given the name of a sql INSERT statement, returns the SQL for that
   statement. Currently the file the SQL statements are stored in
   is: app-mvhub/conf/sql_insert.lib, though you should not depend
   on this location and you should not need to depend on this location
   as finding the SQL::Library file is handled internally for you.
           
 
=head2 get_sql_select_statement($SQL_SELECT_STATMENT_NAME)

   Given the name of a sql SELECT statement, returns the SQL for that
   statement. Currently the file the SQL statements are stored in
   is: app-mvhub/conf/sql_select.lib, though you should not depend
   on this location and you should not need to depend on this location
   as finding the SQL::Library file is handled internally for you.

=head2 get_sql_update_statement

   Given the name of a sql UPDATE statement, returns the SQL for that
   statement. Currently the file the SQL statements are stored in
   is: app-mvhub/conf/sql_update.lib, though you should not depend
   on this location and you should not need to depend on this location
   as finding the SQL::Library file is handled internally for you.
   
=head2 get_sql_delete_statement

   Given the name of a sql DELETE statement, returns the SQL for that
   statement. Currently the file the SQL statements are stored in
   is: app-mvhub/conf/sql_delete.lib, though you should not depend
   on this location and you should not need to depend on this location
   as finding the SQL::Library file is handled internally for you.
  
=head2 insert_one_record
 
 TODO
 
=head2 remove_agency
 
 TODO
 
=head2 select_agency_record
 
 TODO
 
=head2 select_program_record
 
 TODO
 
=head2 update_one_record
 
 TODO
 
                                                                                               
=head1 DIAGNOSTICS

Error messages are listed with documentation for sub routines.

=head1 CONFIGURATION AND ENVIRONMENT

MVHub::Utils does not require configuration beyond that used by MVHub itself

=head1 DEPENDENCIES

Carp, DBI, HTML::Strip

=head1 BUGS AND LIMITATIONS

There are no known bugs in this module.
Please report problems to help@thecsl.org
Patches are welcome.

=head1 AUTHOR

Community Software Lab help@thecsl.org
=head1 INCOMPATIBILITIES

No known incompatibilities


