package MVHub::Utils::Setup;

use strict;
use warnings;

use Carp;
use Config::Simple;
use DateTime;
use DBI;
use IO::File;
use IO::Prompt;
use File::Basename;
use File::Copy qw/move/;
use Params::Validate qw/:all/;

use autodie qw/ :all move /;

use MVHub::Utils::DB;
use MVHub::Utils::ConfigSimple;

our @EXPORT_OK = qw(

    @SITE_CODES
    add_apache_config
    add_dns
    determine_user_config_filename_from
    get_dev_home_dirs
    get_dev_home_files
    get_production_data
    get_testing_data
    make_config_files_for
    make_database_user
    make_databases_for
    make_dev_home_files
    make_dirs_from
    load_production_data
    load_test_data_for
    remove_cron_jobs
    restart_services
    update_etc_files
);

our @SITE_CODES = qw/ mvh nsp /;

sub _append_to_file {
    ( my $file, my $append_lines ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { type => SCALAR }
    );

    my $tmp_file = "$file.tmp";

    open( my $IN, "<", $file );
    my @lines = <$IN>;
    close($IN);
    push @lines, $append_lines;

    open( my $OUT, ">", $tmp_file );
    print $OUT @lines;

    close($OUT);

    # man File::Copy
    move( $tmp_file, $file );
}

sub _add_calculable_values {
    ( my $template_cfg, my $username, my $site_code )
        = Params::Validate::validate_pos(
        @_,
        { type => HASHREF, isa => 'Config::Simple' },
        { type => SCALAR },
        { type => SCALAR }
        );

    $template_cfg->param( 'DATABASE.database_name', "$username.$site_code" );
    $template_cfg->param( 'DATABASE.database_user', "$username" );
    $template_cfg->param( 'NOTIFICATION.admin_email',
        "$username\@thecsl.org" );
    $template_cfg->param( 'NOTIFICATION.dev_email', "$username\@thecsl.org" );
    $template_cfg->param( 'SITE.website_name',
        "$site_code.$username.testing123.net" );
    $template_cfg->param( 'SITE.website_code',     "$site_code" );
    $template_cfg->param( 'RELATIVE_PATH.log_dir', "log/$site_code/" );
    $template_cfg->param( 'RELATIVE_PATH.reports_dir',
        "reports/$site_code/" );
    $template_cfg->param( 'RELATIVE_PATH.reports_dev_dir', "dev/" );

    return $template_cfg;
}

sub _add_site_specific_values {
    ( my $template_cfg, my %site_specific_values )
        = Params::Validate::validate_pos(
        @_,
        { type => HASHREF, isa => 'Config::Simple' },
        (1) x ( @_ - 1 )
        );

    foreach my $key ( keys %site_specific_values ) {
        $template_cfg->param( $key, $site_specific_values{$key} );
    }
    return $template_cfg;
}

sub _add_user_generated_values {
    ( my $template_cfg, my $user_cfg, my $overwrite, my @keys )
        = Params::Validate::validate_pos(
        @_,
        { type => HASHREF, isa => 'Config::Simple' },
        { type => HASHREF, isa => 'Config::Simple' },
        { type => SCALAR },
        (1) x ( @_ - 3 )
        );

    foreach my $key (@keys) {
        if ( $overwrite || !( $user_cfg->param($key) ) ) {
            prompt("Please enter a value for $key: ");
            $template_cfg->param( $key, $_ );
        }
    }
    return $template_cfg;
}

sub _db_exists {
    ( my $dbh, my $db_name ) = Params::Validate::validate_pos(
        @_,
        { type => HASHREF, can => 'selectall_arrayref' },
        { type => SCALAR },
    );

    my $sql
        = MVHub::Utils::DB::get_sql_select_statement('PG_DATABASE_X_DATNAME');

    my $matching_dbs_aref = $dbh->selectall_arrayref( $sql, {}, ($db_name) );
    return scalar @$matching_dbs_aref;

}

sub _db_user_found {
    ( my $dbh, my $username ) = Params::Validate::validate_pos(
        @_,
        { type => HASHREF, can => 'selectall_arrayref' },
        { type => SCALAR },
    );

    my $sql = MVHub::Utils::DB::get_sql_select_statement('PG_USER_X_USENAME');

    my $matching_db_users_aref
        = $dbh->selectall_arrayref( $sql, {}, ($username) );
    return scalar @$matching_db_users_aref;
}

sub _db_user_has_needed_roles {
    ( my $dbh, my $username ) = Params::Validate::validate_pos(
        @_,
        { type => HASHREF, can => 'selectall_arrayref' },
        { type => SCALAR },
    );

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'PG_ROLES_X_ROLNAME_CREATEROLE_CREATEDB');

    my $role_rows_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, ($username) );
    my $result = $$role_rows_aref[0]{rolcreaterole}
        && $$role_rows_aref[0]{rolcreatedb};

    return $result ? 1 : 0;
}

sub _db_user_has_superuser_role {
    ( my $dbh, my $username ) = Params::Validate::validate_pos(
        @_,
        { type => HASHREF, can => 'selectall_arrayref' },
        { type => SCALAR },
    );

    my $sql = MVHub::Utils::DB::get_sql_select_statement(
        'PG_ROLES_X_ROLNAME_ROLSUPER');

    my $role_rows_aref
        = $dbh->selectall_arrayref( $sql, { Slice => {} }, ($username) );
    return $$role_rows_aref[0]{rolsuper} ? 1 : 0;
}

sub _do_or_die {
    ( my $cmd ) = Params::Validate::validate_pos( @_, { type => SCALAR }, );

    $cmd =~ s/^\s+//;
    my @cmd = split /\s+/, $cmd;
    system(@cmd);
}

sub _get_dbh_for_template1_db {
    ( my $cfg ) = Params::Validate::validate_pos( @_, { can => 'param' }, );

    my $database_username = $cfg->param('DATABASE.database_user');
    my $auth              = $cfg->param('DATABASE.database_magic_word');
    my $host              = $cfg->param('DATABASE.database_host');
    my $port              = $cfg->param('DATABASE.database_port');

    my $data_source = "dbi:Pg:dbname=template1;host=$host;port=$port;";

    my %attr = (
        'PrintError'         => 1,
        'RaiseError'         => 1,
        'AutoCommit'         => 1,
        'ShowErrorStatement' => 1,
    );

    my $dbh = DBI->connect( $data_source, $database_username, $auth, \%attr );
    return $dbh;

}

sub _get_site_specific_values {
    ( my $site_code )
        = Params::Validate::validate_pos( @_, { type => SCALAR }, );

    my %hash_of_site_specific_values = (
        'mvh' => {
            'SITE.location' => 'Massachusetts Merrimack Valley',
            'SITE.search_box_help_text' =>
                q/(e.g. family counseling) or organizations (YMCA or UTEC)/,
            'SITE.website_description' => 'Merrimack Valley Hub',
            'NOTIFICATION.team_name'   => 'Merrimack Valley Hub Team'
        },
        'nsp' => {
            'SITE.location' => 'Massachusetts North Shore',
            'SITE.search_box_help_text' =>
                q/(e.g. family counseling) or organizations (YMCA or NSCAP)/,
            'SITE.website_description' => 'North Shore Portal',
            'NOTIFICATION.team_name'   => 'North Shore Portal Team'
        }
    );
    return %{ $hash_of_site_specific_values{$site_code} };
}

sub _make_empty_config_file_named {
    ( my $filename )
        = Params::Validate::validate_pos( @_, { type => SCALAR }, );

    open my $out_fh, '>', $filename
        or die "failed to open $filename for writing error:$!\n";
    print $out_fh "#Config file\n [ignore]\n";
    close $out_fh
        or die "failed to close $filename error:$!\n";
    return 1;
}

sub _merge_conf_objects {
    ( my $template_cfg, my $user_cfg ) = Params::Validate::validate_pos(
        @_,
        { can => 'param' },
        { can => [qw/param delete/] },
    );

    foreach my $key ( $template_cfg->param() ) {
        if ( !( $user_cfg->param($key) ) ) {
            my $value = $template_cfg->param($key);
            $value = $value ? $value : 'BAD_MISSING';
            $user_cfg->param( $key, $value );
        }
    }

    #remove keys no longer used
    foreach my $key ( $user_cfg->param() ) {
        if ( !defined $template_cfg->param($key) ) {
            $user_cfg->delete($key);
        }
    }

    return $user_cfg;
}

sub _string_already_in_file {
    ( my $host, my $filename ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { type => SCALAR },
    );

    open( my $in_fh, '<', $filename );

    my @contents = <$in_fh>;
    my $regex    = qr/$host/;
    return scalar grep { $_ =~ $regex } @contents;
}

sub _run_sql_db_update_scripts {
    my @scripts = @_;

    foreach my $script (@scripts) {
        my $filename = File::Basename::basename($script);
        print "Executing $filename...";
        my $output = `psql -f $script 2>&1`;
        if ( $output =~ /COMMIT/ ) {
            print "SUCCESS\n";
        }
        elsif ( $output =~ /ROLLBACK/ ) {
            print "FAIL\n";
            print
                "The following OUTPUT occured while processing $filename:\n";
            die "$output\n\n****Processing $filename FAILED****\n";
        }
    }

}

sub _set_db_ENV_vars {
    ( my $cfg ) = Params::Validate::validate_pos( @_, { can => 'param' }, );

    # so psql stops on error
    $ENV{ON_ERROR_STOP} = 1;

    $ENV{PGDATABASE} = $cfg->param('DATABASE.database_name');
    $ENV{PGHOST}     = $cfg->param('DATABASE.database_host');
    $ENV{PGPASSWORD} = $cfg->param('DATABASE.database_magic_word');
    $ENV{PGUSER}     = $cfg->param('DATABASE.database_user');
}

sub add_dns {
    ( my $username )
        = Params::Validate::validate_pos( @_, { type => SCALAR }, );

    foreach my $prefix (@SITE_CODES) {
        my $host = "$prefix.$username";
        my $FILE = '/etc/bind/db.testing123.net';

        my $line = sprintf( "%-15s IN      A       129.63.96.180\n", $host );
        if ( _string_already_in_file( $host, $FILE ) ) {
            warn "$host found in $FILE, skipping append\n";
            return;
        }

        _append_to_file( $FILE, $line );
    }
}

sub add_apache_config {
    ( my $username )
        = Params::Validate::validate_pos( @_, { type => SCALAR }, );

    foreach my $prefix (@SITE_CODES) {
        my $FILE = '/etc/apache2/sites-available/mvhub.data';
        my $host = "$prefix.$username.testing123.net";

        my $line = sprintf( "Use MVhub %-15s $prefix $host\n", $username );

        if ( _string_already_in_file( $host, $FILE ) ) {
            warn "$host found in $FILE, skipping append\n";
            return;
        }

        _append_to_file( $FILE, $line );
    }
}

sub apply_database_updates_for {
    ( my $username, my $cfg ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { can  => 'param' },
    );

    foreach my $prefix (@SITE_CODES) {
        my $update_user_cfg_filename
            = determine_user_config_filename_from( $username, $prefix, $cfg );
        my $update_user_cfg = MVHub::Utils::ConfigSimple::create_config_from(
            $update_user_cfg_filename);
        my $dbh = MVHub::Utils::DB::get_dbh($update_user_cfg_filename);

        my $db_version   = get_db_version($dbh);
        my $scripts_path = $cfg->param("ABSOLUTE_PATH.setup_db_dir");
        my @db_scripts   = get_sql_db_update_scripts(
            version => $db_version,
            path    => $scripts_path
        );

        if ( scalar @db_scripts ) {
            print "Applying $prefix database updates...\n";
            _set_db_ENV_vars($update_user_cfg);
            _run_sql_db_update_scripts(@db_scripts);
        }
        else {
            print "$prefix database up-to-date at Version $db_version.\n";
        }
    }
}

sub determine_user_config_filename_from {
    ( my $username, my $site_code, my $cfg ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { type => SCALAR },
        { can  => 'param' }
    );

    my $base_dir = $cfg->param('BASE.dir');

    # Conf directory hard-coded - Least evil of available choices
    # Less cumbersome then adding a conf_dir key in the config file.
    return "$base_dir/$username/conf/$site_code.conf";
}

sub get_db_version {
    ( my $dbh )
        = Params::Validate::validate_pos( @_, { can => 'selectrow_array' }, );

    my $version;

    my $exist_sql = MVHub::Utils::DB::get_sql_select_statement(
        'INFORMATION_SCHEMA_X_ONE');

    my $table_exists = ( $dbh->selectrow_array($exist_sql) )[0];

    if ($table_exists) {
        my $sql
            = MVHub::Utils::DB::get_sql_select_statement('VERSION_LOG_X_MAX');

        $version = ( $dbh->selectrow_array($sql) )[0];
    }
    else {
        $version = 0;
    }

    return $version;
}

sub get_sql_db_update_scripts {
    Params::Validate::validate(
        @_,
        {   path    => { type => SCALAR },
            version => { type => SCALAR },
        }
    );

    my %args = @_;

    my @raw_scripts = glob "$args{path}/*.sql";
    my @scripts;

    foreach my $script ( sort @raw_scripts ) {

        if ( $script =~ m#\/(\d{3})_[\w\-]+\.sql# ) {
            my $script_version = int $1;

            if ( $script_version > $args{'version'} ) {
                push @scripts, $script;
            }
        }
        else {
            die
                "ERROR: '$script' does not conform to naming standard: nnn_word_description.sql\n";
        }
    }
    return sort @scripts;
}

sub get_dirs_with {
    ( my $base_dir, my $username ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { type => SCALAR },
    );

    my @dirs = (
        {   dir         => 'conf',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        },
        {   dir         => 'cover_db',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        },
        {   dir         => 'dev/profile',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        },
        {   dir         => 'reports/nsp/visitors',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=r-x',
        },
        {   dir         => 'reports/nsp/analog',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=r-x',
        },
        {   dir         => 'reports/mvh/analog',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        },
        {   dir         => 'reports/mvh/visitors',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        },
        {   dir         => 'source-code',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        },
        {   dir         => 'log/mvh',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        },
        {   dir         => 'log/nsp',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        },
        {   dir         => 'tmp',
            owner       => '',
            group       => '',
            permissions => 'u=rwx,g=rwsx,o=rx',
        }
    );
    foreach my $dir_href (@dirs) {
        $dir_href->{owner} = $username
            if not $dir_href->{owner};
        $dir_href->{group} = $username
            if not $dir_href->{group};

        $dir_href->{dir} = "$base_dir/$username/" . $dir_href->{dir};
    }

    my $user_dir_href = {
        dir         => "$base_dir/$username/",
        owner       => $username,
        group       => $username,
        permissions => 'u=rwx,g=rwsx,o=rx'
    };
    unshift @dirs, $user_dir_href;

    return @dirs;
}

# requires rysnc to be setup as daemon on csl-db-02
# and staging.inside host to have privs to read
# from csl-db-02 via rsync
sub get_production_data {
    my $cfg = shift;
    my $cmd;

    my $tmp_dir = $cfg->param('ABSOLUTE_PATH.tmp_dir');

    my $today = ` date +%Y-%m-%d`;
    chomp $today;

    print "Getting recent backup of production data...";
    foreach my $site_code (@SITE_CODES) {

        $cmd
            = " rsync csl-db-02.inside::backup/postgres/${today}*${site_code}*dump ";
        $cmd .= " | tail -n 1";
        chomp $cmd;

        my $last_dumpfile = `$cmd`;
        $last_dumpfile = ( split / +/, $last_dumpfile )[-1];
        chomp $last_dumpfile;

        $cmd = "rsync csl-db-02.inside::backup/postgres/";
        $cmd .= "${last_dumpfile} ";
        $cmd .= "$tmp_dir/${site_code}_production_data.dump";

        system($cmd);
    }
}

sub load_production_data {
    ( my $username, my $cfg ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { can  => 'param' },
    );

    my $cmd;
    my $tmp_dir = $cfg->param('ABSOLUTE_PATH.tmp_dir');

    print "Loading production data...";
    foreach my $suffix (@SITE_CODES) {
        my $cfg_filename
            = determine_user_config_filename_from( $username, $suffix, $cfg );
        my $user_cfg
            = MVHub::Utils::ConfigSimple::create_config_from($cfg_filename);

        _set_db_ENV_vars($user_cfg);

        $cmd
            = "pg_restore -O -x -d $username.$suffix ${tmp_dir}/${suffix}_production_data.dump >/dev/null";

        ( system($cmd) == 0 )
            or warn "failed to load ${suffix}_production_data.dump";
    }
    print "done\n";
}

sub load_test_data_for {
    ( my $username, my $cfg ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { can  => 'param' },
    );

    my $path_to_project_tools
        = $cfg->param('ABSOLUTE_PATH.project_tools_dir');
    my $cmd;

    print "Loading test data...";
    foreach my $suffix (@SITE_CODES) {
        my $cfg_filename
            = determine_user_config_filename_from( $username, $suffix, $cfg );
        my $user_cfg
            = MVHub::Utils::ConfigSimple::create_config_from($cfg_filename);

        _set_db_ENV_vars($user_cfg);

        $cmd
            = "psql -f ${path_to_project_tools}/test_${suffix}_db.sql >/dev/null";

        ( system($cmd) == 0 )
            or warn "failed to load data for $username.$suffix";
    }
    print "done\n";
}

sub make_config_files_for {
    ( my $username, my $cfg, my $overwrite ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { can  => 'param' },
        { type => SCALAR }
    );

    my @keys_for_user_generated_values = qw/ DATABASE.database_magic_word /;

    my $template_config_file
        = $cfg->param('ABSOLUTE_PATH.template_conf_dir') . "/template.conf";
    my $template_cfg = new Config::Simple($template_config_file)
        or die "failed to create config object from $template_config_file";

    foreach my $site_code (@SITE_CODES) {

        my $cfg_filename
            = determine_user_config_filename_from( $username, $site_code,
            $cfg );
        my $user_cfg = new Config::Simple( syntax => 'ini' );
        if ( !-e $cfg_filename ) {
            _make_empty_config_file_named($cfg_filename);
        }
        $user_cfg->read($cfg_filename);

        print "Updating Config: $cfg_filename\n";
        my %site_specific_values = _get_site_specific_values($site_code);

        $template_cfg = _add_site_specific_values( $template_cfg,
            %site_specific_values );
        $template_cfg
            = _add_calculable_values( $template_cfg, $username, $site_code );
        $template_cfg = _add_user_generated_values( $template_cfg, $user_cfg,
            $overwrite, @keys_for_user_generated_values );

        if ($overwrite) {
            $user_cfg->write("$cfg_filename.bak");
            $user_cfg = $template_cfg;
        }
        else {
            _merge_conf_objects( $template_cfg, $user_cfg );
        }

        $user_cfg->write($cfg_filename);
    }
}

sub get_dev_home_dirs {
    ( my $username )
        = Params::Validate::validate_pos( @_, { type => SCALAR }, );

    my @dirs = (
        {   dir         => "/home/$username/.bazaar",
            owner       => "$username",
            group       => "$username",
            permissions => 'u=rwx,g=rwx,o=rx',
        },
        {   dir         => "/home/$username/.ssh",
            owner       => "$username",
            group       => "$username",
            permissions => 'u=rwx,g=,o=',
        }
    );
    return @dirs;
}

sub get_dev_home_files {
    ( my $username, my $email_address, my $launchpad_login )
        = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { type => SCALAR },
        { type => SCALAR },
        );

    my $result = ( getpwnam("$username") )[6];
    croak "$username not a valid username\n" if !( defined $result );
    my $name = ( split ",", $result )[0];

    my @files = (
        {   dest        => "/home/$username/.bazaar/bazaar.conf",
            owner       => "$username",
            group       => "$username",
            permissions => 'u=rw,g=rw,o=r',
            contents    => <<"CONTENT",
[DEFAULT]
email = $name <$email_address>
launchpad_username = $launchpad_login
mail_client = mutt

[ALIASES]
recentlog = log -r-20..-1 --short --forward
ll = log -r-10..-1 --line --forward
CONTENT
        },
        {   dest        => "/home/$username/.bazaar/locations.conf",
            owner       => "$username",
            group       => "$username",
            permissions => 'u=rw,g=rw,o=r',
            contents    => <<"CONTENT",
[/var/www/mvhub/$username/source-code]
push_location = lp:~$launchpad_login/mvhub/
push_location:policy = appendpath
public_branch = lp:~$launchpad_login/mvhub/
public_branch:policy = appendpath
submit_to = merge\@code.launchpad.net
submit_branch = lp:mvhub

[/var/www/mvhub/$username/source-code/trunk]
push_location = lp:mvhub/trunk
public_branch = lp:mvhub/trunk
CONTENT
        },
        {   dest        => "/home/$username/.ssh/config",
            owner       => "$username",
            group       => "$username",
            permissions => 'u=rw,g=,o=',
            contents    => <<'CONTENT',
Host bazaar.launchpad.net
IdentityFile ~/.ssh/id_rsa_launchpad

Host *
IdentityFile ~/.ssh/id_rsa
CONTENT
        }
    );
    return @files;
}

sub make_database_user {
    ( my $username, my $cfg ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { can  => 'param' },
    );

    my $dbh = _get_dbh_for_template1_db($cfg);

    if ( !_db_user_found( $dbh, $username ) ) {
        my $sql
            = "CREATE USER $username WITH PASSWORD 'test' CREATEDB CREATEROLE";
        $dbh->do($sql);
    }
}

sub make_databases_for {
    ( my $username, my $cfg ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { can  => 'param' },
    );

    my $dbh = _get_dbh_for_template1_db($cfg);
    my $sql;
    my $user_running_cmd = $cfg->param('DATABASE.database_user');

    print "(re)making databases for $username...";

    if ( !_db_user_has_needed_roles( $dbh, $user_running_cmd ) ) {
        warn
            "\nSkipping db stuff $user_running_cmd lacks CREATEDB or CREATEROLE\n";
        return;
    }

    if ( ( $user_running_cmd ne $username )
        && !_db_user_has_superuser_role( $dbh, $user_running_cmd ) )
    {
        my $msg = qq/
        skipping db stuff $user_running_cmd can't drop table owned by $username 
        without role SUPERUSER
        /;
        warn $msg;
        return;
    }

    foreach my $suffix (@SITE_CODES) {
        if ( _db_exists( $dbh, "$username.$suffix" ) ) {
            $sql = qq{DROP DATABASE "$username.${suffix}"};
            $dbh->do($sql);
        }

        # kludge to avoid 'db being currently being accessed' errs
        sleep 1;

        $sql
            = qq{CREATE DATABASE "$username.${suffix}" OWNER "$username" ENCODING='SQL_ASCII' };
        $dbh->do($sql);

    }
    print "done\n";
}

sub make_dev_home_files {
    my @files = @_;
    croak 'missing param: @files' if !( scalar @files );
    foreach my $file_href (@files) {
        if ( -e $file_href->{dest} ) {
            if ( -e "$file_href->{dest}.default" ) {
                print "$file_href->{dest}.default exists: skipping file\n";
                next;
            }
            print "$file_href->{dest} exists: creating a default instead.\n";
            $file_href->{dest} = "$file_href->{dest}.default";
        }
        open my $FILE, ">$file_href->{dest}" or croak "open FILE failed: $!";
        print $FILE $file_href->{contents};
        close $FILE or croak "close FILE failed: $!";
        my $chown_cmd
            = "chown $file_href->{owner}:$file_href->{group} $file_href->{dest}";
        my $chmod_cmd = "chmod $file_href->{permissions} $file_href->{dest}";
        _do_or_die($chown_cmd);
        _do_or_die($chmod_cmd);
    }
}

sub make_dirs_from {
    ( my $dir_aref_hrefs )
        = Params::Validate::validate_pos( @_, { type => ARRAYREF }, );

    my @dir_hrefs = @$dir_aref_hrefs;

    foreach my $dir_href (@dir_hrefs) {
        my $mkdir_cmd = "mkdir -p $dir_href->{dir}";
        my $chown_cmd
            = "chown -R $dir_href->{owner}:$dir_href->{group} $dir_href->{dir}";
        if ( $dir_href->{dir} =~ /log/ ) {

        # root owns log files it creates don't try to change ownership on them
            $chown_cmd
                = "chown $dir_href->{owner}:$dir_href->{group} $dir_href->{dir}";
        }
        else {
            $chown_cmd
                = "chown -R $dir_href->{owner}:$dir_href->{group} $dir_href->{dir}";
        }
        my $chmod_cmd = "chmod $dir_href->{permissions} $dir_href->{dir}";
        _do_or_die($mkdir_cmd);
        _do_or_die($chown_cmd);
        _do_or_die($chmod_cmd);
    }
    print "done\n";
}

sub restart_services {
    system('apache2ctl restart');
    my $is_apache_running = `ps aux | grep 'apache2' | grep -v grep`;
    die "Apache failed to start" if !($is_apache_running);
}

sub remove_cron_jobs {
    print 'Removing mvhub-cron from /etc...\n';
    unlink '/etc/cron.d/mvhub-cron';
}

sub run_tests {
    ( my $username, my $cfg ) = Params::Validate::validate_pos(
        @_,
        { type => SCALAR },
        { can  => 'param' },
    );

    print "Running tests...\n";
    my $project_tools_path = $cfg->param('ABSOLUTE_PATH.project_tools_dir');
    my $cmd                = "$project_tools_path/bin/mv_su_test $username";
    ( system($cmd) == 0 ) or die "Failed to run tests\n";
}

sub update_etc_files {
    ( my $cfg ) = Params::Validate::validate_pos( @_, { can => 'param' }, );

    my $etc_path = $cfg->param('ABSOLUTE_PATH.setup_etc_dir');
    print "Updating contents of /etc...\n";
    my $cmd = "cp -vur $etc_path /";
    ( system($cmd) == 0 ) or die "Failed to copy etc files\n";
}

