#!/usr/bin/perl -w

use strict;
use Test::More tests => 7;
use MVHub::HTMLTemplateUtils;

use Test::NoWarnings;
use IO::Scalar;
use TestHelper qw/make_test_file/;

use English '-no_match_vars';
local $OUTPUT_RECORD_SEPARATOR;    # enable localized slurp mode

my ( $content, $test_fh, $test_name, $params_href, $output );

# setup for first few tests

my $template_file = "$$.tmp";
$content
    = qq{<html><body>You typed in <TMPL_VAR NAME="variable">.</body></html>};
TestHelper::make_test_file( $template_file, $content );

########
$test_name = 'pass a param for a simple TMPL_VAR , get it back';
########

$params_href = { variable => "Hello World", };

# MVHub::HTMLTemplateUtils::format_template() can only to some output when printing
# we have to send it's output to a file (sigh)

$output  = '';
$test_fh = new IO::Scalar \$output
    or die "failed to open file handle";
select $test_fh;

MVHub::HTMLTemplateUtils::format_template(
    filename   => $template_file,
    param_href => $params_href,
    print      => 1,
);

select STDOUT;
like( $output, qr{<html><body>You typed in Hello World.</body></html>},
    $test_name );

########
$test_name = 'output, w/default header';
########

$params_href = { variable => "some text", };

$output  = '';
$test_fh = new IO::Scalar \$output
    or die "failed to open file handle";
select $test_fh;

MVHub::HTMLTemplateUtils::format_template(
    filename   => $template_file,
    param_href => $params_href,
    print      => 1
);
select STDOUT;

like(
    $output,
    qr{Content-Type: text/html\s+<html><body>You typed in some text.</body></html>},
    $test_name
);

########
$test_name = 'print output with header and default doctype';
########

$params_href = { variable => "an incomplete sentence", };

$output  = '';
$test_fh = new IO::Scalar \$output
    or die "failed to open file handle";
select $test_fh;

MVHub::HTMLTemplateUtils::format_template(
    filename   => $template_file,
    param_href => $params_href,
    print      => 1,
    doctype    => 1
);

select STDOUT;

like( $output, qr/^\s*<!DOCTYPE HTML PUBLIC/m, $test_name );

########
$test_name = 'output with header and custom doctype';
########

$params_href = { variable => "nothing", };

my $doctype
    = qq{<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">};

$output  = '';
$test_fh = new IO::Scalar \$output
    or die "failed to open file handle";
select $test_fh;

MVHub::HTMLTemplateUtils::format_template(
    filename   => $template_file,
    param_href => $params_href,
    print      => 1,
    doctype    => $doctype
);

select STDOUT;
is( $output,
    qq{Content-Type: text/html\n\n<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">\n<html><body>You typed in nothing.</body></html>},
    $test_name
);

########
$test_name = 'output using unnested TMPL_LIST';
########

$content
    = qq{<html><body><TMPL_LOOP NAME="mainloop">Name: <TMPL_VAR NAME="name"><br>Info: <TMPL_VAR NAME="info"><p></TMPL_LOOP></body></html>};
TestHelper::make_test_file( $template_file, $content );

my $list_params = {
    mainloop => [
        { name => "a",   info => "the first letter" },
        { name => "b",   info => "another letter" },
        { name => "zed", info => "the last letter (sometimes)" },
    ],
};

ok( MVHub::HTMLTemplateUtils::format_template(
        filename   => $template_file,
        param_href => $list_params
        ) eq
        qq{<html><body>Name: a<br>Info: the first letter<p>Name: b<br>Info: another letter<p>Name: zed<br>Info: the last letter (sometimes)<p></body></html>},
    $test_name
);

########
$test_name = 'output using nested TMPL_LIST';
########

# concat without linebreaks
$content = '';
$content .= qq{<html><body><TMPL_LOOP NAME="mainloop">};
$content .= qq{<TMPL_VAR NAME="general">:<br>};
$content .= qq{<TMPL_LOOP NAME="sudoloop">};
$content .= qq{\t- <TMPL_VAR NAME="description"><br></TMPL_LOOP><br>};
$content .= qq{</TMPL_LOOP></body></html>};

$params_href = { variable => "some text", };

TestHelper::make_test_file( $template_file, $content );

$params_href = {
    mainloop => [
        {   general  => "fire",
            sudoloop => [
                { description => "hot" },
                { description => "may burn" },
                { description => "gives off light" },
            ],
        },
        {   general  => "ice",
            sudoloop => [
                { description => "cold" },
                { description => "solid form of water" },
                { description => "fun to skate on" },
            ],
        },
    ],
};

is( MVHub::HTMLTemplateUtils::format_template(
        filename   => $template_file,
        param_href => $params_href,
    ),
    qq{<html><body>fire:<br>\t- hot<br>\t- may burn<br>\t- gives off light<br><br>ice:<br>\t- cold<br>\t- solid form of water<br>\t- fun to skate on<br><br></body></html>},
    $test_name
);

#######
# cleanup
#######

unlink $template_file;
