#!/usr/bin/perl 

our $VERSION = 0.1;

package TestHelper::History;

use warnings;
use strict;

use Exporter;
our @ISA       = qw/Exporter/;
our @EXPORT_OK = qw/
    files_needing_retest
    load_test_history
    set_time_tested
    store_test_history
    /;

use FreezeThaw;
use Params::Validate;

use MVHub::Utils::ConfigSimple qw/create_config_from /;

my $cfg     = create_config_from( $ENV{MV_CONFIG_FILE} );
my $tmp_dir = $cfg->param('ABSOLUTE_PATH.tmp_dir');

$TestHelper::TEST_HISTORY_FILE = "$tmp_dir/test_history.txt";

sub files_needing_retest {
    my %a = Params::Validate::validate(
        @_,
        {   raw_files_aref => 1,
            test_name      => 1,
            history_href   => 1,
        }
    );
    my ( $raw_files_aref, $test_name, $history_href )
        = @a{qw/raw_files_aref test_name history_href /};

    # allow for files modified during lengthy test runs
    # allow file timestamps modified by version control
    my $SLOP = 20;

    my @retest_files;

    foreach my $filename (@$raw_files_aref) {
        my $modification_time = $SLOP + ( stat $filename )[9];

        next
            if exists $$history_href{$filename}
                && exists $$history_href{$filename}{$test_name}
                && $modification_time < $$history_href{$filename}{$test_name};

        push @retest_files, $filename;
    }
    return \@retest_files;
}

sub set_time_tested {
    my %a = Params::Validate::validate(
        @_,
        {   tested_files_aref => 1,
            test_name         => 1,
            history_href      => 1,
        }
    );
    my ( $tested_files_aref, $test_name, $history_href )
        = @a{qw/tested_files_aref test_name history_href /};

    my $test_time = time();
    foreach my $filename (@$tested_files_aref) {
        $$history_href{$filename}{$test_name} = $test_time;
    }

    return $history_href;
}

sub load_test_history {
    my $history_filename = shift || $TestHelper::TEST_HISTORY_FILE;

    return {} if !-e $history_filename;

    my $test_history = {};
    open( my $IN, $history_filename );
    my $raw_history = <$IN>;
    close $IN;

    my ($history_href) = FreezeThaw::thaw($raw_history);
    return $history_href;
}

sub store_test_history {
    my ( $thawed_history_href, $history_filename )
        = Params::Validate::validate_pos( @_, 1, 0 );
    $history_filename ||= $TestHelper::TEST_HISTORY_FILE;

    my $frozen_history = FreezeThaw::freeze($thawed_history_href);
    open my $OUT, '>', $history_filename;
    print $OUT $frozen_history;
    close $OUT;
}

# All good modules return true
1;

__DATA__

=head1 TestHelper::History
 
TestHelper::History - subs to avoid running lengthy tests on unchanged 
 
=head1 VERSION
 
This documentation refers to TestHelper::History version 0.1
 
=head1 SYNOPSIS

 #!/usr/bin/perl

 use strict; 
 use warnings;

 use Test::More;
 use TestHelper::History;
 
  # get a data structure with files 
  # and when tests were run on them
  my $test_history_href = TestHelper::History::load_test_history();

  my @raw_files = `find . -name '*.pm'`;
  chomp @raw_files;

  # given the name of the test,
  # check @raw_files to see if
  # that test has run since
  my $files_to_test_aref = TestHelper::History::files_needing_retest(
    test_name      => 'our_time_consuming_test',
    raw_files_aref => \@raw_files,
    history_href   => $test_history_href,
 );

    # run our tests, get back names of files we tested
    # ran_tests_on() is example only
    my @tested_files =ran_tests_on($files_to_test_aref);

 # update our history data structure with
 # time();
 $test_history_href = TestHelper::History::set_time_tested(
    tested_files_aref => \@tested_files,
    test_name         => 'our_time_consuming_test',
    history_href      => $test_history_href,
 );

 # save our results to disc
 TestHelper::History::store_test_history($test_history_href);

=head1 DESCRIPTION

There are several tests that take a long time to run. If we cache 
the results of of these tests, we don't have to re-run the tests
until the files that are tested change.
 
The cache file is removed by running the mv_clean command.
 
=head1 SUBROUTINES/METHODS 

=over

=item load_test_history()

=item load_test_history($file_with_history)

=back

  This function loads the cached history for all tested files, if you 
  don't supply a filename, it loads from a default file in the 
  application tmp  directory. If there is no file, it creates an empty data structure.

=over

=item files_needing_retest

=back

  my $files_to_test_aref = files_needing_retest(
    test_name      => 'our_time_consuming_test',
    raw_files_aref => \@raw_files,
    history_href   => $test_history_href,
 );

Given the name of the test, the files you want to check for past test runs and a history 
data structure, return an array reference to a list of files that need to be tested 
because their modification time is greater than the last time they were tested.

=over

=item set_time_tested

=back

 $test_history_href = set_time_tested(
    tested_files_aref => \@tested_files,
    test_name         => 'our_time_consuming_test',
    history_href      => $test_history_href,
 );

=over

=item store_test_history

=back

Set the time we ran this test so the next time we run tests, we won't have rerun them.

 store_test_history($test_history_href);

  save our results to disc

 
=head1 DIAGNOSTICS
 
TODO

A list of every error and warning message that the module can generate 
(even the ones that will "never happen"), with a full explanation of 
each problem, one or more likely causes, and any suggested remedies. 
(See also QUOTE \" " INCLUDETEXT "13_ErrorHandling" 
"XREF83683_Documenting_Errors_"\! Documenting Errors QUOTE \" " QUOTE " 
in Chapter " in Chapter INCLUDETEXT "13_ErrorHandling" "XREF40477__"\! 
13.)

=head1 INCOMPATIBILITIES
 
There are no known incompatibilites
 
=head1 BUGS AND LIMITATIONS
 
There are no known bugs in this module. 
Please report problems to help@thecsl.org
Patches are welcome.
 
=head1 AUTHOR

Community Software Lab help@thecsl.org 
 
=head1 LICENCE AND COPYRIGHT

Copyright (c) 2004-2010 Community Software Lab, Inc ( http://thecsl.org )

The GNU Affero General Public License, version 3.0 or later


=cut
